"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parkingCodesOverLapCheck = exports.extractNumberFromParkingCode = exports.separateDoubleAndSingleSpots = exports.compareOldAndNewParkingCodes = exports.generateRandomPassword = exports.isCoordinatesInRadius = exports.isDate90DaysOld = exports.isPasswordValid = exports.validatePassword = exports.validateEmail = void 0;
const bcrypt_1 = __importDefault(require("bcrypt"));
const emailRegex = /^[\w-]+(\.[\w-]+)*@([\w-]+\.)+[a-zA-Z]{2,7}$/;
const passwordRegex = /^(?=(?:.*[a-z]){1,})(?=(?:.*[A-Z]){1,})(?=(?:.*[!@#$%^&*()-+]){1,})(?=(?:.*[0-9]){1,}).{8,}$/;
const degreesToRadians = (degrees) => {
    return degrees * Math.PI / 180;
};
const distanceInKmBetweenEarthCoordinates = (lat1, lon1, lat2, lon2) => {
    const earthRadiusKm = 6371;
    const dLat = degreesToRadians(lat2 - lat1);
    const dLon = degreesToRadians(lon2 - lon1);
    lat1 = degreesToRadians(lat1);
    lat2 = degreesToRadians(lat2);
    const a = Math.sin(dLat / 2) * Math.sin(dLat / 2) + Math.sin(dLon / 2) * Math.sin(dLon / 2) * Math.cos(lat1) * Math.cos(lat2);
    const c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
    return earthRadiusKm * c;
};
const validateEmail = (email) => {
    return emailRegex.test(email);
};
exports.validateEmail = validateEmail;
const validatePassword = (password) => {
    // return passwordRegex.test(password)
    const hasLowercase = /[a-z]/.test(password);
    const hasUppercase = /[A-Z]/.test(password);
    const hasSpecialChar = /[!@#$%^&*]/.test(password);
    const hasNumber = /[0-9]/.test(password);
    const criteriaMet = [hasLowercase, hasUppercase, hasSpecialChar, hasNumber].filter(Boolean).length >= 3;
    return criteriaMet;
};
exports.validatePassword = validatePassword;
const isPasswordValid = (password, hashedPassword) => {
    return bcrypt_1.default.compareSync(password, hashedPassword);
};
exports.isPasswordValid = isPasswordValid;
const isDate90DaysOld = (targetDate) => {
    const ninetyDaysAgo = new Date();
    ninetyDaysAgo.setDate(ninetyDaysAgo.getDate() - 90);
    return targetDate <= ninetyDaysAgo;
};
exports.isDate90DaysOld = isDate90DaysOld;
const isCoordinatesInRadius = (center, plottedPoint, radius) => {
    const distance = distanceInKmBetweenEarthCoordinates(plottedPoint.latitude, plottedPoint.longitude, center.latitude, center.longitude);
    return distance <= radius;
};
exports.isCoordinatesInRadius = isCoordinatesInRadius;
const generateRandomPassword = () => {
    return new Promise((resolve, reject) => {
        const lowercaseChars = 'abcdefghijklmnopqrstuvwxyz';
        const uppercaseChars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
        const specialChars = '!@#$%^&*()-+';
        const numericChars = '0123456789';
        const allChars = lowercaseChars + uppercaseChars + specialChars + numericChars;
        const getRandomChar = (charSet) => charSet[Math.floor(Math.random() * charSet.length)];
        const getRandomFromSet = (charSets) => getRandomChar(charSets[Math.floor(Math.random() * charSets.length)]);
        const password = [lowercaseChars, uppercaseChars, specialChars, numericChars]
            .map(getRandomChar)
            .concat(Array.from({ length: 4 }, () => getRandomFromSet([lowercaseChars, uppercaseChars, specialChars, numericChars])))
            .sort(() => Math.random() - 0.5)
            .join('');
        resolve(password);
    });
};
exports.generateRandomPassword = generateRandomPassword;
const compareOldAndNewParkingCodes = (oldArray, newArray) => {
    const removedItems = oldArray.filter(item => !newArray.includes(item));
    const addedItems = newArray.filter(item => !oldArray.includes(item));
    const unchangedItems = oldArray.filter(item => newArray.includes(item));
    const changedItems = [];
    unchangedItems.forEach(item => {
        const oldIndex = oldArray.indexOf(item);
        const newIndex = newArray.indexOf(item);
        if (oldIndex !== newIndex) {
            changedItems.push(item);
        }
        // You can also add additional checks here if attributes of the items need updating
    });
    return {
        removed: removedItems,
        added: addedItems,
        unchanged: unchangedItems,
        changed: changedItems
    };
};
exports.compareOldAndNewParkingCodes = compareOldAndNewParkingCodes;
const separateDoubleAndSingleSpots = (array) => {
    const { singleSpots, doubleSpots } = array.reduce((acc, item) => {
        if (item.includes("-")) {
            const [prefix, suffix] = item.split("-");
            const index = acc.doubleSpots.findIndex((group) => group[0].startsWith(prefix));
            if (index !== -1) {
                acc.doubleSpots[index].push(item);
            }
            else {
                acc.doubleSpots.push([item]);
            }
        }
        else {
            acc.singleSpots.push(item);
        }
        return acc;
    }, { singleSpots: [], doubleSpots: [] });
    return { singleSpots, doubleSpots };
};
exports.separateDoubleAndSingleSpots = separateDoubleAndSingleSpots;
const extractNumberFromParkingCode = (str) => {
    var _a;
    const numberPart = (_a = str.match(/\d+/)) === null || _a === void 0 ? void 0 : _a[0];
    return numberPart ? numberPart : '';
};
exports.extractNumberFromParkingCode = extractNumberFromParkingCode;
const parkingCodesOverLapCheck = (arr1, arr2, arr3, arr4) => {
    const allItems = [...arr1, ...arr2, ...arr3, ...arr4];
    const uniqueItems = new Set(allItems);
    return uniqueItems.size !== allItems.length;
};
exports.parkingCodesOverLapCheck = parkingCodesOverLapCheck;
