import bcrypt from 'bcrypt'
import { Parser } from 'json2csv'
import path from 'path';
import { promises } from 'fs'


export const generateTemporaryPassword = (): { expiry: Date, password: string } => {
  const uppercaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  const lowercaseLetters = 'abcdefghijklmnopqrstuvwxyz';
  const digits = '0123456789';
  let result = '';
  let typesCount = 0; // Counter to track the number of character types used in the password
  // Function to check if the password contains characters from at least three different types
  const hasThreeTypes = () => {
    return (
      /[A-Z]/.test(result) && // Check for uppercase letters
      /[a-z]/.test(result) && // Check for lowercase letters
      /\d/.test(result) // Check for digits
    );
  };
  // Generate password until it contains characters from at least three different types
  while (!hasThreeTypes()) {
    result = '';
    typesCount = 0;
    for (let i = 0; i < 10; i++) {
      const randomIndex = Math.floor(Math.random() * 3); // Randomly select a character type
      switch (randomIndex) {
        case 0:
          result += uppercaseLetters.charAt(Math.floor(Math.random() * uppercaseLetters.length));
          typesCount++;
          break;
        case 1:
          result += lowercaseLetters.charAt(Math.floor(Math.random() * lowercaseLetters.length));
          typesCount++;
          break;
        case 2:
          result += digits.charAt(Math.floor(Math.random() * digits.length));
          typesCount++;
          break;
      }
    }
  }
  const date = new Date();
  date.setMinutes(date.getMinutes() + 30);
  return { expiry: date, password: result };
};

export const generateTemporaryPinCode = (): { expiry: Date, password: string } => {
  let result = '';
  for (let i = 0; i < 5; i++) {
    result += Math.floor(Math.random() * 10);
  }
  const date = new Date();
  date.setMinutes(date.getMinutes() + 30);
  return { expiry: date, password: result };
}

export const generateHashPassword = async (password: string): Promise<string> => {
  const salt = await bcrypt.genSalt(10)
  return await bcrypt.hash(password, salt)
}

export const generateLevelCode = (floorNumber: number): string => {
  return String.fromCharCode(floorNumber + 64)
}

export const generateParkingCodes = (floorNumber: number, numSpotsPerFloor: number, startFrom?: number): string[] => {
  const floorCodes: string[] = [];
  for (let spot = startFrom ? startFrom : 1; spot <= numSpotsPerFloor; spot++) {
    const floorCode = String.fromCharCode(floorNumber + 64);
    const spotCode = spot.toString().padStart(2, '0');
    const parkingCode = `${floorCode}${spotCode}`;
    floorCodes.push(parkingCode);
  }
  return floorCodes;
}

export const registrationSmsGenerator = (firstName: string, generateRandom: boolean, generatedPassword: { password: string, expiry: Date }, password: string): string => {
  return `
${firstName}, ברוך הבא לספוטפארק, אפליקציית ניהול החניונים החדשה של חברת החשמל.\n
הוראות ההתקנה והשימוש הועברו אליך במצגת במייל הארגוני.\n
יש להוריד את האפליקציה החדשה מהחנות , מצ"ב קישורים:\n
קישור לבעלי אייפון: https://apps.apple.com/us/app/spotpark/id6475667550 \n
קישור לבעלי אנדרואיד: https://play.google.com/store/apps/details?id=com.spotpark.ysb&pli=1 \n
לאחר הורדת האפליקציה החדשה מהחנות יש למלא את  שם המשתמש שלך (מספר הטלפון הנייד שלך) והסיסמה הזמנית שלך היא: ${generateRandom ? generatedPassword.password : password}.\n
לאחר מכן, יש לשנות סיסמא כפי שמוסבר במצגת שקיבלת.\n
בכל בעיה שאתה נתקל אתה מוזמן לפנות לאדמיניסטרטור שלך כפי שמופיע במצגת שקיבלת.\n
לאחר ההתחברות לאפליקציה החדשה יש להסיר את האפליקציה הישנה שיצאה משימוש.\n
שים לב כי מספר החניה שלך נשאר ללא שינוי.\n
בהצלחה לכולנו!`
}

export const doubleParkingNotificationTextGenerator = (registration: string, parkingLot: string, name: string, phoneNumber: { countryCode: string, number: string }) => {
  return `רכב ${registration} חוסם אותך בחניה כפולה בחניון ${parkingLot}. מספר הטלפון שלו הוא ${phoneNumber?.number}. שם הנהג: ${name}`
}

export const generateComplaintsReportCsv = async (complaints: any[]) => {
  const fields = [
    // { label: 'מזהה תלונה', value: '_id' },
    { label: 'שם חניון', value: 'parkingLot.name' },
    // { label: 'זהות המשתמש', value: 'complaintBy._id' },
    { label: 'שם מלא', value: 'complaintBy.firstName' },
    { label: 'טלפון', value: 'complaintBy.phoneNumber.number' },
    { label: 'מייל', value: 'complaintBy.email' },
    { label: 'תיאור הבעיה', value: 'details' },
    { label: 'סטטוס', value: 'status' },
    { label: 'נוצר ב', value: 'createdAt' },
    // { label: 'מעודכן ב', value: 'updatedAt' },
  ]
  const parser = new Parser({ fields, })
  const csv = parser.parse(complaints)
  const directory = path.join(__dirname, '../../../public/reports/complaints')
  await promises.mkdir(directory, { recursive: true })
  const filePath = path.join(directory, `complaint-report-${Date.now()}${Date.now()}.csv`)
  const createdPath = 'public/' + filePath?.split('public/')[1]
  const file = await promises.writeFile(filePath, csv, 'utf-8')
  return {
    file,
    filePath: createdPath,
  }
}

export const generateUserReportCsv = async (users: any[]) => {
  const newUsers = await Promise.all(users.map(async (user) => {
    const object = {
      ...user?.toObject(),
      isSpotFrozen: '',
      levels: ''
    }
    if (user?.parkingLots?.length > 0) {
      const parkingLots = await Promise.all(user?.parkingLots?.map((parkingLot: { name: string }) => parkingLot?.name))
      object.parkingLots = parkingLots.join(', ')
    } else object.parkingLots = ''
    if (user?.designatedParkingSpot?.length > 0) {
      const frozenArray: string[] = []
      const levelsArray: string[] = []
      const designatedParkingSpot = await Promise.all(user?.designatedParkingSpot?.map((spot: { parkingCode: string, isPermanent: boolean, level: { name: string } }) => {
        frozenArray.push(spot.isPermanent ? 'כן' : 'לא')
        // frozenArray.push(spot.isPermanent ? 'true' : 'false')
        levelsArray.push(spot.level.name)
        return spot?.parkingCode
      }))
      object.isSpotFrozen = frozenArray.join(', ')
      object.designatedParkingSpot = designatedParkingSpot.join(', ')
      object.levels = levelsArray.join(', ')
    } else object.designatedParkingSpot = ''
    object.deviceToken = user?.deviceToken ? true : false
    object.electricCharging = user?.electricCharging ? 'כן' : 'לא'
    object.disabled = user?.disabled ? 'כן' : 'לא'
    object.phoneNumber = user?.phoneNumber?.number
    return object
  }))
  const fields = [
    // { label: 'זהות המשתמש', value: '_id' },
    { label: 'שם מלא', value: 'firstName' },
    { label: 'טלפון', value: 'phoneNumber' },
    { label: 'חניונים', value: 'parkingLots' },
    { label: 'אזורים', value: 'levels' },
    { label: 'מספר חניה', value: 'designatedParkingSpot' },
    { label: 'קפואה?', value: 'isSpotFrozen' },
    { label: 'תאריך התחברות ראשון', value: 'firstLoginDate' },
    { label: 'חשמלי', value: 'electricCharging' },
    { label: 'נגיש', value: 'disabled' },
    { label: 'אישר התראות', value: 'deviceToken' },
    // { label: 'נוצר ב', value: 'createdAt' },
    // { label: 'מעודכן ב', value: 'updatedAt' },
  ]
  const parser = new Parser({ fields, })
  const csv = parser.parse(newUsers)
  const directory = path.join(__dirname, '../../../public/reports/users')
  await promises.mkdir(directory, { recursive: true })
  const filePath = path.join(directory, `users-report-${Date.now()}${Date.now()}.csv`)
  const createdPath = 'public/' + filePath?.split('public/')[1]
  const file = await promises.writeFile(filePath, csv, 'utf-8')
  return {
    file: file,
    filePath: createdPath,
  }
}

export const generateGuestUserReportCsv = async (users: any[]) => {
  const fields = [
    { label: 'שם מלא', value: 'firstName' },
    { label: 'מספר טלפון', value: 'phoneNumber.number' },
    { label: 'רישום רכב', value: 'vehicles' },
    { label: 'נגיש', value: 'disabled' },
    { label: 'מקום חניה שמור', value: 'designatedParkingSpot' },
    { label: 'תאריך הגעה', value: 'reservations' }
  ]
  const parser = new Parser({ fields, })
  const csv = parser.parse(users)
  const directory = path.join(__dirname, '../../../public/reports/guest-users')
  await promises.mkdir(directory, { recursive: true })
  const filePath = path.join(directory, `guest-users-report-${Date.now()}${Date.now()}.csv`)
  const createdPath = 'public/' + filePath?.split('public/')[1]
  const file = await promises.writeFile(filePath, csv, 'utf-8')
  return {
    file,
    filePath: createdPath,
  }
}

export const generateParkingLotReportCsv = async (parkingSpots: any[]) => {
  const fields = [
    { label: 'שם חניון', value: 'parkingLot.name' },
    { label: 'שם אזור', value: 'level.name' },
    { label: 'מספר חניה', value: 'parkingCode' },
    { label: 'סוג חניה', value: 'spotType' },
    { label: 'קפואה?', value: 'isPermanent' },
    { label: 'סטטוס', value: 'status' },
    { label: 'שמור למשתמש', value: 'designatedTo.firstName' },
    { label: 'תפוס על ידי', value: 'isOccupied' },
  ]
  const parser = new Parser({ fields, })
  const csv = parser.parse(parkingSpots)
  const directory = path.join(__dirname, '../../../public/reports/parking-lot')
  await promises.mkdir(directory, { recursive: true })
  const filePath = path.join(directory, `parking-lot-report-${Date.now()}${Date.now()}.csv`)
  const createdPath = 'public/' + filePath?.split('public/')[1]
  const file = await promises.writeFile(filePath, csv, 'utf-8')
  return {
    file,
    filePath: createdPath,
  }
}