import { Request, } from 'express';
import { addTime, getCorrectDateTime, getJewishHolidayEndDate, getStartAndEndOfDay, getTimeDifference, getTodayTomorrowYesterday, isJewishHolidayShabbat, isTimeBetween } from './dates';
import { ObjectId } from 'mongodb';
import ParkingSpot, { IParkingSpotInterface } from '../models/parking-spot';
import Reservation from '../models/reservation';
import Absence from '../models/absence';
import { Schema } from 'mongoose';
import User, { IUserInterface } from '../models/user';
import Log from '../models/log';
import LogTypes from '../services/constants/log-types';
import { sendNotification } from '../middlewares/notifications';
import { NotificationTypes } from '../services/constants/notification';
import { IHoliday } from '../models/holiday';
import { reservationEnums } from '../v2.9.3/reservations/controllers';
// Assuming availableSpots is an array of objects with a property named 'parkingCode'
// Define a custom sorting function
export const customSort = (a: any, b: any) => {
    // Extract the numeric parts of the parking codes
    const numA = parseInt(a.parkingCode.substring(1));
    const numB = parseInt(b.parkingCode.substring(1));
    // Extract the alphabetical parts of the parking codes
    const alphaA = a.parkingCode.substring(0, 1);
    const alphaB = b.parkingCode.substring(0, 1);
    // First, sort alphabetically
    if (alphaA < alphaB) return -1;
    if (alphaA > alphaB) return 1;
    // If the alphabetical parts are equal, sort numerically
    if (numA < numB) return -1;
    if (numA > numB) return 1;
    // If both parts are equal, retain the original order
    return 0;
};

export const customSortDescending = (a: any, b: any) => {
    // Extract the numeric parts of the parking codes
    const numA = parseInt(a.parkingCode.substring(1));
    const numB = parseInt(b.parkingCode.substring(1));
    // Extract the alphabetical parts of the parking codes
    const alphaA = a.parkingCode.substring(0, 1);
    const alphaB = b.parkingCode.substring(0, 1);
    // First, sort alphabetically in descending order
    if (alphaA > alphaB) return -1;
    if (alphaA < alphaB) return 1;
    // If the alphabetical parts are equal, sort numerically in descending order
    if (numA > numB) return -1;
    if (numA < numB) return 1;
    // If both parts are equal, retain the original order
    return 0;
}

export const customSortNameAscending = (a: any, b: any) => {
    // Handle null values in designatedTo
    if (a.designatedTo === null && b.designatedTo === null) {
        return 0; // If both are null, retain the original order
    } else if (a.designatedTo === null) {
        return -1; // If only a is null, place it before b
    } else if (b.designatedTo === null) {
        return 1; // If only b is null, place it before a
    } else {
        // Compare by designatedTo.firstName
        return a.designatedTo.firstName.localeCompare(b.designatedTo.firstName);
    }
};

export const customSortNameDescending = (a: any, b: any) => {
    // Handle null values in designatedTo
    if (a.designatedTo === null && b.designatedTo === null) {
        return 0; // If both are null, retain the original order
    } else if (a.designatedTo === null) {
        return 1; // If only a is null, place it after b
    } else if (b.designatedTo === null) {
        return -1; // If only b is null, place it after a
    } else {
        // Compare by designatedTo.firstName in reverse order
        return b.designatedTo.firstName.localeCompare(a.designatedTo.firstName);
    }
};

export const customSortNameReservedByAscending = (a: any, b: any) => {
    // Handle null values in designatedTo
    if (a.reservedBy === null && b.reservedBy === null) {
        return 0; // If both are null, retain the original order
    } else if (a.reservedBy === null) {
        return -1; // If only a is null, place it before b
    } else if (b.reservedBy === null) {
        return 1; // If only b is null, place it before a
    } else {
        // Compare by designatedTo.firstName
        return a.reservedBy.firstName.localeCompare(b.reservedBy.firstName);
    }
};

export const customSortNameReservedByDescending = (a: any, b: any) => {
    // Handle null values in designatedTo
    if (a.reservedBy === null && b.reservedBy === null) {
        return 0; // If both are null, retain the original order
    } else if (a.reservedBy === null) {
        return 1; // If only a is null, place it after b
    } else if (b.reservedBy === null) {
        return -1; // If only b is null, place it after a
    } else {
        // Compare by designatedTo.firstName in reverse order
        return b.reservedBy.firstName.localeCompare(a.reservedBy.firstName);
    }
};

export const customSortPhoneNumberAscending = (a: any, b: any) => {
    if (a.reservedBy === null) return 1;
    if (b.reservedBy === null) return -1;
    if (a.reservedBy.phoneNumber.number < b.reservedBy.phoneNumber.number) return -1;
    if (a.reservedBy.phoneNumber.number > b.reservedBy.phoneNumber.number) return 1;
    return 0;
}

export const customSortPhoneNumberDescending = (a: any, b: any) => {
    if (a.reservedBy === null) return 1;
    if (b.reservedBy === null) return -1;
    if (a.reservedBy.phoneNumber.number < b.reservedBy.phoneNumber.number) return 1;
    if (a.reservedBy.phoneNumber.number > b.reservedBy.phoneNumber.number) return -1;
    return 0;
}

export const customSortPhoneNumberDesignatedToAscending = (a: any, b: any) => {
    if (a.designatedTo === null) return 1;
    if (b.designatedTo === null) return -1;
    if (a.designatedTo.phoneNumber.number < b.designatedTo.phoneNumber.number) return -1;
    if (a.designatedTo.phoneNumber.number > b.designatedTo.phoneNumber.number) return 1;
    return 0;
}

export const customSortPhoneNumberDesignatedToDescending = (a: any, b: any) => {
    if (a.designatedTo === null && b.designatedTo === null) return 0;
    if (a.designatedTo === null) return -1;
    if (b.designatedTo === null) return 1;
    if (a.designatedTo.phoneNumber.number < b.designatedTo.phoneNumber.number) return 1;
    if (a.designatedTo.phoneNumber.number > b.designatedTo.phoneNumber.number) return -1;
    return 0;
}


// Custom sort function for sorting in ascending order based on "status"
export const customSortStatusAscending = (a: any, b: any) => {
    console.log(a.status, b.status, 'customSortStatusAscending')
    const statusOrder: any = {
        'reserved': 4,
        'released': 3,
        'free': 2,
        'designated-not-released': 1
    };
    return statusOrder[a.status] - statusOrder[b.status];
};

// Custom sort function for sorting in descending order based on "status"
export const customSortStatusDescending = (a: any, b: any) => {
    const statusOrder: any = {
        'reserved': 1,
        'released': 2,
        'free': 3,
        'designated-not-released': 4
    };
    return statusOrder[b.status] - statusOrder[a.status];
};

export const findAvailableSpots = async (req: Request, level?: string | Schema.Types.ObjectId): Promise<{ totalCount: number; availableSpots: typeof ParkingSpot[]; }> => {
    const { _id }: { _id?: string } = req.params;
    const { reservationDate, type, isDisabled, isElectric, }: { level?: string, reservationDate?: Date, type?: string, isAvailable?: boolean | string, isDisabled?: boolean | string, isElectric?: boolean | string } = req.query
    const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate ? reservationDate : new Date())
    const sort: {} = { parkingCode: 1 }
    let filter: any = {
        level: level ? level : new ObjectId(_id),
        $or: [{ isPermanent: false }, { isPermanent: undefined }]
    }
    if (type != undefined) filter.spotType = type
    if (isDisabled == 'false') filter.spotType = { $nin: ['disabled'] };
    if (isElectric == 'false') filter.spotType = { $nin: ['electric'] };
    if (isDisabled == 'false' && isElectric == 'false') filter.spotType = { $nin: ['disabled', 'electric'] };
    let outerWithNoInnerReserved: number = 0
    const totalSpots = await ParkingSpot.find(filter).sort(sort)
    const reservedSpots = await Reservation.find({ level: level ? level : new ObjectId(_id), reservationDate: { $gte: startOfDay, $lte: endOfDay }, parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] } })
    const availableSpots = await Promise.all(totalSpots.map(async (spot: any) => {
        const isIncludedInReserved = reservedSpots.findIndex((reservedSpot: any) => reservedSpot.parkingSpot.toString() == spot._id.toString())
        // If No Reservation
        // if (isIncludedInReserved < 0) {
        //     const isOuter = spot.outerToSpot != null
        //     console.log(isOuter, 'isOuter')
        //     // If 17-B
        //     if (isOuter) {
        //         // Check Reservation on 17-A
        //         const isReservationOnInner = await Promise.resolve(reservedSpots.findIndex((reservedSpot: any) => reservedSpot.parkingSpot.toString() == spot.outerToSpot.toString()))
        //         // If Reservation on 17-A, return 17-B
        //         console.log(isReservationOnInner, 'isReservationOnInner')
        //         if (isReservationOnInner >= 0) return spot
        //         else outerWithNoInnerReserved += 1      // else return nothing and increase count by one for double spots i.e returned will be 17-A and count will be +1 for double
        //     } else {
        //         if (spot.spotType == 'designated' && spot.designatedTo != null) {
        //             const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate ? reservationDate : new Date())
        //             const spotAbsence = await Absence.findOne({ designatedSpot: spot._id, $and: [{ startDate: { $lte: reservationDate }, endDate: { $gte: reservationDate } },] })
        //             const isUserPresent = spot.userPresenceDate >= startOfDay && spot.userPresenceDate <= endOfDay
        //             if (spotAbsence && !isUserPresent) return spot
        //         } else return spot
        //     }
        // }
        if (isIncludedInReserved < 0) {
            if (spot.spotType == 'designated' && spot.designatedTo != null) {
                const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate ? reservationDate : new Date())
                const spotAbsence = await Absence.findOne({ designatedSpot: spot._id, $and: [{ startDate: { $lte: reservationDate }, endDate: { $gte: reservationDate } },] })
                const isUserPresent = spot.userPresenceDate >= startOfDay && spot.userPresenceDate <= endOfDay
                if (spotAbsence && !isUserPresent) return spot
            } else {
                const isOuter = spot.outerToSpot != null
                // If 17-B
                if (isOuter) {
                    // Check Reservation on 17-A
                    const isReservationOnInner = await Promise.resolve(reservedSpots.findIndex((reservedSpot: any) => reservedSpot.parkingSpot.toString() == spot.outerToSpot.toString()))
                    // If Reservation on 17-A, return 17-B
                    if (isReservationOnInner >= 0) return spot
                    else outerWithNoInnerReserved += 1
                } else return spot
            }
        }
    }))
    const filteredSpots = await Promise.all(availableSpots.filter((spot: any) => spot != null))
    const sorted = filteredSpots.sort(customSort)
    return { availableSpots: sorted, totalCount: sorted.length + outerWithNoInnerReserved }
}

const getFirstRegistration = (item: any): string | null => {
    if (item.reservedBy && item.reservedBy.vehicles.length > 0) {
        console.log('reservedBy', item.reservedBy.vehicles)
        return item.reservedBy.vehicles[0].registration;
    } else if (item.designatedTo && item.designatedTo.vehicles.length > 0) {
        console.log('designatedTo', item.designatedTo.vehicles)
        return item.designatedTo.vehicles[0].registration;
    } else if ((item.reservedBy && item.reservedBy.vehicles.length === 0) || (item.designatedTo && item.designatedTo.vehicles.length === 0)) {
        return '';
    }
    return null;
};

export const customSortRegistrationAscending = (a: any, b: any) => {
    const aReg = getFirstRegistration(a);
    const bReg = getFirstRegistration(b);

    if (aReg === null) return 1;
    if (bReg === null) return -1;
    if (aReg === '') return 1;
    if (bReg === '') return -1;
    return aReg.localeCompare(bReg);
}

export const customSortRegistrationDescending = (a: any, b: any) => {
    const aReg = getFirstRegistration(a);
    const bReg = getFirstRegistration(b);

    if (aReg === null) return -1;
    if (bReg === null) return 1;
    if (aReg === '') return -1;
    if (bReg === '') return 1;
    return bReg.localeCompare(aReg);
}

const sendNotificationForDesignatedSpotConfirmation = async (user: any, spot: any) => {
    if (user?.deviceToken && spot?.parkingLot?.messageSendTime) {
        {
            try {
                const notification = user?.deviceToken ? await sendNotification(user?.deviceToken,
                    // `נא לאשר האם אתה משתמש חניה השמורה שלך בחניון ${spot?.parkingLot?.name}`,
                    `נא לאשר האם אתה משתמש בחניה השמורה שלך בחניון  ${ spot?.parkingLot?.name}`,
                    NotificationTypes.designatedSpotConfirmation, null, [user?._id]) : null
                await Log.create({
                    user: user._id,
                    parkingLot: spot?.parkingLot?._id,
                    level: spot?.level,
                    parkingSpot: spot?._id,
                    status: 'success',
                    request: {
                        messageSendTime: spot?.parkingLot?.messageSendTime,
                        absenceSendTime: spot?.parkingLot?.absenceSendTime,
                        userPresenceDate: spot?.userPresenceDate,
                        deviceToken: user?.deviceToken,
                    },
                    response: notification,
                    type: LogTypes.designatedCronJob.notificationSent,
                })
                if (!process.env.DB_COLLECTION?.includes('-')) console.log('designated notification send status', notification)
            } catch (error) {
                await Log.create({
                    user: user._id,
                    parkingLot: spot?.parkingLot?._id,
                    level: spot?.level,
                    parkingSpot: spot?._id,
                    status: 'error',
                    request: {
                        messageSendTime: spot?.parkingLot?.messageSendTime,
                        absenceSendTime: spot?.parkingLot?.absenceSendTime,
                        userPresenceDate: spot?.userPresenceDate,
                    },
                    type: LogTypes.designatedCronJob.notificationFailedToSend,
                })
            }
        }
    } else {
        await Log.create({
            user: user._id,
            parkingLot: spot?.parkingLot?._id,
            level: spot?.level,
            parkingSpot: spot?._id,
            status: 'error',
            request: {
                messageSendTime: spot?.parkingLot?.messageSendTime,
                absenceSendTime: spot?.parkingLot?.absenceSendTime,
                userPresenceDate: spot?.userPresenceDate,
            },
            type: LogTypes.designatedCronJob.notificationDetailsMissing,
        })
    }
}

const timeMatchCheck = (spot: { parkingLot: { messageSendTime: string, absenceSendTime: string } }, time: string, sendMessage: boolean, createAbsence: boolean) => {
    return (spot?.parkingLot?.messageSendTime == time && sendMessage) || (spot?.parkingLot?.absenceSendTime == time && createAbsence)
}
const userPresenceCheck = (spot: { _id: ObjectId, parkingLot: { reservationTime: 'same-day' | 'day-before' }, userPresenceDate: Date }) => {
    const { today, tomorrow } = getTodayTomorrowYesterday()
    const { startOfDay, endOfDay } = getStartAndEndOfDay(spot?.parkingLot?.reservationTime == 'same-day' ? today : tomorrow)
    return (spot.userPresenceDate == null || !(spot.userPresenceDate >= startOfDay && endOfDay >= spot.userPresenceDate))
}
const absenceCheck = async (spot: { _id: ObjectId, parkingLot: { reservationTime: 'same-day' | 'day-before' } }) => {
    const { today, tomorrow } = getTodayTomorrowYesterday()
    return await Absence.findOne({
        designatedSpot: spot?._id,
        startDate: { $lte: spot?.parkingLot?.reservationTime == 'same-day' ? today : tomorrow },
        endDate: { $gte: spot?.parkingLot?.reservationTime == 'same-day' ? today : tomorrow },
        status: 'pending'
    })
}

// export const performDesignatedSpotAction = async (time: string, sendMessage: boolean, createAbsence: boolean) => {
//     try {
//         const today = getCorrectDateTime(new Date())
//         const yesterday = getCorrectDateTime(new Date())
//         yesterday.setUTCDate(yesterday.getUTCDate() - 1)
//         const shabbatCheck = await isJewishHolidayShabbat(today)
//         const holiday = await getJewishHolidayEndDate(today)
//         const holidayYesterday = await getJewishHolidayEndDate(yesterday)
//         if (shabbatCheck && shabbatCheck.isShabbat) {
//             console.log('It is Shabbat, not doing anything')
//         } else {
//             const users = await User.find({
//                 $and: [{ designatedParkingSpot: { $exists: true, $ne: null } }, { designatedParkingSpot: { $ne: [] } }]
//             }).populate({
//                 path: 'designatedParkingSpot',
//                 populate: {
//                     path: 'parkingLot',
//                     select: ['messageSendTime', 'reservationTime', 'absenceSendTime', 'name']
//                 }
//             })
//             if (users?.length > 0) {
//                 await Promise.all(users.map(async (user: IUserInterface) => {
//                     try {
//                         const spotsNoPresence: any[] = [];
//                         if (user.designatedParkingSpot?.length > 0) {
//                             await Promise.all(user.designatedParkingSpot.map(async (spot: any) => {
//                                 try {
//                                     // Check All Conditions
//                                     const findUserAbsenceForToday = await absenceCheck(spot)
//                                     const isTimeMatching = timeMatchCheck(spot, time, sendMessage, createAbsence)
//                                     const isUserAbsentOnDate = userPresenceCheck(spot)
//                                     // Check if the spot is valid for notification/absence
//                                     const isSpotValidForNotification = isTimeMatching && findUserAbsenceForToday == null && isUserAbsentOnDate
//                                     // Check if Holiday was today and spot is valid
//                                     const isHolidaySpotValid = holiday && findUserAbsenceForToday == null && isUserAbsentOnDate

//                                     if (!process.env.DB_COLLECTION?.includes('-')) {
//                                         if (!holiday) {
//                                             console.log(createAbsence ? `isSpotValidForAbsence ${spot?._id}` : `isSpotValidForNotification ${spot?._id}`, isSpotValidForNotification)
//                                             console.log('--------------------------------')
//                                         }
//                                         console.log(spot?.parkingLot?.messageSendTime, 'spot?.parkingLot?.messageSendTime', spot?.parkingLot?.name)
//                                         console.log('--------------------------------')
//                                     }
//                                     if (isSpotValidForNotification && !spot?.isPermanent) {
//                                         const index = spotsNoPresence.findIndex(item => item._id == spot._id)
//                                         if (index < 0) {
//                                             await Log.create({
//                                                 user: user._id,
//                                                 parkingLot: spot?.parkingLot?._id,
//                                                 level: spot?.level,
//                                                 parkingSpot: spot?._id,
//                                                 status: 'success',
//                                                 request: {
//                                                     sendNotification: sendMessage,
//                                                     createAbsence,
//                                                     isSpotValidForNotification,
//                                                     currentTime: time,
//                                                     messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                     absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                     userPresenceDate: spot.userPresenceDate,
//                                                     spotsNoPresence,
//                                                     spotsNoPresenceIndex: index,
//                                                 },
//                                                 type: sendMessage ? LogTypes.designatedCronJob.notificationValidity : LogTypes.designatedCronJob.absenceValidity,
//                                             })
//                                             spotsNoPresence.push(spot)
//                                         } else {
//                                             await Log.create({
//                                                 user: user._id,
//                                                 parkingLot: spot?.parkingLot?._id,
//                                                 level: spot?.level,
//                                                 parkingSpot: spot?._id,
//                                                 status: 'error',
//                                                 request: {
//                                                     sendNotification: sendMessage,
//                                                     createAbsence,
//                                                     isSpotValidForNotification,
//                                                     currentTime: time,
//                                                     messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                     absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                     userPresenceDate: spot.userPresenceDate,
//                                                     spotsNoPresenceIndex: index,
//                                                     spotsNoPresence,
//                                                 },
//                                                 type: sendMessage ? LogTypes.designatedCronJob.notificationValidity : LogTypes.designatedCronJob.absenceValidity,
//                                             })
//                                         }
//                                     } else if (isHolidaySpotValid && !spot?.isPermanent) {
//                                         const parkingLotReservationTime = getTimeDifference(spot?.parkingLot?.messageSendTime, holiday.endTime)
//                                         const isParkingLotTimePassed = parkingLotReservationTime.minutes < 0 ? true : false
//                                         if (spot?.parkingLot?.reservationTime == 'same-day' && isParkingLotTimePassed) {
//                                             await Log.create({
//                                                 user: user._id,
//                                                 parkingLot: spot?.parkingLot?._id,
//                                                 level: spot?.level,
//                                                 parkingSpot: spot?._id,
//                                                 status: 'error',
//                                                 request: {
//                                                     sendNotification: sendMessage,
//                                                     createAbsence,
//                                                     isSpotValidForNotification,
//                                                     currentTime: time,
//                                                     messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                     absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                     userPresenceDate: spot.userPresenceDate,
//                                                     spotsNoPresence,
//                                                     isParkingLotTimePassed,
//                                                     parkingLotReservationTime,
//                                                     holiday,
//                                                 },
//                                                 type: LogTypes.designatedCronJob.parkingLotTimePassedOnHoliday,
//                                             })
//                                             console.log('Parking lot time has passed for same-day and today holiday ended at', holiday.endTime)
//                                             console.log('--------------------------------')
//                                         }
//                                         else {
//                                             const timeDifference = getTimeDifference(time, holiday.endTime)
//                                             if (timeDifference.minutes == 60 && sendMessage) {
//                                                 await Log.create({
//                                                     user: user._id,
//                                                     parkingLot: spot?.parkingLot?._id,
//                                                     level: spot?.level,
//                                                     parkingSpot: spot?._id,
//                                                     status: 'success',
//                                                     request: {
//                                                         sendNotification: sendMessage,
//                                                         createAbsence,
//                                                         isSpotValidForNotification,
//                                                         currentTime: time,
//                                                         messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                         absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                         userPresenceDate: spot.userPresenceDate,
//                                                         spotsNoPresence,
//                                                         timeDifference,
//                                                         holiday,
//                                                     },
//                                                     type: sendMessage ? LogTypes.designatedCronJob.notificationValidityHoliday : LogTypes.designatedCronJob.absenceValidityHoliday,
//                                                 })
//                                                 spotsNoPresence.push(spot)
//                                                 if (!process.env.DB_COLLECTION?.includes('-')) {
//                                                     console.log(isHolidaySpotValid, 'isHolidaySpotValidForNotification', spot?._id)
//                                                     console.log('--------------------------------')
//                                                 }
//                                             } else if (createAbsence) {
//                                                 const timerDuration = getTimeDifference(spot?.parkingLot.absenceSendTime, spot?.parkingLot.messageSendTime)
//                                                 const timeAfterOneHourOfShabbat = addTime(holiday.endTime, '01:00')
//                                                 const durationEndTime = addTime(timeAfterOneHourOfShabbat, timerDuration.formatted)
//                                                 if (durationEndTime == time) {
//                                                     spotsNoPresence.push(spot)
//                                                 }
//                                                 await Log.create({
//                                                     user: user._id,
//                                                     parkingLot: spot?.parkingLot?._id,
//                                                     level: spot?.level,
//                                                     parkingSpot: spot?._id,
//                                                     status: durationEndTime == time ? 'success' : 'error',
//                                                     request: {
//                                                         sendNotification: sendMessage,
//                                                         createAbsence,
//                                                         isSpotValidForNotification,
//                                                         isHolidaySpotValid,
//                                                         currentTime: time,
//                                                         messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                         absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                         userPresenceDate: spot.userPresenceDate,
//                                                         spotsNoPresence,
//                                                         timeDifference,
//                                                         holiday,
//                                                     },
//                                                     type: LogTypes.designatedCronJob.absenceValidityHoliday,
//                                                 })
//                                                 // console.log(isHolidaySpotValid, 'isHolidaySpotValidForAbsence', spot?._id)
//                                                 // console.log('--------------------------------')
//                                             } else {
//                                                 if (!process.env.DB_COLLECTION?.includes('-')) {
//                                                     console.log(createAbsence ? `isSpotValidForAbsenceOnHoliday ${spot?._id}` : `isSpotValidForNotificationOnHoliday ${spot?._id}`, isSpotValidForNotification)
//                                                     console.log('--------------------------------')
//                                                 }
//                                                 await Log.create({
//                                                     user: user._id,
//                                                     parkingLot: spot?.parkingLot?._id,
//                                                     level: spot?.level,
//                                                     parkingSpot: spot?._id,
//                                                     status: 'error',
//                                                     request: {
//                                                         sendNotification: sendMessage,
//                                                         createAbsence,
//                                                         isSpotValidForNotification,
//                                                         isHolidaySpotValid,
//                                                         currentTime: time,
//                                                         messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                         absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                         userPresenceDate: spot.userPresenceDate,
//                                                         spotsNoPresence,
//                                                         timeDifference,
//                                                         holiday,
//                                                     },
//                                                     type: LogTypes.designatedCronJob.notificationValidityHoliday,
//                                                 })
//                                             }
//                                         }
//                                     } else if (spot?.parkingLot?.messageSendTime == time || spot?.parkingLot?.absenceSendTime == time) {
//                                         if (!process.env.DB_COLLECTION?.includes('-')) {
//                                             console.log(`Not eligible to send ${sendMessage ? 'Notification' : 'Absence'}`, spot?._id)
//                                             console.log('--------------------------------')
//                                         }
//                                         await Log.create({
//                                             user: user._id,
//                                             parkingLot: spot?.parkingLot?._id,
//                                             level: spot?.level,
//                                             parkingSpot: spot?._id,
//                                             status: 'error',
//                                             request: {
//                                                 sendNotification: sendMessage,
//                                                 createAbsence: createAbsence,
//                                                 isSpotValidForNotification,
//                                                 isHolidaySpotValid,
//                                                 currentTime: time,
//                                                 messageSendTime: spot?.parkingLot?.messageSendTime,
//                                                 absenceSendTime: spot?.parkingLot?.absenceSendTime,
//                                                 userPresenceDate: spot.userPresenceDate,
//                                                 spotsNoPresence,
//                                                 findUserAbsenceForToday,
//                                                 isSpotPermanent: spot?.isPermanent,
//                                             },
//                                             type: sendMessage ? LogTypes.designatedCronJob.notificationNotEligible : LogTypes.designatedCronJob.absenceNotEligible,
//                                         })
//                                     }
//                                 } catch (err) {
//                                     console.log('Error in mapping user designated spots in performAbsenceAction', err)
//                                 }
//                             }))
//                             if (spotsNoPresence.length > 0) {
//                                 if (sendMessage) {
//                                     console.log('Sending notification to user: ' + user?._id)
//                                     await Promise.all(spotsNoPresence.map(async (spot: any) => {
//                                         if (user?.deviceToken) await sendNotificationForDesignatedSpotConfirmation(user, spot)
//                                         else console.log(user?._id, 'does not have device token')
//                                     }))
//                                 }
//                                 if (createAbsence && user?._id && user?.designatedParkingSpot.length > 0) {
//                                     await Promise.all(spotsNoPresence.map(async (spot: any) => {
//                                         console.log('Sending absence to user: ' + user?._id)
//                                         const dateForAbsence = new Date()
//                                         if (spot?.parkingLot?.reservationTime == 'day-before') dateForAbsence.setUTCDate(dateForAbsence.getUTCDate() + 1)
//                                         const { startOfDay, endOfDay } = getStartAndEndOfDay(dateForAbsence)
//                                         const absence = new Absence({
//                                             user: user?._id,
//                                             designatedSpot: spot?._id,
//                                             startDate: startOfDay,
//                                             endDate: endOfDay,
//                                             note: 'המשתמש לא הגיב בתוך פרק הזמן שצוין.',
//                                             type: 'system-generated'
//                                         })
//                                         await absence.save()
//                                         await ParkingSpot.findByIdAndUpdate(spot?._id, { userPresenceDate: null });
//                                     }));
//                                 }
//                             }
//                         }
//                     } catch (err) {
//                         console.log('Error in mapping users in performAbsenceAction', err)
//                     }
//                 }));
//             }
//             if (!process.env.DB_COLLECTION?.includes('-')) {
//                 console.log('All spots checked', users?.length);
//                 console.log('===================');
//                 console.log('===================');
//             }
//         }
//     } catch (err) {
//         console.log('Error executing performAbsenceAction', err)
//     }
// }

const processUserDesignatedSpots = async (user: IUserInterface, time: string, sendMessage: boolean, createAbsence: boolean, holiday: IHoliday | null, holidayYesterday: IHoliday | null) => {
    const spotsNoPresence: any[] = [];
    if (user.designatedParkingSpot?.length > 0) {
        await Promise.all(user.designatedParkingSpot.map(async (spot: any) => {
            try {
                // Check All Conditions
                const findUserAbsenceForToday = await absenceCheck(spot)
                const isTimeMatching = timeMatchCheck(spot, time, sendMessage, createAbsence)
                const isUserAbsentOnDate = userPresenceCheck(spot)
                let parkingLotReservationTime = { minutes: 0, formatted: '00:00' }
                if (holiday) parkingLotReservationTime = getTimeDifference(spot?.parkingLot?.messageSendTime, holiday.endTime)
                const isParkingLotTimePassed = parkingLotReservationTime.minutes < 0 ? true : false
                // Check if the spot is valid for notification/absence
                const isSpotValidForNotification = isTimeMatching && findUserAbsenceForToday == null && isUserAbsentOnDate
                // Check if Holiday was today and spot is valid
                const isHolidaySpotValid = holiday && findUserAbsenceForToday == null && isUserAbsentOnDate
                // Check if the time is between mid night and 3 hours after
                const isTimeAfterMidNight = isTimeBetween(time, '00:00', '03:00')
                // Check if parking lot notification is on
                const isParkingLotNotificationAllowed = spot?.parkingLot?.sendPushNotification != false ? true : false
                if (isSpotValidForNotification && isParkingLotNotificationAllowed && !spot?.isPermanent && !isParkingLotTimePassed) {
                    const index = spotsNoPresence.findIndex(item => item._id == spot._id)
                    if (index < 0) {
                        // Log here
                        spotsNoPresence.push(spot)
                    } else {
                        // Log here
                    }
                } else if (isHolidaySpotValid && isParkingLotNotificationAllowed && !spot?.isPermanent) {
                    if (spot?.parkingLot?.reservationTime == 'same-day' && isParkingLotTimePassed) {
                        console.log('Parking lot time has passed for same-day and today holiday ended at', holiday.endTime)
                        console.log('--------------------------------')
                    } else {
                        const timeDifference = getTimeDifference(time, holiday.endTime)
                        if (timeDifference.minutes == 60 && isParkingLotTimePassed && sendMessage) {
                            spotsNoPresence.push(spot)
                            if (!process.env.DB_COLLECTION?.includes('-')) {
                                console.log(isHolidaySpotValid, 'isHolidaySpotValidForNotification', spot?._id)
                                console.log('--------------------------------')
                            }
                        } else if (createAbsence) {
                            const timerDuration = getTimeDifference(spot?.parkingLot.absenceSendTime, spot?.parkingLot.messageSendTime)
                            const timeAfterOneHourOfShabbat = addTime(holiday.endTime, '01:00')
                            const durationEndTime = addTime(timeAfterOneHourOfShabbat, timerDuration.formatted)
                            if (durationEndTime == time) {
                                spotsNoPresence.push(spot)
                            }
                            // console.log(isHolidaySpotValid, 'isHolidaySpotValidForAbsence', spot?._id)
                            // console.log('--------------------------------')
                        } else {
                            if (!process.env.DB_COLLECTION?.includes('-')) {
                                console.log(createAbsence ? `isSpotValidForAbsenceOnHoliday ${spot?._id}` : `isSpotValidForNotificationOnHoliday ${spot?._id}`, isSpotValidForNotification)
                                console.log('--------------------------------')
                            }
                        }
                    }
                } else if (spot?.parkingLot?.messageSendTime == time || spot?.parkingLot?.absenceSendTime == time) {
                    if (!process.env.DB_COLLECTION?.includes('-')) {
                        console.log(`Not eligible to send ${sendMessage ? 'Notification' : 'Absence'}`, spot?._id)
                        console.log('--------------------------------')
                    }
                } else if (isTimeAfterMidNight && holidayYesterday && createAbsence && isParkingLotNotificationAllowed) {
                    const timeAfterOneHourOfShabbat = addTime(holidayYesterday.endTime, '01:00')
                    const durationEndTime = addTime(timeAfterOneHourOfShabbat, spot?.parkingLot?.designatedReservationDuration)
                    const isDurationEndTimeAfterMidNight = isTimeBetween(durationEndTime, '00:00', '03:00')
                    // console.log(isDurationEndTimeAfterMidNight,'isDurationEndTimeAfterMidNight')
                    // console.log(durationEndTime,'durationEndTime')
                    // console.log(time,'time')
                    if (isDurationEndTimeAfterMidNight) {
                        if (durationEndTime == time) {
                            // spotsNoPresence.push(spot)
                            console.log('Sending absence to user: ' + user?._id)
                            const dateForAbsence = new Date()
                            const { startOfDay, endOfDay } = getStartAndEndOfDay(dateForAbsence)
                            const absence = new Absence({
                                user: user?._id,
                                designatedSpot: spot?._id,
                                startDate: startOfDay,
                                endDate: endOfDay,
                                note: 'המשתמש לא הגיב בתוך פרק הזמן שצוין.',
                                type: 'system-generated'
                            })
                            await absence.save()
                            await ParkingSpot.findByIdAndUpdate(spot?._id, { userPresenceDate: null });
                        }
                    }
                }
            } catch (err) {
                console.log('Error in mapping user designated spots in performAbsenceAction', err)
            }
        }))
        if (spotsNoPresence.length > 0) {
            if (sendMessage) {
                console.log('Sending notification to user: ' + user?._id)
                await Promise.all(spotsNoPresence.map(async (spot: any) => {
                    if (user?.deviceToken) await sendNotificationForDesignatedSpotConfirmation(user, spot)
                    else console.log(user?._id, 'does not have device token')
                }))
            }
            if (createAbsence && user?._id && user?.designatedParkingSpot.length > 0) {
                await Promise.all(spotsNoPresence.map(async (spot: any) => {
                    console.log('Sending absence to user: ' + user?._id)
                    const dateForAbsence = new Date()
                    if (spot?.parkingLot?.reservationTime == 'day-before') dateForAbsence.setUTCDate(dateForAbsence.getUTCDate() + 1)
                    const { startOfDay, endOfDay } = getStartAndEndOfDay(dateForAbsence)
                    const absence = new Absence({
                        user: user?._id,
                        designatedSpot: spot?._id,
                        startDate: startOfDay,
                        endDate: endOfDay,
                        note: 'המשתמש לא הגיב בתוך פרק הזמן שצוין.',
                        type: 'system-generated'
                    })
                    await absence.save()
                    await ParkingSpot.findByIdAndUpdate(spot?._id, { userPresenceDate: null });
                }));
            }
        }
    }
}

export const performDesignatedSpotAction = async (time: string, sendMessage: boolean, createAbsence: boolean) => {
    try {
        const { today, yesterday } = getTodayTomorrowYesterday()
        const shabbatCheck = await isJewishHolidayShabbat(today)
        const holiday = await getJewishHolidayEndDate(today)
        const holidayYesterday = await getJewishHolidayEndDate(yesterday)
        if (shabbatCheck && shabbatCheck.isShabbat) {
            console.log('It is Shabbat, not doing anything')
        } else {
            const users = await User.find({ $and: [{ designatedParkingSpot: { $exists: true, $ne: null } }, { designatedParkingSpot: { $ne: [] } }] })
                .populate({
                    path: 'designatedParkingSpot',
                    populate: {
                        path: 'parkingLot',
                        select: ['name', 'reservationTime', 'messageSendTime', 'absenceSendTime', 'designatedReservationDuration', 'sendPushNotification',],
                    }
                })
            if (users?.length > 0) {
                await Promise.all(users.map(async (user: IUserInterface) => {
                    try {
                        await processUserDesignatedSpots(user, time, sendMessage, createAbsence, holiday, holidayYesterday)
                    } catch (err) {
                        console.log('Error in mapping users in performAbsenceAction', err)
                    }
                }));
            }
            if (!process.env.DB_COLLECTION?.includes('-')) {
                console.log('All spots checked', users?.length);
                console.log('===================');
                console.log('===================');
            }
        }
    } catch (err) {
        console.log('Error executing performAbsenceAction', err)
    }
}