import bcrypt from 'bcrypt'
const emailRegex = /^[\w-]+(\.[\w-]+)*@([\w-]+\.)+[a-zA-Z]{2,7}$/
const passwordRegex = /^(?=(?:.*[a-z]){1,})(?=(?:.*[A-Z]){1,})(?=(?:.*[!@#$%^&*()-+]){1,})(?=(?:.*[0-9]){1,}).{8,}$/

const degreesToRadians = (degrees: number): number => {
    return degrees * Math.PI / 180;
}

const distanceInKmBetweenEarthCoordinates = (lat1: number, lon1: number, lat2: number, lon2: number): number => {
    const earthRadiusKm = 6371
    const dLat = degreesToRadians(lat2 - lat1)
    const dLon = degreesToRadians(lon2 - lon1)
    lat1 = degreesToRadians(lat1)
    lat2 = degreesToRadians(lat2)
    const a = Math.sin(dLat / 2) * Math.sin(dLat / 2) + Math.sin(dLon / 2) * Math.sin(dLon / 2) * Math.cos(lat1) * Math.cos(lat2)
    const c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a))
    return earthRadiusKm * c;
}

export const validateEmail = (email: string): boolean => {
    return emailRegex.test(email)
}
export const validatePassword = (password: string): boolean => {
    // return passwordRegex.test(password)
    const hasLowercase = /[a-z]/.test(password);
    const hasUppercase = /[A-Z]/.test(password);
    const hasSpecialChar = /[!@#$%^&*]/.test(password);
    const hasNumber = /[0-9]/.test(password);
    const criteriaMet = [hasLowercase, hasUppercase, hasSpecialChar, hasNumber].filter(Boolean).length >= 3;
    return criteriaMet;
}

export const isPasswordValid = (password: string, hashedPassword: string): boolean => {
    return bcrypt.compareSync(password, hashedPassword)
}

export const isDate90DaysOld = (targetDate: Date): boolean => {
    const ninetyDaysAgo = new Date();
    ninetyDaysAgo.setDate(ninetyDaysAgo.getDate() - 90);
    return targetDate <= ninetyDaysAgo;
}

export const isCoordinatesInRadius = (center: { latitude: number, longitude: number }, plottedPoint: { latitude: number, longitude: number }, radius: number): boolean => { // Radius in KMs
    const distance = distanceInKmBetweenEarthCoordinates(plottedPoint.latitude, plottedPoint.longitude, center.latitude, center.longitude)
    return distance <= radius
}

export const generateRandomPassword = (): Promise<string> => {
    return new Promise((resolve, reject) => {
        const lowercaseChars = 'abcdefghijklmnopqrstuvwxyz';
        const uppercaseChars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
        const specialChars = '!@#$%^&*()-+';
        const numericChars = '0123456789';
        const allChars = lowercaseChars + uppercaseChars + specialChars + numericChars;
        const getRandomChar = (charSet: string): string => charSet[Math.floor(Math.random() * charSet.length)];
        const getRandomFromSet = (charSets: string[]): string => getRandomChar(charSets[Math.floor(Math.random() * charSets.length)]);
        const password = [lowercaseChars, uppercaseChars, specialChars, numericChars]
            .map(getRandomChar)
            .concat(Array.from({ length: 4 }, () => getRandomFromSet([lowercaseChars, uppercaseChars, specialChars, numericChars])))
            .sort(() => Math.random() - 0.5)
            .join('');
        resolve(password)
    })
}

export const compareOldAndNewParkingCodes = (oldArray: string[], newArray: string[]): { removed: string[], added: string[], unchanged: string[], changed: string[] } => {
    const removedItems: string[] = oldArray.filter(item => !newArray.includes(item));
    const addedItems: string[] = newArray.filter(item => !oldArray.includes(item));
    const unchangedItems: string[] = oldArray.filter(item => newArray.includes(item));
    const changedItems: string[] = [];
    unchangedItems.forEach(item => {
        const oldIndex = oldArray.indexOf(item);
        const newIndex = newArray.indexOf(item);
        if (oldIndex !== newIndex) {
            changedItems.push(item);
        }
        // You can also add additional checks here if attributes of the items need updating
    });
    return {
        removed: removedItems,
        added: addedItems,
        unchanged: unchangedItems,
        changed: changedItems
    };
}

export const separateDoubleAndSingleSpots = (array: string[]): { singleSpots: string[], doubleSpots: string[][] } => {
    const { singleSpots, doubleSpots }: any = array.reduce((acc: any, item) => {
        if (item.includes("-")) {
            const [prefix, suffix] = item.split("-");
            const index = acc.doubleSpots.findIndex((group: any) => group[0].startsWith(prefix));
            if (index !== -1) {
                acc.doubleSpots[index].push(item);
            } else {
                acc.doubleSpots.push([item]);
            }
        } else {
            acc.singleSpots.push(item);
        }
        return acc;
    }, { singleSpots: [], doubleSpots: [] });
    return { singleSpots, doubleSpots };
}

export const extractNumberFromParkingCode = (str: string): string => {
    const numberPart = str.match(/\d+/)?.[0];
    return numberPart ? numberPart : '';
}

export const parkingCodesOverLapCheck = (arr1: string[], arr2: string[], arr3: string[], arr4: string[]): boolean => {      // Returns true if overlapping
    const allItems = [...arr1, ...arr2, ...arr3, ...arr4];
    const uniqueItems = new Set(allItems);
    return uniqueItems.size !== allItems.length;
}