import cron from "node-cron";
import { Schema } from "mongoose";
import {
  getCorrectDateTime,
  getStartAndEndOfDay,
  getTodayTomorrowYesterday,
} from "../../helpers/dates";
import Reservation from "../../models/reservation";
import moment from "moment";
import User from "../../models/user";
import Absence from "../../models/absence";
import { sendNotification } from "../../middlewares/notifications";
import { NotificationTypes } from "../constants/notification";
import Announcement from "../../models/announcement";
import { performDesignatedSpotAction } from "../../helpers";
import { reservationEnums } from "../../v2.9.3/reservations/controllers";
import ParkingSpot from "../../models/parking-spot";

export const findUsersWithMultipleReservations = async (
  getReservations: boolean
): Promise<any[]> => {
  try {
    const { startOfDay, endOfDay } = getStartAndEndOfDay(new Date());
    const filters = [
      {
        $match: {
          reservationDate: {
            $gte: new Date(startOfDay),
            $lte: new Date(endOfDay),
          },
          parkingStatus: { $in: ["reserved", "occupied"] },
        },
      },
      {
        $lookup: {
          from: "users", // Assuming 'users' is the name of the User collection
          localField: "user",
          foreignField: "_id",
          as: "userData",
        },
      },
      {
        $unwind: "$userData", // Flatten the userData array
      },
      {
        $group: {
          _id: "$user",
          count: { $sum: 1 },
          user: { $first: "$userData" }, // Include user data in the result
          reservations: getReservations && { $push: "$$ROOT" },
        },
      },
      {
        $match: {
          count: { $gt: 1 },
        },
      },
      {
        $project: {
          _id: 1,
          count: 1,
          user: {
            _id: 1,
            email: 1,
            phoneNumber: 1,
          },
          reservations: {
            $cond: {
              if: getReservations,
              then: "$reservations",
              else: [], // Set reservations to null if count is not greater than 1
            },
          },
        },
      },
    ];
    const reservations = await Reservation.aggregate(filters);
    return reservations;
  } catch (err) {
    console.log(err);
    return [];
  }
};
const findAbsencesEndedLastNight = async () => {
  const today = getCorrectDateTime(new Date());
  await Absence.updateMany(
    { endDate: { $lte: today } },
    { status: "completed" }
  );
};

const testingLastDayUsers = async () => {
  console.log("Here");
};

const findReservationsWithElectricLimit = async (time: string) => {
  try {
    const today = new Date();
    const { startOfDay, endOfDay } = getStartAndEndOfDay(today);
    const reservations = await Reservation.find({
      forMoving: false,
      parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] },
      reservationDate: { $gte: startOfDay, $lte: endOfDay },
    })
      .populate({
        path: "parkingLot",
        select: ["chargingTimeLimit", "chargingTimeEnd"],
        // match: {
        //     $and: [
        //         { parkingLot: { $exists: true } },
        //         { chargingTimeLimit: true },
        //         { chargingTimeEnd: time }
        //     ]
        // },
      })
      .populate({
        path: "user",
        select: ["phoneNumber", "deviceToken"],
      })
      .populate({
        path: "moveTo",
        select: ["parkingSpot"],
        populate: {
          path: "parkingSpot",
          select: "parkingCode",
        },
      });
    if (reservations?.length > 0) {
      await Promise.all(
        reservations.map(async (reservation: any) => {
          if (
            reservation?.user?.deviceToken &&
            reservation?.parkingLot &&
            reservation?.parkingLot?.chargingTimeEnd == time &&
            reservation?.moveTo?.parkingSpot?.parkingCode
          ) {
            const notification = reservation?.user?.deviceToken
              ? await sendNotification(
                  reservation?.user?.deviceToken,
                  "הזמן המוקצה עבור החניה החשמלית שלך הסתיים, אנא פנה את הרכב למקום חניה פנוי",
                  NotificationTypes.electricReservationTimeLimitEnd,
                  { reservation: String(reservation?._id) },
                  [reservation?.user?._id]
                )
              : null;
            // console.log(notification, 'findReservationsWithElectricLimitNotification')
            console.log(
              `notification send status(${notification}) \nfor reservation ID: ` +
                reservation?._id
            );
          }
        })
      );
      // console.log(reservations?.length, 'findReservationsWithElectricLimit')
    }
  } catch (error) {
    console.log(error, "error while finding electric with limit");
  }
};
const findAnnouncementsWithTime = async () => {
  const currentDate = getCorrectDateTime(new Date());
  let filter: any = {
    startDate: currentDate,
  };
  const announcements = await Announcement.find(filter).sort({ updatedAt: -1 });
  if (announcements.length > 0) {
    const announcementsText: string[] = [];
    const parkingLots = await Promise.all(
      announcements.map(announcement => {
        announcementsText.push(announcement.content);
        return announcement.parkingLots;
      })
    );
    if (parkingLots.length > 0) {
      const allParkingLots = parkingLots.flat();
      const users = await User.find({
        parkingLots: { $in: allParkingLots },
        $and: [
          { deviceToken: { $ne: null } },
          { deviceToken: { $ne: undefined } },
        ],
      });
      if (users.length > 0) {
        const deviceTokens: string[] = [];
        const userIds: Schema.Types.ObjectId[] = [];
        await Promise.all(
          users.map(user => {
            userIds?.push(user?._id);
            if (user.deviceToken != null) deviceTokens.push(user.deviceToken);
          })
        );
        announcementsText.map(async content => {
          const notifications = await sendNotification(
            deviceTokens,
            content,
            NotificationTypes.announcement,
            null,
            userIds
          );
        });
      }
    }
  }
};

const deleteGuestUsers = async () => {
  const { yesterday } = getTodayTomorrowYesterday();
  const { endOfDay } = getStartAndEndOfDay(yesterday);
  const filter: any = { reservationDate: { $lte: endOfDay } };
  const users = await User.find({
    isGuest: true,
  });
  users.forEach(async user => {
    const reservations = await Reservation.find({ user: user?._id, ...filter });
    await Promise.all(
      reservations?.map(async reservation => {
        await ParkingSpot.findByIdAndUpdate(reservation?.parkingSpot, {
          designatedTo: null,
          userPresenceDate: null,
        });
        await Reservation.findByIdAndDelete(reservation?._id);
      })
    );
    const remainingReservations = await Reservation.find({ user: user?._id });
    if (!remainingReservations.length) {
      await User.deleteOne({ _id: user._id });
    }
  });
};

const scheduleCronJob = () => {
  const oneAmTask = cron.schedule("0 1 * * *", async () => {
    console.log("Cron Job Started, 1am");
    await findAbsencesEndedLastNight();
    await deleteGuestUsers();
  });
  const everyHourTask = cron.schedule("0 * * * *", async () => {
    console.log("Cron Job Started, hourly,", getCorrectDateTime(new Date()));
    const date = moment();
    await findReservationsWithElectricLimit(date.format("HH:mm"));
    // const currentHour = date.format('HH:mm')
    // const threeHoursBefore = date.subtract(3, 'hours').format('HH:mm')
    // console.log('performAbsenceAction sending notification for', currentHour)
    // console.log('performAbsenceAction creating absence for', threeHoursBefore)
    // await performAbsenceAction(currentHour, true)
    // await performAbsenceAction(threeHoursBefore, false, true)
  });

  const everyMinuteTask = cron.schedule("* * * * *", async () => {
    console.log(
      "Cron Job Started, every minute,",
      getCorrectDateTime(new Date())
    );
    const date = moment();
    const currentTime = date.format("HH:mm");
    await findAnnouncementsWithTime();
    await performDesignatedSpotAction(currentTime, true, false);
    await performDesignatedSpotAction(currentTime, false, true);
  });

  console.log("Cron job scheduled");
};

export { scheduleCronJob };
