import { Request, Response, } from 'express';
import { Schema, Types } from 'mongoose';
import HttpMessages from '../../services/constants/http-messages';
import HttpStatus from '../../services/constants/http-statuses';
import Admin from '../../models/admin';
import ParkingLot from '../../models/parking-lot';
import Level from '../../models/lot-level';
import ParkingSpot from '../../models/parking-spot';
import Reservation from '../../models/reservation';
import { getStartAndEndOfDay } from '../../helpers/dates';
import User from '../../models/user';
import { reservationEnums } from '../reservations/controllers';
import { customSort } from '../../helpers';
import Absence from '../../models/absence';
import { generateParkingLotReportCsv } from '../../helpers/generators';

export const createParkingLotService = async (req: Request, res: Response) => {
    try {
        const {
            name,
            chargingTimeLimit,
            chargingTimeStart,
            chargingTimeEnd,
            location,
            admins,
            allotmentMethod,
            allotmentFile,
            accessMapFile,
            avatar,
            maxReservationDistance,
            messageSendTime,
            absenceSendTime,
            messageSendTimeOfDay,
            reservationTime,
            smsSendTime,
            designatedReservationDuration,
            sendPushNotification,
        }: {
            name?: string,
            chargingTimeLimit?: boolean,
            chargingTimeStart?: string,
            chargingTimeEnd?: string,
            location?: { latitude?: number, longitude?: number },
            admins?: Types.ObjectId[],
            allotmentMethod?: string,
            allotmentFile?: string,
            accessMapFile?: string,
            avatar?: string,
            maxReservationDistance: number,
            messageSendTime: string,
            absenceSendTime: string,
            messageSendTimeOfDay: 'before-evening' | 'morning',
            reservationTime: 'day-before' | 'same-day',
            smsSendTime: string,
            designatedReservationDuration?: string,
            sendPushNotification?: boolean
        } = req.body;
        if (name &&
            (chargingTimeLimit == true || chargingTimeLimit == false) &&
            location &&
            admins &&
            avatar &&
            maxReservationDistance > 0 &&
            allotmentMethod &&
            accessMapFile &&
            messageSendTime &&
            absenceSendTime &&
            designatedReservationDuration &&
            // messageSendTimeOfDay &&
            reservationTime
        ) {
            const parkingLot = new ParkingLot({
                name,
                avatar,
                chargingTimeLimit,
                chargingTimeStart,
                chargingTimeEnd,
                location,
                allotmentMethod,
                allotmentFile: allotmentFile ? allotmentFile : null,
                accessMapFile,
                admins,
                maxReservationDistance,
                messageSendTime,
                absenceSendTime,
                messageSendTimeOfDay,
                reservationTime,
                smsSendTime,
                designatedReservationDuration,
                sendPushNotification,
            })
            await Promise.all(admins.map(async (admin: Types.ObjectId) => {
                const saved = await Admin.findByIdAndUpdate(admin, { $push: { parkingLots: parkingLot._id } })
                return saved
            }))
            await parkingLot.save()
            // let createdLevels: any[] = []
            if (parkingLot.allotmentMethod == 'manual') {
                // createdLevels = await createManualSpotsHelper(parkingLot._id, parkingLot)
                // res.status(HttpStatus.CREATED).json({ parkingLot, levels: createdLevels });
                res.status(HttpStatus.CREATED).json({ parkingLot });
            } else res.status(HttpStatus.CREATED).json({ parkingLot });
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.parkingLot.PARKING_LOT_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAllParkingLotsService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10, currentAdmin }: { page?: number, limit?: number, currentAdmin?: string, } = req.query
        let filter: any = {}
        if (currentAdmin) {
            const admin = await Admin.findById(currentAdmin)
            if (admin?.role == 'sub-admin') {
                filter.admins = { $in: [currentAdmin] }
            }
        }
        const totalCount = await ParkingLot.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        const parkingLots = await ParkingLot.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
        if (parkingLots.length > 0) {
            const counts = await Promise.all(parkingLots.map(async (parkingLot) => {
                // const lot = await Reservation.countDocuments({ parkingLot: parkingLot._id })
                const regular = await ParkingSpot.countDocuments({ parkingLot: parkingLot._id, spotType: 'regular' })
                const designated = await ParkingSpot.countDocuments({ parkingLot: parkingLot._id, spotType: 'designated' })
                const electric = await ParkingSpot.countDocuments({ parkingLot: parkingLot._id, spotType: 'electric' })
                const disabled = await ParkingSpot.countDocuments({ parkingLot: parkingLot._id, spotType: 'disabled' })
                return {
                    ...parkingLot.toObject(),
                    regular,
                    designated,
                    electric,
                    disabled,
                }
            }))
            res.status(HttpStatus.OK).json({ totalPages, totalCount, parkingLots: counts })
        } else res.status(HttpStatus.OK).json({ totalPages, totalCount, parkingLots: [] })
        // res.status(HttpStatus.OK).json({ totalPages, totalCount, parkingLots });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getReportService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { startDate = new Date(), endDate = new Date(), }: { startDate?: Date, endDate?: Date, } = req.query
        let filter: any = { parkingLot: _id }
        const parkingLot = await ParkingLot.findById(_id)
        if (parkingLot) {
            const sort: {} = { parkingCode: 1 }
            const { startOfDay, } = getStartAndEndOfDay(startDate)
            const { endOfDay } = getStartAndEndOfDay(endDate)
            const parkingSpots: any = await ParkingSpot.find(filter).sort(sort).populate({ path: 'designatedTo', select: ['firstName'], }).populate({ path: 'level', select: ['name'] }).populate({ path: 'parkingLot', select: ['name'] })
            const reservations: any = await Reservation.find({ parkingLot: _id, parkingStatus: { $nin: [reservationEnums[3], reservationEnums[4]] }, reservationDate: { $gte: startOfDay, $lte: endOfDay } }).select(['reservationDate', 'parkingSpot']).populate({ path: 'user', select: ['firstName'] })
            const finalSpots = await Promise.all(parkingSpots.map(async (spot: any) => {
                const reservationIndex = await Promise.resolve(reservations.findIndex((reservation: any) => String(reservation.parkingSpot) == String(spot._id)))
                const spotType = spot?.spotType == 'regular' ? 'רגיל' : spot?.spotType == 'designated' ? 'שמור' : spot?.spotType == 'electric' ? 'חשמלי' : 'נגיש'
                if (reservationIndex < 0) return {
                    ...spot?.toObject(),
                    spotType,
                    reservation: null,
                    status: spot.designatedTo ? 'לא פנוי' : 'פנוי',
                    // status: spot.designatedTo ? 'שמור' : 'פנוי',
                    isOccupied: 'לא תפוסה',
                    isPermanent: spot?.isPermanent ? 'כן' : 'לא'
                    // status: spot.designatedTo ? 'designated-no-reservation' : 'available-for-reservation',
                }
                else return {
                    ...spot?.toObject(),
                    spotType,
                    reservation: reservations[reservationIndex],
                    status: 'לא פנוי',
                    isOccupied: reservations[reservationIndex]?.user?.firstName,
                    // isOccupied: 'תפוס',
                    isPermanent: spot?.isPermanent ? 'כן' : 'לא'
                    // status: 'reserved',
                }
            }))
            const { filePath } = await generateParkingLotReportCsv(finalSpots)
            res.status(HttpStatus.OK).json({ path: process.env.LIVE_BASE_URL + filePath })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
    } catch (err) {
        console.log('error', err)
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSingleParkingLotService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const parkingLot = await ParkingLot.findById(_id).populate({ path: 'admins', select: ['email', 'firstName', 'lastName', 'role'] })
        if (parkingLot) res.status(HttpStatus.OK).json(parkingLot);
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAdminParkingLotsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { page = 1, limit = 10 }: { page?: number, limit?: number } = req.query
        const admin = await Admin.findById(_id).populate('parkingLots').select('parkingLots').sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
        if (admin) {
            const { parkingLots } = admin
            res.status(HttpStatus.OK).json({ parkingLots })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getUserParkingLotsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { page = 1, limit = 10 }: { page?: number, limit?: number } = req.query
        const user = await User.findById(_id)
            .populate({
                path: 'parkingLots',
                options: { sort: { updatedAt: -1 } }
            })
        if (user) {
            let { parkingLots }: any = user
            const mainParkingLotIndex = parkingLots.findIndex((lot: any) => lot._id.equals(user.mainParkingLot))
            if (mainParkingLotIndex !== -1) {
                const mainParkingLot = parkingLots.splice(mainParkingLotIndex, 1)[0]
                parkingLots.unshift(mainParkingLot)
            }
            res.status(HttpStatus.OK).json({ parkingLots })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getUserParkingLotAdminsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const user = await User.findById(_id)
            .populate({
                path: 'parkingLots',
                options: { sort: { updatedAt: -1 }, },
                select: ['admins'],
                populate: { path: 'admins', select: ['phoneNumber', 'firstName', 'lastName', 'middleName', 'role'] },
            })
        if (user) {
            let { parkingLots }: any = user
            const mainParkingLotIndex = parkingLots.findIndex((lot: any) => lot._id.equals(user.mainParkingLot))
            if (mainParkingLotIndex !== -1) {
                const mainParkingLot = parkingLots.splice(mainParkingLotIndex, 1)[0]
                parkingLots.unshift(mainParkingLot)
            }
            const admins: any[] = []
            await Promise.all(parkingLots.map(async (lot: any) => {
                await Promise.all(lot.admins.map((admin: any) => admins.push(admin)))
            }))
            res.status(HttpStatus.OK).json(admins)
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getParkingLotCountsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { page = 1, limit = 10, }: { page?: number, limit?: number, } = req.query
        const { startOfDay, endOfDay } = getStartAndEndOfDay(new Date())
        let filter: any = {}
        if (_id) {
            const admin = await Admin.findById(_id)
            if (admin?.role == 'sub-admin') {
                filter.admins = { $in: [admin._id] }
            }
        }
        const parkingLots = await ParkingLot.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1)).select(['name', 'location',])
        const totalCount = await ParkingLot.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        if (parkingLots.length > 0) {
            const counts = await Promise.all(parkingLots.map(async (parkingLot) => {
                const lot = await Reservation.countDocuments({ parkingLot: parkingLot._id, reservationDate: { $gte: startOfDay, $lte: endOfDay }, parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] } })
                const total = await ParkingSpot.countDocuments({ parkingLot: parkingLot._id, })
                return {
                    parkingLot,
                    total: total,
                    reserved: lot,
                }
            }))
            res.status(HttpStatus.OK).json({ totalPages, totalCount, counts })
        } else res.status(HttpStatus.OK).json([])
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const updateParkingLotService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const {
            name,
            chargingTimeLimit,
            chargingTimeStart,
            chargingTimeEnd,
            location,
            admins,
            allotmentMethod,
            allotmentFile,
            accessMapFile,
            avatar,
            maxReservationDistance,
            smsSendTime,
            messageSendTime,
            absenceSendTime,
            messageSendTimeOfDay,
            reservationTime,
            designatedReservationDuration,
            sendPushNotification,
        }: {
            name?: string,
            chargingTimeLimit?: boolean,
            chargingTimeStart?: string,
            chargingTimeEnd?: string,
            location?: { latitude?: number, longitude?: number },
            admins?: Types.ObjectId[],
            allotmentMethod?: string,
            allotmentFile?: string,
            accessMapFile?: string,
            avatar?: string,
            maxReservationDistance?: number,
            smsSendTime?: string,
            messageSendTime: string,
            absenceSendTime: string,
            messageSendTimeOfDay: 'before-evening' | 'morning',
            reservationTime: 'day-before' | 'same-day',
            designatedReservationDuration?: string,
            sendPushNotification?: boolean;
        } = req.body;
        const beforeUpdate = await ParkingLot.findById(_id)
        const parkingLot = await ParkingLot.findByIdAndUpdate(_id, {
            name,
            chargingTimeLimit,
            chargingTimeStart,
            chargingTimeEnd,
            location,
            admins,
            allotmentMethod,
            allotmentFile,
            accessMapFile,
            avatar,
            maxReservationDistance,
            smsSendTime,
            messageSendTime,
            absenceSendTime,
            messageSendTimeOfDay,
            reservationTime,
            designatedReservationDuration,
            sendPushNotification,
        }, { omitUndefined: true, new: true, })
        if (parkingLot) {
            if (admins) {
                if (beforeUpdate && beforeUpdate?.admins?.length > 0) {
                    await Promise.all(beforeUpdate?.admins.map(async (admin: Schema.Types.ObjectId) => {
                        const removing = await Admin.findByIdAndUpdate(admin, { $pull: { parkingLots: parkingLot._id } }, { new: true })
                        return removing
                    }))
                }
                await Promise.all(admins.map(async (admin: Types.ObjectId) => {
                    const saved = await Admin.findByIdAndUpdate(admin, { $push: { parkingLots: parkingLot._id } }, { new: true })
                    return saved
                }))
            }
            res.status(HttpStatus.OK).json(parkingLot)
        }
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const deleteParkingLotService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const parkingLotReservations = await Reservation.countDocuments({ parkingLot: _id, parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] } })
        const designatedAssigned = await ParkingSpot.countDocuments({ parkingLot: _id, designatedTo: { $ne: null } })
        if (parkingLotReservations > 0) res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.parkingLot.DELETE_FAILURE_REASON_UNCOMPLETED_RESERVATION })
        else if (designatedAssigned > 0) res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.parkingLot.DELETE_FAILURE_REASON_DESIGNATED })
        else {
            const parkingLot = await ParkingLot.findByIdAndDelete(_id);
            if (parkingLot) {
                await Level.deleteMany({ parkingLot: _id })
                await ParkingSpot.deleteMany({ parkingLot: _id })
                await ParkingLot.updateMany({ admins: { $in: parkingLot.admins } }, { $pull: { parkingLots: _id } })
                // await Promise.all(parkingLot.admins.map(async (admin: Types.ObjectId) => {
                //     const removed = await Admin.findByIdAndUpdate(admin, { $pull: { parkingLots: _id } });
                // }))
                res.status(HttpStatus.OK).json({ message: HttpMessages.parkingLot.PARKING_LOT_DELETED_SUCCESS })
            }
            else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}