import { Request, Response, } from 'express';
import { addTime, getCorrectDateTime, getJewishHolidayEndDate, getTimeDifference, getTodayTomorrowYesterday, isJewishHolidayShabbat, isTimeBetween, } from '../../helpers/dates/index';
import HttpMessages from '../../services/constants/http-messages';
import HttpStatus from '../../services/constants/http-statuses';
import { getStartAndEndOfDay, } from '../../helpers/dates'
import { isCoordinatesInRadius } from '../../helpers/validators'
import ParkingLot, { IParkingLotInterface } from '../../models/parking-lot';
import ParkingSpot, { IParkingSpotInterface } from '../../models/parking-spot';
import Reservation from '../../models/reservation';
import { Schema, } from 'mongoose';
import { sendMessage } from '../../middlewares/email-messaging';
import User from '../../models/user';
import moment from 'moment';
import Absence from '../../models/absence';
import { sendNotification } from '../../middlewares/notifications';
import { NotificationTypes } from '../../services/constants/notification';
import { ObjectId } from 'mongodb';
import { doubleParkingNotificationTextGenerator } from '../../helpers/generators';
import { reservationEnums } from './controllers';

const findAvailableSpots = async (req: Request, spotType: 'regular' | 'designated' | 'electric' | 'disabled', level?: Schema.Types.ObjectId) => {
    if (spotType == 'designated') return await findAvailableDesignatedSpots(req)
    else {
        const { parkingLot, reservationDate, }: { parkingLot: string, user: string, reservationDate: Date, } = req.body
        const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
        const query: any = {
            parkingLot,
            spotType,
            isPermanent: false,
        }
        if (level) query.level = level
        const parkingLotObject = await ParkingLot.findById(parkingLot)
        if (parkingLotObject) {
            const sort: {} = parkingLotObject.allotmentMethod == 'manual' ? { _id: 1 } : { parkingCode: 1 }
            // const sort: {} = parkingLotObject.allotmentMethod == 'manual' ? { createdAt: 1 } : { parkingCode: 1 }
            const totalSpots = await ParkingSpot.find(query).sort(sort)
            const reservedSpots = await Reservation.find({ parkingLot, reservationDate: { $gte: startOfDay, $lte: endOfDay }, parkingStatus: { $in: ['reserved', 'occupied'] } })
            const availableSpots: any[] = await Promise.all(totalSpots.filter((spot: any) => {
                const isIncludedInReserved = reservedSpots.findIndex((reservedSpot: any) => reservedSpot.parkingSpot.toString() == spot._id.toString())
                if (isIncludedInReserved < 0) return spot
            }))
            return { availableSpots, reservedSpots }
        } else return { availableSpots: [], reservedSpots: [] }
    }
}

const findAvailableDesignatedSpots = async (req: Request) => {
    const { parkingLot, reservationDate, }: { parkingLot: string, user: string, reservationDate: Date, } = req.body
    const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
    const query = {
        parkingLot,
        spotType: 'designated',
        isPermanent: false,
    }
    const parkingLotObject = await ParkingLot.findById(parkingLot)
    if (parkingLotObject) {
        const sort: {} = parkingLotObject.allotmentMethod == 'manual' ? { _id: 1 } : { parkingCode: 1 }
        // const sort: {} = parkingLotObject.allotmentMethod == 'manual' ? { createdAt: 1 } : { parkingCode: 1 }
        const totalSpots = await ParkingSpot.find(query).sort(sort)
        const reservedSpots = await Reservation.find({ parkingLot, reservationDate: { $gte: startOfDay, $lte: endOfDay }, parkingStatus: { $in: ['reserved', 'occupied'] } })
        const possibleAvailable: any[] = await Promise.all(totalSpots.filter((spot: any) => {
            const isIncludedInReserved = reservedSpots.findIndex((reservedSpot: any) => reservedSpot.parkingSpot.toString() == spot._id.toString())
            if (isIncludedInReserved < 0) return spot
        }))
        const availableSpots: IParkingSpotInterface[] = []
        if (possibleAvailable.length > 0) {
            await Promise.all(possibleAvailable.map(async (spot: IParkingSpotInterface) => {
                const findAbsence = await Absence.find({ designatedSpot: spot._id, startDate: { $lte: reservationDate }, endDate: { $gte: reservationDate }, status: 'pending' })
                if (findAbsence.length > 0 || spot.designatedTo == null) { availableSpots.push(spot) }
            }))
        }
        return { availableSpots, reservedSpots }
    } else return { availableSpots: [], reservedSpots: [] }
}

const createElectricReservationWithTimeLimit = async (electricSpot: IParkingSpotInterface, otherAvailableSpot: IParkingSpotInterface, lot: IParkingLotInterface, req: Request, res: Response) => {
    const { parkingLot, user, reservationDate, isElectricChargingRequired, isDisabled, language = "heb" }: { parkingLot: string, user: string, reservationDate: Date, isElectricChargingRequired: boolean, isDisabled: boolean, location: { latitude: number, longitude: number }, language?: string, } = req.body
    const electricReservation = new Reservation({
        user,
        parkingLot,
        level: electricSpot.level,
        parkingSpot: electricSpot._id,
        reservationDate: new Date(reservationDate),
        isElectricChargingRequired,
        isDisabled: isDisabled ? isDisabled : false,
        parkingStatus: 'reserved',
    })
    const regularReservation = new Reservation({
        user,
        parkingLot,
        level: otherAvailableSpot.level,
        parkingSpot: otherAvailableSpot._id,
        reservationDate: new Date(reservationDate),
        isElectricChargingRequired,
        isDisabled: isDisabled ? isDisabled : false,
        parkingStatus: 'reserved',
        forMoving: true,
    })
    const savedReservation = await regularReservation.save()
    electricReservation.moveTo = savedReservation._id
    const save = await electricReservation.save()
    regularReservation.moveFrom = save._id
    await regularReservation.save()
    const userFromDb = await User.findById(user).select(['phoneNumber', 'firstName'])
    let sentSms: any = null
    const messageSentForBlocking = await checkIfSpotOuter(electricSpot, reservationDate)
    if (userFromDb && userFromDb?.phoneNumber) {
        const message = `A reservation has been made by ${userFromDb.firstName} for date ${moment(new Date(reservationDate)).format('YYYY-MM-DD')}. The parking lot has charging limit from ${lot.chargingTimeStart} to ${lot.chargingTimeEnd}. You will have to move your vehicle to the next assigned spot after the charge limit has been reached.`
        sentSms = await sendMessage(userFromDb.phoneNumber, message)
    }
    res.status(HttpStatus.CREATED).json({ reservation: save, sentSms, messageSentForBlocking })
}

const getUserAbsenceStatusForDate = async (req: Request, res: Response): Promise<{ designatedSpots: any }> => {
    return new Promise(async (resolve) => {
        const { _id }: { _id?: string } = req.params
        const { checkDate, checkTime }: { checkDate?: Date, checkTime?: string } = req.query
        const user = await User.findById(_id).populate({
            path: 'designatedParkingSpot',
            select: ['-level', '-spotType', '-designatedTo'],
            populate: {
                path: 'parkingLot',
                select: ['name', 'reservationTime', 'messageSendTime', 'absenceSendTime', 'designatedReservationDuration'],
            }
        }).select(['designatedParkingSpot'])
        const designatedParkingSpots = user?.designatedParkingSpot ? (Array.isArray(user?.designatedParkingSpot) ? user?.designatedParkingSpot : [user?.designatedParkingSpot]) : []
        let designatedSpots: any[] = []
        await Promise.all(designatedParkingSpots.map(async (spot: any) => {
            const dateToCheck = spot?.parkingLot?.reservationTime == 'day-before' ? moment(checkDate).add(1, 'days').toDate() : checkDate ? checkDate : new Date()
            const date = checkDate ? new Date(dateToCheck) : new Date()
            const { startOfDay, endOfDay } = getStartAndEndOfDay(date)
            const userPresenceDate = spot.userPresenceDate;
            const isUserPresent = userPresenceDate != null && userPresenceDate >= startOfDay && userPresenceDate <= endOfDay
            const absence = await Absence.find({ user: _id, designatedSpot: spot._id, startDate: { $lte: date }, endDate: { $gte: date }, status: 'pending' })
            if (absence.length == 0 && !isUserPresent && !spot.isPermanent) {
                const { yesterday } = getTodayTomorrowYesterday()
                const shabbatDateCheck = getCorrectDateTime(new Date())
                const isShabbat = await isJewishHolidayShabbat(shabbatDateCheck)
                const holiday = await getJewishHolidayEndDate(shabbatDateCheck)
                const holidayYesterday = await getJewishHolidayEndDate(yesterday)
                // console.log(isShabbat, 'isShabbat')
                if (!isShabbat?.isShabbat) {
                    const timerDuration = getTimeDifference(spot?.parkingLot.absenceSendTime, spot?.parkingLot.messageSendTime)
                    const checkTimeDiffWithMessageTime = getTimeDifference(checkTime ? checkTime : '05:00', spot?.parkingLot.messageSendTime)
                    const absenceTimeDiffWithCheckTime = getTimeDifference(spot?.parkingLot.absenceSendTime, checkTime ? checkTime : '05:00')
                    // Check if the time is between mid night and 3 hours after
                    const isTimeAfterMidNight = isTimeBetween(checkTime ? checkTime : '05:00', '00:00', '03:00')

                    let isParkingLotTimeBeforeShabbat = false
                    if (holiday) {
                        const timeAfterOneHourOfShabbat = addTime(holiday.endTime, '01:00')
                        isParkingLotTimeBeforeShabbat = getTimeDifference(spot?.parkingLot.messageSendTime, timeAfterOneHourOfShabbat).minutes < 0 ? true : false
                        console.log(timeAfterOneHourOfShabbat, isParkingLotTimeBeforeShabbat, getTimeDifference(spot?.parkingLot.messageSendTime, timeAfterOneHourOfShabbat).minutes)
                    }
                    if (checkTimeDiffWithMessageTime.minutes >= 0 && absenceTimeDiffWithCheckTime.minutes > 0 && !isParkingLotTimeBeforeShabbat)
                        designatedSpots.push({ ...spot.toObject(), status: false, timerDuration, timeLeft: absenceTimeDiffWithCheckTime })
                    else if (holiday) {
                        const timeAfterOneHourOfShabbat = addTime(holiday.endTime, '01:00')     // Acts as messageSendTime
                        const parkingLotReservationTime = getTimeDifference(spot?.parkingLot?.messageSendTime, timeAfterOneHourOfShabbat)
                        const isParkingLotTimePassed = parkingLotReservationTime.minutes < 0 ? true : false
                        if (spot?.parkingLot?.reservationTime == 'day-before' && isParkingLotTimePassed) {
                            const endTimeInShabbat = addTime(timeAfterOneHourOfShabbat, timerDuration.formatted)        // Acts as absenceSendTime
                            const checkTimeDiffWithMessageTime = getTimeDifference(checkTime ? checkTime : '05:00', timeAfterOneHourOfShabbat)
                            const absenceTimeDiffWithCheckTime = getTimeDifference(endTimeInShabbat, checkTime ? checkTime : '05:00')
                            if (checkTimeDiffWithMessageTime.minutes >= 0) {
                                // console.log(absenceTimeDiffWithCheckTime,endTimeInShabbat, 'absencetime')
                                // if (checkTimeDiffWithMessageTime.minutes >= 0 && absenceTimeDiffWithCheckTime.minutes > 0){
                                designatedSpots.push({
                                    ...spot.toObject(),
                                    parkingLot: {
                                        ...spot.toObject().parkingLot,
                                        messageSendTime: timeAfterOneHourOfShabbat
                                    }, status: false,
                                    timerDuration,
                                    timeLeft: absenceTimeDiffWithCheckTime
                                })
                            }
                        }
                    } else if (isTimeAfterMidNight && holidayYesterday) {
                        const timeAfterOneHourOfShabbat = addTime(holidayYesterday.endTime, '01:00')
                        const durationEndTime = addTime(timeAfterOneHourOfShabbat, spot?.parkingLot?.designatedReservationDuration)
                        const isDurationEndTimeAfterMidNight = isTimeBetween(durationEndTime, '00:00', '03:00')
                        if (isDurationEndTimeAfterMidNight) {
                            const absenceTimeDiffWithCheckTime = getTimeDifference(durationEndTime, checkTime ? checkTime : '05:00')
                            // console.log(isDurationEndTimeAfterMidNight, 'isDurationEndTimeAfterMidNight')
                            // console.log(timeAfterOneHourOfShabbat, 'timeAfterOneHourOfShabbat')
                            // console.log(durationEndTime, 'durationEndTime')
                            // console.log(absenceTimeDiffWithCheckTime, 'absenceTimeDiffWithCheckTime')
                            const date = getCorrectDateTime(new Date())
                            const { startOfDay, endOfDay } = getStartAndEndOfDay(date)
                            const userPresenceDate = spot.userPresenceDate;
                            const isUserPresent = userPresenceDate != null && userPresenceDate >= startOfDay && userPresenceDate <= endOfDay
                            const absence = await Absence.findOne({ user: _id, designatedSpot: spot._id, startDate: { $lte: date }, endDate: { $gte: date }, status: 'pending' })
                            // console.log(absence, 'absence', isUserPresent, 'isUserPresent')
                            if (absenceTimeDiffWithCheckTime.minutes > 0 && !absence && !isUserPresent) {
                                designatedSpots.push({
                                    ...spot.toObject(),
                                    parkingLot: {
                                        ...spot.toObject().parkingLot,
                                        messageSendTime: timeAfterOneHourOfShabbat
                                    },
                                    reservationDate: date,
                                    status: false,
                                    timerDuration,
                                    timeLeft: absenceTimeDiffWithCheckTime
                                })
                            }
                        }
                    }
                }
            }
        }))
        const sorted = designatedSpots.filter(spot => spot != null).sort((a, b) => {
            const timeA = a?.parkingLot?.messageSendTime;
            const timeB = b?.parkingLot?.messageSendTime;
            if (timeA < timeB) return -1;
            if (timeA > timeB) return 1;
            return 0;
        })
        resolve({ designatedSpots: sorted })
    })
}

const checkIfSpotHasReservations = async (spot: any, reservationDate: Date) => {
    try {
        const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
        const reservations = await Reservation.countDocuments({
            parkingSpot: spot?._id,
            reservationDate: { $gte: startOfDay, $lte: endOfDay },
            parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] }
        })
        if (reservations > 0) return true
        else return false
    } catch (error) {
        return true
    }

}

const checkIfSpotOuter = async (spot: IParkingSpotInterface, reservationDate: Date) => {
    if (spot.outerToSpot) {
        const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
        const outer: any = await Reservation.findOne({
            parkingSpot: spot._id,
            reservationDate: { $gte: startOfDay, $lte: endOfDay },
            parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] }
        }).populate({ path: 'user', select: ['firstName', 'lastName', 'phoneNumber', 'vehicle', 'vehicles', 'deviceToken'], populate: 'vehicles' })
            .populate({ path: 'parkingLot', select: ['name'] })
        const inner: any = await Reservation.findOne({
            parkingSpot: spot.outerToSpot,
            reservationDate: { $gte: startOfDay, $lte: endOfDay },
            parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] }
        }).populate({ path: 'user', select: ['firstName', 'lastName', 'phoneNumber', 'vehicle', 'vehicles', 'deviceToken'], populate: 'vehicles' })
            .populate({ path: 'parkingLot', select: ['name'] })
        if (!outer || !inner) return null
        else if (outer?.user?.phoneNumber && inner?.user?.phoneNumber) {
            const messageForInner = doubleParkingNotificationTextGenerator(
                outer?.user?.vehicles?.length > 0 ? outer?.user?.vehicles[0]?.registration : '',
                inner?.parkingLot?.name,
                `${outer?.user?.firstName} ${outer?.user?.lastName}`,
                outer?.user?.phoneNumber
            )
            const messageForOuter = doubleParkingNotificationTextGenerator(
                inner?.user?.vehicles?.length > 0 ? inner?.user?.vehicles[0]?.registration : '',
                outer?.parkingLot?.name,
                `${inner?.user?.firstName} ${inner?.user?.lastName}`,
                inner?.user?.phoneNumber
            )
            const sentToInner = inner?.user?.deviceToken ? sendNotification(inner?.user?.deviceToken, messageForInner, NotificationTypes.outerSpotReservation, { reservationId: inner._id.toString() }, [inner?.user?._id]) : null
            const sentToOuter = outer?.user?.deviceToken ? sendNotification(outer?.user?.deviceToken, messageForOuter, NotificationTypes.outerSpotReservation, { reservationId: outer._id.toString() }, [outer?.user?._id]) : null
            return { sentToInner, sentToOuter }
        }
        else return null
    } else return null
}

export const createReservationService = async (req: Request, res: Response) => {
    try {
        const { parkingLot, user, reservationDate, isElectricChargingRequired, location, language = "heb", parkingSpot, keepDesignatedSpot }: { parkingLot: string, user: string, reservationDate: Date, isElectricChargingRequired: boolean, location: { latitude: number, longitude: number }, language?: string, parkingSpot: string, keepDesignatedSpot: boolean } = req.body
        if (parkingLot && user && reservationDate && isElectricChargingRequired != undefined && location?.latitude && location?.longitude) {
            const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
            const parkingLotReservations = await Reservation.find({
                user,
                parkingLot,
                forMoving: false,
                parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] },
                $and: [{ reservationDate: { $gte: startOfDay } }, { reservationDate: { $lte: endOfDay } }]
            })
            const userReservations = await Reservation.find({
                user,
                forMoving: false,
                parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] },
                $and: [{ reservationDate: { $gte: startOfDay } }, { reservationDate: { $lte: endOfDay } }]
            })
            let absenceOnDesignated: null | typeof Absence = null
            // const designatedCheck = await ParkingSpot.findOne({
            //     parkingLot,
            //     designatedTo: user,
            //     // isPermanent: true
            // })
            const designatedSpot = await ParkingSpot.findOne({
                parkingLot,
                designatedTo: user,
                // isPermanent: false
            })
            if (designatedSpot) {
                absenceOnDesignated = await Absence.findOne({ designatedSpot, status: 'pending', startDate: { $lte: reservationDate }, endDate: { $gte: reservationDate } })
            }
            if (parkingLotReservations.length > 0) {
                res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_ALREADY_IN_PARKING_LOT_HEB : HttpMessages.reservation.RESERVATION_ALREADY_IN_PARKING_LOT })
            } else if (userReservations.length >= 2) {
                res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.USER_MAX_RESERVATION_HEB : HttpMessages.reservation.USER_MAX_RESERVATION });
            } else if (designatedSpot && (designatedSpot.isPermanent || (absenceOnDesignated == null && !keepDesignatedSpot))) {
                res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.DESIGNATED_ALREADY_IN_PARKING_LOT_HEB : HttpMessages.reservation.DESIGNATED_ALREADY_IN_PARKING_LOT })
            } else {
                const lot = await ParkingLot.findById(parkingLot)
                if (lot) {
                    const isInRadius = isCoordinatesInRadius(lot.location, location, lot.maxReservationDistance)
                    if (isInRadius || keepDesignatedSpot) {
                        if (lot.allotmentMethod == 'user-selected') {
                            const spot = await ParkingSpot.findById(parkingSpot)
                            if (spot) {
                                if (await checkIfSpotHasReservations(spot, reservationDate))
                                    res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED_HEB : HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED })
                                else {
                                    if (lot.chargingTimeLimit && spot.spotType == 'electric') {
                                        const { availableSpots: regularAvailable, } = await findAvailableSpots(req, 'regular', spot.level)
                                        if (regularAvailable.length > 0) {
                                            await createElectricReservationWithTimeLimit(spot, regularAvailable[0], lot, req, res)
                                        } else {
                                            const { availableSpots: designatedAvailable, } = await findAvailableSpots(req, 'designated', spot.level)
                                            if (designatedAvailable.length > 0) {
                                                await createElectricReservationWithTimeLimit(spot, designatedAvailable[0], lot, req, res)
                                            } else {
                                                const electricSpotsError: string = language == "heb" ? HttpMessages.reservation.RESERVATION_SPOTS_ALL_BOOKED_ELECTRIC_HEB : HttpMessages.reservation.RESERVATION_SPOTS_ALL_BOOKED_ELECTRIC
                                                const normalSpotsError: string = language == "heb" ? HttpMessages.reservation.RESERVATION_EXTRA_SPOT_NOT_FOUND_HEB : HttpMessages.reservation.RESERVATION_EXTRA_SPOT_NOT_FOUND
                                                res.status(HttpStatus.NOT_FOUND).json({ message: isElectricChargingRequired ? electricSpotsError : normalSpotsError })
                                            }
                                        }
                                    } else if (spot.spotType == 'designated') {
                                        const date = reservationDate ? new Date(reservationDate) : new Date()
                                        const absence = await Absence.find({ designatedSpot: spot._id, startDate: { $lte: date }, endDate: { $gte: date }, status: 'pending' })
                                        if (absence.length > 0 || spot.designatedTo == null || String(spot.designatedTo) == user) {
                                            const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
                                            const reservations = await Reservation.find({ user: { $ne: user }, parkingSpot: spot._id, reservationDate: { $gte: startOfDay, $lte: endOfDay }, parkingStatus: { $in: ['reserved', 'occupied'] } })
                                            if (reservations.length > 0) res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED_HEB : HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED })
                                            else {
                                                const newReservation = new Reservation({
                                                    user,
                                                    parkingLot,
                                                    level: spot.level,
                                                    parkingSpot: spot._id,
                                                    reservationDate: new Date(reservationDate),
                                                    isElectricChargingRequired,
                                                    parkingStatus: 'reserved',
                                                })
                                                const savedReservation = await newReservation.save()
                                                const messageSentForBlocking = await checkIfSpotOuter(spot, reservationDate)
                                                if (keepDesignatedSpot) {
                                                    const parkingSpot = await ParkingSpot.findByIdAndUpdate(spot._id, { userPresenceDate: reservationDate }, { omitUndefined: true, new: true });
                                                    res.status(HttpStatus.CREATED).json({ reservation: savedReservation, designatedSpot: parkingSpot, messageSentForBlocking })
                                                } else res.status(HttpStatus.CREATED).json({ reservation: savedReservation, designatedSpot: spot, messageSentForBlocking })

                                            }
                                        } else res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_SPOT_NOT_RELEASED_HEB : HttpMessages.reservation.RESERVATION_SPOT_NOT_RELEASED })
                                    } else {
                                        const newReservation = new Reservation({
                                            user,
                                            parkingLot,
                                            level: spot.level,
                                            parkingSpot: spot._id,
                                            reservationDate: new Date(reservationDate),
                                            isElectricChargingRequired,
                                            parkingStatus: 'reserved',
                                        })
                                        const savedReservation = await newReservation.save()
                                        const messageSentForBlocking = await checkIfSpotOuter(spot, reservationDate)
                                        res.status(HttpStatus.CREATED).json({ reservation: savedReservation, messageSentForBlocking })
                                    }
                                }
                            } else res.status(HttpStatus.NOT_FOUND).json({ message: language == 'heb' ? HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND_HEB : HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
                        }
                    } else {
                        const radiusError: string = language == "heb" ? HttpMessages.reservation.RESERVATION_NOT_IN_RESERVATION_RADIUS_HEB : HttpMessages.reservation.RESERVATION_NOT_IN_RESERVATION_RADIUS
                        res.status(HttpStatus.FORBIDDEN).json({ message: radiusError })
                    }
                } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
            }
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_DETAILS_MISSING })
    } catch (err) {
        console.log(err, 'err')
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const getAllReservationsService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10, parkingStatus }: { page?: number, limit?: number, parkingStatus?: string } = req.query
        let filter: any = {}
        if (parkingStatus) filter['parkingStatus'] = parkingStatus
        const totalCount = await Reservation.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        const reservations = await Reservation.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
        res.status(HttpStatus.OK).json({ totalPages, totalCount, reservations })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const getUserReservationsService = async (req: Request, res: Response) => {
    try {
        const { _id: user }: { _id?: string } = req.params
        const { page = 1, limit = 10, parkingStatus }: { page?: number, limit?: number, parkingStatus?: string } = req.query
        if (user) {
            const { startOfDay } = getStartAndEndOfDay(new Date())
            let filter: any = { user, forMoving: false, parkingStatus: { $nin: [reservationEnums[3], reservationEnums[4],] }, reservationDate: { $gte: startOfDay } }
            if (parkingStatus) filter['parkingStatus'] = parkingStatus
            const totalCount = await Reservation.countDocuments(filter);
            const totalPages = Math.ceil(totalCount / limit);
            const reservations = await Reservation.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
                .populate({ path: 'parkingLot', select: ['name',] })
                .populate({ path: 'level', select: ['entrance'] })
                .populate({ path: 'parkingSpot', select: ['parkingCode'] })
            const { designatedSpots } = await getUserAbsenceStatusForDate(req, res)
            res.status(HttpStatus.OK).json({ totalPages, totalCount, reservations, designatedSpots })
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const getParkingLotReservationsService = async (req: Request, res: Response) => {
    try {
        const { _id: parkingLot }: { _id?: string } = req.params
        const { page = 1, limit = 10, parkingStatus }: { page?: number, limit?: number, parkingStatus?: string } = req.query
        if (parkingLot) {
            let filter: any = { parkingLot }
            if (parkingStatus) filter['parkingStatus'] = parkingStatus
            const totalCount = await Reservation.countDocuments(filter);
            const totalPages = Math.ceil(totalCount / limit);
            const reservations = await Reservation.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
            res.status(HttpStatus.OK).json({ totalPages, totalCount, reservations })
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const getLevelReservationsService = async (req: Request, res: Response) => {
    try {
        const { _id: level }: { _id?: string } = req.params
        const { page = 1, limit = 10, parkingStatus }: { page?: number, limit?: number, parkingStatus?: string } = req.query
        if (level) {
            let filter: any = { level }
            if (parkingStatus) filter['parkingStatus'] = parkingStatus
            const totalCount = await Reservation.countDocuments(filter);
            const totalPages = Math.ceil(totalCount / limit);
            const reservations = await Reservation.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
            res.status(HttpStatus.OK).json({ totalPages, totalCount, reservations })
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const getSpotAvailabilityService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { reservationDate, language = 'heb' }: { reservationDate?: Date, language?: string } = req.query
        const parkingSpot = await ParkingSpot.findById(_id)
        if (parkingSpot) {
            const date = reservationDate ? new Date(reservationDate) : new Date()
            const { startOfDay, endOfDay } = getStartAndEndOfDay(date)
            const reservations = await Reservation.find({ parkingSpot, reservationDate: { $gte: new Date(startOfDay), $lte: new Date(endOfDay) }, parkingStatus: { $in: ['reserved', 'occupied'] } })
            if (reservations?.length > 0) res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED_HEB : HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED })
            else res.status(HttpStatus.OK).json({ isAvailable: true })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const getSingleReservationService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const reservation: any = await Reservation.findById(_id)
            .populate('parkingLot', ['name', 'accessMapFile', 'avatar', 'chargingTimeLimit', 'chargingTimeStart', 'chargingTimeEnd', 'reservationTime', 'messageSendTime', 'messageSendTimeOfDay',])
            .populate('level', ['level', 'map', 'entrance',])
            .populate('parkingSpot', ['parkingCode', 'name', 'avatar', 'spotType', 'innerToSpot', 'outerToSpot',])
            .populate({ path: 'moveTo', populate: { path: 'parkingSpot', select: ['parkingCode', 'name', 'avatar',] }, select: ['parkingSpot', 'reservationDate', 'parkingStatus'] })
            .populate({ path: 'moveFrom', populate: { path: 'parkingSpot', select: ['parkingCode', 'name', 'avatar',] }, select: ['parkingSpot', 'reservationDate', 'parkingStatus'] })
        if (reservation) {
            const outerReservation = async () => {
                if (reservation.parkingSpot.innerToSpot) {
                    const outerSpot = await ParkingSpot.findById(reservation.parkingSpot.innerToSpot)
                    if (outerSpot) {
                        const { startOfDay, endOfDay } = getStartAndEndOfDay(reservation.reservationDate)
                        const outerReservation = await Reservation.findOne({
                            parkingLot: reservation.parkingLot,
                            level: reservation.level,
                            parkingSpot: outerSpot._id,
                            reservationDate: { $gte: startOfDay, $lte: endOfDay },
                            parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] }
                        }).populate({ path: 'user', select: ['phoneNumber', 'firstName', 'lastName', 'vehicle', 'vehicles'], populate: 'vehicles' })
                        if (outerReservation) return outerReservation
                        else return null
                    } else return null
                } else return null
            }
            const innerReservation = async () => {
                if (reservation.parkingSpot.outerToSpot) {
                    const innerSpot = await ParkingSpot.findById(reservation.parkingSpot.outerToSpot)
                    if (innerSpot) {
                        const { startOfDay, endOfDay } = getStartAndEndOfDay(reservation.reservationDate)
                        const innerReservation = await Reservation.findOne({
                            parkingLot: reservation.parkingLot,
                            level: reservation.level,
                            parkingSpot: innerSpot._id,
                            reservationDate: { $gte: startOfDay, $lte: endOfDay },
                            parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] }
                        }).populate({ path: 'user', select: ['phoneNumber', 'firstName', 'lastName', 'vehicle', 'vehicles'], populate: 'vehicles' })
                        if (innerReservation) return innerReservation
                        else return null
                    } else return null
                } else return null
            }
            res.status(HttpStatus.OK).json({ reservation, outerReservation: await outerReservation(), innerReservation: await innerReservation() })
        }
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.reservation.RESERVATION_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const updateReservationService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { parkingStatus, user, reservationDate }: { parkingStatus: string, user?: ObjectId, reservationDate: Date } = req.body
        if (reservationEnums.includes(parkingStatus)) {
            const reservation: any = await Reservation.findById(_id).populate({ path: 'parkingSpot', select: ['designatedTo', 'spotType'] })
            if (reservation) {
                if (parkingStatus == reservationEnums[2]) {
                    if (reservation.moveTo) {
                        reservation.parkingStatus = reservationEnums[2]
                        await Reservation.findByIdAndUpdate(reservation.moveTo, { parkingStatus: 'occupied' })
                        await reservation.save()
                    } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_INVALID_STATUS })
                }
                if (reservationEnums.includes(parkingStatus)) {
                    reservation.parkingStatus = parkingStatus
                    await reservation.save()
                    if (reservation?.moveTo && (parkingStatus == reservationEnums[3] || parkingStatus == reservationEnums[4])) await Reservation.findByIdAndUpdate(reservation?.moveTo, { parkingStatus })
                    if (parkingStatus == reservationEnums[3] && user && reservation?.parkingSpot?.designatedTo?.toString() == user) {
                        const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate)
                        await ParkingSpot.findByIdAndUpdate(reservation?.parkingSpot, { userPresenceDate: null }, { new: true })
                        const absence = new Absence({
                            user: user,
                            designatedSpot: reservation?.parkingSpot._id,
                            startDate: startOfDay,
                            endDate: endOfDay,
                            type: 'system-generated',
                            note: 'הזמנת מקום חניה הושלמה על ידי המשתמש.',
                        })
                        await absence.save()
                    }
                    res.status(HttpStatus.OK).json(reservation)
                } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_INVALID_STATUS })
            } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const changeReservedSpotService = async (req: Request, res: Response) => {
    try {
        const { _id } = req.params
        const { language = "heb", parkingSpot }: { language?: string, parkingSpot: Schema.Types.ObjectId } = req.body
        const reservation = await Reservation.findById(_id)
        if (reservation) {
            const spot = await ParkingSpot.findById(parkingSpot)
            if (spot) {
                if (await checkIfSpotHasReservations(spot, reservation.reservationDate)) res.status(HttpStatus.FORBIDDEN).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED_HEB : HttpMessages.reservation.RESERVATION_SELECTED_SPOT_ALREADY_BOOKED })
                else {
                    const saved = await Reservation.findByIdAndUpdate(_id, { parkingSpot })
                    res.status(HttpStatus.OK).json(saved)
                }
            } else res.status(HttpStatus.NOT_FOUND).json({ message: language == 'heb' ? HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND_HEB : HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: language == 'heb' ? HttpMessages.reservation.RESERVATION_NOT_FOUND_HEB : HttpMessages.reservation.RESERVATION_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}
export const deleteReservationService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const reservation = await Reservation.findByIdAndDelete(_id)
        if (reservation) res.status(HttpStatus.OK).json({ message: HttpMessages.reservation.RESERVATION_DELETED_SUCCESS })
        else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.reservation.RESERVATION_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}