import { Request, Response, } from 'express';
import HttpStatus from '../../services/constants/http-statuses';
import HttpMessages from '../../services/constants/http-messages';
import Announcement from '../../models/announcement';
import Admin from '../../models/admin';
import User from '../../models/user';
interface IUpdateAnnouncementInterface {
    parkingLots?: string,
    startDate?: Date,
    endDate?: Date,
    title?: string,
    content?: string,
    status?: string
}

export const createAnnouncementService = async (req: Request, res: Response) => {
    try {
        const { parkingLots, startDate, endDate, title, content, }: { parkingLots?: string[], startDate?: Date, endDate?: Date, title?: string, content?: string, status?: string } = req.body;
        if (parkingLots && startDate && endDate && title && content) {
            if (new Date(startDate) > new Date(endDate)) res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.announcement.ANNOUNCEMENT_WRONG_DATE })
            else {
                const announcement = new Announcement({
                    parkingLots,
                    startDate,
                    endDate,
                    title,
                    content,
                    status: 'active',
                })
                await announcement.save()
                res.status(HttpStatus.CREATED).json({ announcement })
            }
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.announcement.ANNOUNCEMENT_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAdminsAnnouncementsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { page = 1, limit = 10, status, currentDate, }: { page?: number, limit?: number, status?: string, currentDate?: Date, } = req.query
        if (_id) {
            const admin = await Admin.findById(_id).select(['parkingLots'])
            if (admin) {
                const { parkingLots } = admin
                let filter: any = { parkingLots: { $in: parkingLots }, }
                if (status) filter.status = status
                if (currentDate) {
                    if (currentDate) {
                        filter = {
                            ...filter,
                            $or: [{ startDate: { $gte: currentDate } }, { endDate: { $gte: currentDate } }]
                        }
                    }
                }
                const totalCount = await Announcement.countDocuments(filter);
                const totalPages = Math.ceil(totalCount / limit);
                const announcements = await Announcement.find(filter).sort({ updatedAt: -1, }).limit(limit).skip(limit * (page - 1)).populate({ path: 'parkingLots', select: ['name', 'avatar'] })
                res.status(HttpStatus.OK).json({ totalPages, totalCount, announcements });
            } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAllAnnouncementsService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10, status, currentDate, parkingLots }: { page?: number, limit?: number, status?: string, currentDate?: Date, parkingLots?: string[] } = req.query
        let filter: any = {}
        if (status) filter.status = status
        if (parkingLots && parkingLots.length > 0) filter.parkingLots = { $in: Array.isArray(parkingLots) ? parkingLots : [parkingLots] }
        if (currentDate) {
            if (currentDate) {
                filter = {
                    ...filter,
                    $or: [{ startDate: { $gte: currentDate } }, { endDate: { $gte: currentDate } }]
                }
            }
        }
        const totalCount = await Announcement.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        const announcements = await Announcement.find(filter).sort({ updatedAt: -1, }).limit(limit).skip(limit * (page - 1)).populate({ path: 'parkingLots', select: ['name', 'avatar'] })
        res.status(HttpStatus.OK).json({ totalPages, totalCount, announcements });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getUserAnnouncementsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { page = 1, limit = 10, status, currentDate, }: { page?: number, limit?: number, status?: string, currentDate?: Date, user?: string } = req.query
        let filter: any = {}
        if (status) filter.status = status
        if (currentDate) {
            filter = {
                ...filter,
                $or: [{ startDate: { $gte: currentDate } }, { endDate: { $gte: currentDate } }]
            }
        }
        const user = await User.findById(_id)
        if (user) {
            const parkingLots = user.parkingLots
            filter.parkingLots = { $in: parkingLots }
            const totalCount = await Announcement.countDocuments(filter);
            const totalPages = Math.ceil(totalCount / limit);
            const unreadCount = await Announcement.countDocuments({ ...filter, $or: [{ readBy: { $nin: [_id] } }, { readBy: undefined }] })
            const announcements = await Announcement.find(filter).sort({ updatedAt: -1, }).limit(limit).skip(limit * (page - 1)).populate({ path: 'parkingLots', select: ['name', 'avatar'] })
            res.status(HttpStatus.OK).json({ totalPages, totalCount, announcements, unreadCount });
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSingleAnnouncementService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const announcement = await Announcement.findById(_id);
        if (announcement) res.status(HttpStatus.OK).json(announcement)
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.announcement.ANNOUNCEMENT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const readUserAnnouncementsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { page = 1, limit = 10, }: { page?: number, limit?: number, user?: string } = req.query
        const { status, currentDate, }: { status?: string, currentDate?: Date, } = req.body
        let filter: any = {}
        if (status) filter.status = status
        const user = await User.findById(_id)
        if (user) {
            const parkingLots = user.parkingLots
            filter.parkingLots = { $in: parkingLots }
            const totalCount = await Announcement.countDocuments(filter);
            const totalPages = Math.ceil(totalCount / limit);
            const announcements = await Announcement.updateMany(
                { ...filter, $or: [{ readBy: { $nin: [_id] } }, { readBy: undefined }] },
                { $push: { readBy: _id } })
            res.status(HttpStatus.OK).json({ totalPages, totalCount, announcements, });
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const updateAnnouncementService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { parkingLots, startDate, endDate, title, content, status }: IUpdateAnnouncementInterface = req.body;
        const updateData: IUpdateAnnouncementInterface = { parkingLots, startDate, endDate, title, content, status, }
        const announcement = await Announcement.findByIdAndUpdate(_id, updateData, { new: true, omitUndefined: true, });
        if (announcement) res.status(HttpStatus.OK).json(announcement)
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.announcement.ANNOUNCEMENT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const deleteAnnouncementService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const announcement = await Announcement.findByIdAndDelete(_id);
        if (announcement) res.status(HttpStatus.OK).json({ message: HttpMessages.announcement.ANNOUNCEMENT_DELETED_SUCCESS });
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.announcement.ANNOUNCEMENT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}