import { Request, Response, } from 'express';
import HttpMessages from '../../services/constants/http-messages';
import HttpStatus from '../../services/constants/http-statuses';
import ParkingLot from '../../models/parking-lot';
import Level from '../../models/lot-level';
import ParkingSpot, { IParkingSpotInterface } from '../../models/parking-spot';
import Absence, { IAbsenceInterface } from '../../models/absence';
import { getStartAndEndOfDay } from '../../helpers/dates';
import Reservation from '../../models/reservation';
import { reservationEnums } from '../reservations/controllers';
import { customSort, findAvailableSpots } from '../../helpers';

export const createParkingSpotService = async (req: Request, res: Response) => {
    try {
        const { parkingLot, name, avatar, level, parkingCode, spotType }: { parkingLot: string, name: string, avatar: string, level: string, parkingCode: string, spotType: string } = req.body;
        if (parkingLot && level && parkingCode) {
            const lot = await ParkingLot.findById(parkingLot)
            if (lot) {
                const parkingLevel = await Level.findById(level)
                if (parkingLevel) {
                    const code = await ParkingSpot.findOne({ parkingLot, level, parkingCode })
                    if (code) {
                        res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.parkingSpot.PARKING_CODE_ALREADY_IN_DB })
                    } else {
                        const parkingSpot = new ParkingSpot({
                            name,
                            avatar,
                            parkingLot,
                            level,
                            spotType,
                            parkingCode,
                        })
                        await parkingSpot.save();
                        res.status(HttpStatus.CREATED).json(parkingSpot);
                    }
                } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.lotLevel.LEVEL_NOT_FOUND })
            } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAllParkingSpotsService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10, type, isAvailable }: { page?: number, limit?: number, type?: string, isAvailable?: boolean | string } = req.query
        let filter: any = {}
        if (type != undefined) filter.spotType = type
        if (type == 'designated' && (isAvailable == 'true' || isAvailable == true)) filter.designatedTo = null
        if (type == 'designated' && (isAvailable == 'false' || isAvailable == false)) filter.designatedTo = { $ne: null }
        const totalCount = await ParkingSpot.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        const parkingSpots = await ParkingSpot.find(filter).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
        res.status(HttpStatus.OK).json({ totalPages, totalCount, parkingSpots });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSingleParkingSpotService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const parkingSpot = await ParkingSpot.findById(_id);
        if (parkingSpot) {
            res.status(HttpStatus.OK).json(parkingSpot);
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getParkingLotParkingSpotsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { page = 1, limit = 10, type, isAvailable, sortParkingCodes }: { page?: number, limit?: number, type?: string, isAvailable?: boolean | string, sortParkingCodes?: boolean | string } = req.query
        let filter: any = { parkingLot: _id }
        if (type != undefined) filter.spotType = type
        if (type == 'designated' && (isAvailable == 'true' || isAvailable == true)) filter.designatedTo = null
        if (type == 'designated' && (isAvailable == 'false' || isAvailable == false)) filter.designatedTo = { $ne: null }
        const parkingLot = await ParkingLot.findById(_id)
        if (parkingLot) {
            const sort: {} = sortParkingCodes == 'true' || sortParkingCodes == true ? { parkingCode: 1 } : { _id: 1 }
            // const sort: {} = parkingLot.allotmentMethod == 'manual' ? { createdAt: 1 } : { parkingCode: 1 }
            const totalCount = await ParkingSpot.countDocuments(filter);
            const totalPages = Math.ceil(totalCount / limit);
            const parkingSpots = await ParkingSpot.find(filter).sort(sort).limit(limit).skip(limit * (page - 1))
            res.status(HttpStatus.OK).json({ totalPages, totalCount, parkingSpots });
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getLotMonitorSpotsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { startDate = new Date(), endDate = new Date(), }: { startDate?: Date, endDate?: Date, } = req.query
        let filter: any = { parkingLot: _id }
        const parkingLot = await ParkingLot.findById(_id)
        if (parkingLot) {
            const sort: {} = { parkingCode: 1 }
            const { startOfDay, } = getStartAndEndOfDay(startDate)
            const { endOfDay } = getStartAndEndOfDay(endDate)
            const levels = await Level.find({ parkingLot: _id })
            const parkingSpots: any = await ParkingSpot.find(filter).sort(sort).populate({
                path: 'designatedTo',
                select: ['-password', '-temporaryPassword'],
                populate: [{
                    path: 'parkingLots',
                }, {
                    path: 'designatedParkingSpot',
                    populate: 'parkingLot'
                }, {
                    path: 'vehicles'
                }]
            })
            const reservations: any = await Reservation.find({ parkingStatus: { $nin: [reservationEnums[3], reservationEnums[4]] }, reservationDate: { $gte: startOfDay, $lte: endOfDay } })
                .populate({
                    path: 'user',
                    select: ['-password', '-temporaryPassword'],
                    populate: [{
                        path: 'parkingLots',
                    }, {
                        path: 'designatedParkingSpot',
                        populate: 'parkingLot'
                    }, {
                        path: 'vehicles'
                    }]
                })
            const finalSpots = await Promise.all(parkingSpots.map(async (spot: any) => {
                const reservationIndex = await Promise.resolve(reservations.findIndex((reservation: any) => String(reservation.parkingSpot) == String(spot._id)))
                const absence = await Absence.findOne({ designatedSpot: spot?._id, startDate: { $lte: startDate }, endDate: { $gte: endDate } })
                if (reservationIndex < 0) {
                    if (spot?.designatedTo?.parkingLots && spot?.designatedTo?.parkingLots?.length > 0) {
                        const userWithDesignatedSpots = await Promise.all(spot?.designatedTo?.parkingLots.map(async (parkingLot: any) => {
                            const designatedSpots = await ParkingSpot.find({ parkingLot: parkingLot._id, spotType: 'designated', designatedTo: spot?.designatedTo?._id });
                            return {
                                ...parkingLot.toObject(),
                                designatedSpots
                            };
                        }))
                        const designatedTo = spot?.designatedTo ? { ...spot?.designatedTo?.toObject(), parkingLots: userWithDesignatedSpots } : null
                        return {
                            ...spot.toObject(),
                            status: absence && designatedTo ? 'released' : 'designated-not-released',
                            designatedTo,
                            reservedBy: null,
                        }
                    } else return {
                        ...spot.toObject(),
                        status: 'free',
                        reservedBy: null,
                    }
                }
                else {
                    const userWithDesignatedSpots = await Promise.all(reservations[reservationIndex].user.parkingLots.map(async (parkingLot: any) => {
                        const designatedSpots = await ParkingSpot.find({ parkingLot: parkingLot._id, spotType: 'designated', designatedTo: null });
                        return {
                            ...parkingLot.toObject(),
                            designatedSpots
                        };
                    }))
                    return {
                        ...spot.toObject(),
                        status: 'reserved',
                        reservedBy: {
                            ...reservations[reservationIndex].user.toObject(),
                            parkingLots: userWithDesignatedSpots
                        },
                    }
                }
            }))
            let totalCount = 0
            let totalReservedCount = 0
            let totalFreeCount = 0
            let totalElectricCount = 0
            let totalDesignatedCount = 0
            let totalRegularCount = 0
            let totalDisabledCount = 0
            const levelsWithSpots = await Promise.all(levels.map(async (level) => {
                const spots = await Promise.all(finalSpots.filter((spot) => String(spot.level) == String(level._id)))
                const reserved = await Promise.all(spots.filter((spot) => spot.status == 'reserved'))
                const free = await Promise.all(spots.filter((spot) => spot.status == 'free'))
                const electric = await Promise.all(spots.filter((spot) => spot.spotType == 'electric'))
                const designated = await Promise.all(spots.filter((spot) => spot.spotType == 'designated'))
                const designatedReleased = await Promise.all(spots.filter((spot) => spot.spotType == 'designated' && spot?.status == 'released'))
                const regular = await Promise.all(spots.filter((spot) => spot.spotType == 'regular'))
                const disabled = await Promise.all(spots.filter((spot) => spot.spotType == 'disabled'))
                totalCount = totalCount + spots.length
                totalReservedCount = totalReservedCount + reserved.length
                totalFreeCount = totalFreeCount + free.length + designatedReleased?.length
                totalElectricCount = totalElectricCount + electric.length
                totalDesignatedCount = totalDesignatedCount + designated.length
                totalRegularCount = totalRegularCount + regular.length
                totalDisabledCount = totalDisabledCount + disabled.length
                return {
                    level: {
                        ...level.toObject(),
                        totalCount: spots.length,
                        reservedCount: reserved.length,
                        electricCount: electric.length,
                        designatedCount: designated.length,
                        regularCount: regular.length,
                        disabledCount: disabled.length,
                        freeCount: free.length + designatedReleased?.length,
                        // freeCount: free.length,
                    },
                    parkingSpots: spots.sort(customSort)
                }
            }))
            res.status(HttpStatus.OK).json({
                totalCount,
                totalReservedCount,
                totalFreeCount,
                totalElectricCount,
                totalDesignatedCount,
                totalRegularCount,
                totalDisabledCount,
                levels: levelsWithSpots,
            });
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getLotLevelSpotsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { page = 1, limit = 10, type, isAvailable, isDisabled, isElectric }: { page?: number, limit?: number, type?: string, isAvailable?: boolean | string, isDisabled?: boolean | string, isElectric?: boolean | string } = req.query
        let filter: any = { level: _id, }
        if (type != undefined) filter.spotType = type
        if (type == 'designated' && isAvailable == 'true') filter.designatedTo = null
        if (type == 'designated' && isAvailable == 'false') filter.designatedTo = { $ne: null }
        if (isDisabled == 'false') filter.spotType = { ...filter.spotType, $nin: ['disabled'] };
        if (isElectric == 'false') filter.spotType = { ...filter.spotType, $nin: ['electric'] };
        if (isDisabled == 'false' && isElectric == 'false') filter.spotType = { $nin: ['disabled', 'electric'] };
        const level = await Level.findById(_id)
        if (level) {
            const parkingLot = await ParkingLot.findById(level.parkingLot)
            if (parkingLot) {
                const sort: {} = { _id: 1 }
                const totalCount = await ParkingSpot.countDocuments(filter);
                const totalPages = Math.ceil(totalCount / limit);
                const parkingSpots = await ParkingSpot.find(filter).sort(sort).limit(limit).skip(limit * (page - 1))
                if (parkingSpots) {
                    res.status(HttpStatus.OK).json({ totalPages, totalCount, parkingSpots });
                } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.lotLevel.LEVEL_NOT_FOUND })
            } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.lotLevel.LEVEL_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getLevelAvailableSpotsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const level = await Level.findById(_id)
        if (level) {
            const parkingLot = await ParkingLot.findById(level.parkingLot)
            if (parkingLot) {
                const { availableSpots, totalCount } = await findAvailableSpots(req)
                res.status(HttpStatus.OK).json({ parkingSpots: availableSpots, totalCount })
            } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.lotLevel.LEVEL_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const updateParkingSpotService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { name, avatar, userPresenceDate }: { name?: string, avatar?: string, userPresenceDate: Date | string, } = req.body;
        const parkingSpot = await ParkingSpot.findByIdAndUpdate(_id, { name, avatar, userPresenceDate }, { omitUndefined: true, new: true });
        if (parkingSpot) {
            let response: { parkingSpot?: IParkingSpotInterface, absence?: IAbsenceInterface } = { parkingSpot }
            if (userPresenceDate == null) {
                const absence = new Absence({
                    user: parkingSpot.designatedTo,
                    designatedSpot: parkingSpot._id,
                    startDate: new Date().setUTCHours(0, 0, 0, 0),
                    endDate: new Date().setUTCHours(23, 59, 59, 999),
                    type: 'system-generated',
                })
                const saved = await absence.save();
                if (saved) response.absence = saved
            }
            res.status(HttpStatus.OK).json(response)
        }
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const deleteParkingSpotService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const parkingSpot = await ParkingSpot.findByIdAndDelete(_id);
        if (parkingSpot) {
            res.status(HttpStatus.OK).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_DELETED_SUCCESS })
        }
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}