import { Request, Response, } from 'express';
import Absence, { IAbsenceInterface } from "../../models/absence";
import HttpMessages from '../../services/constants/http-messages';
import HttpStatus from '../../services/constants/http-statuses';
import { getCorrectDateTime, getTimeRelation } from '../../helpers/dates';
import ParkingSpot from '../../models/parking-spot';
import { reservationEnums } from '../reservations/controllers';
import Reservation from '../../models/reservation';
import moment from 'moment';
const absenceEnums = ['pending', 'completed', 'canceled']

export const createAbsenceService = async (req: Request, res: Response) => {
    try {
        const { user, designatedSpot, startDate, endDate, note, type, }: IAbsenceInterface = req.body;
        if (user && designatedSpot && startDate && endDate) {
            const absence = new Absence({
                user,
                designatedSpot,
                startDate: new Date(startDate).setUTCHours(0, 0, 0, 1),
                endDate: new Date(endDate).setUTCHours(23, 59, 59, 999),
                note,
                type: type ? type : 'user-created',
            })
            const saved = await absence.save()
            res.status(HttpStatus.CREATED).json(saved);
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.absence.ABSENCE_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAllAbsencesService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10, status, user, designatedSpot }: { page?: number, limit?: number, status?: string, user?: string, designatedSpot?: string } = req.query
        let filter: any = {
            endDate: { $gte: getCorrectDateTime(new Date()) },
        }
        if (status) filter.status = status
        if (user) {
            filter.user = user
            filter.type = 'user-created'
        }
        if (designatedSpot) filter.designatedSpot = designatedSpot
        const totalCount = await Absence.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        const absences = await Absence.find(filter).sort({ startDate: -1, }).limit(limit).skip(limit * (page - 1)).populate({ path: 'designatedSpot', select: ['parkingLot', 'parkingCode'], populate: { path: 'parkingLot', select: ['name'] } })
        res.status(HttpStatus.OK).json({ totalPages, totalCount, absences });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSingleAbsenceService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const absence = await Absence.findById(_id);
        if (absence) res.status(HttpStatus.OK).json(absence);
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.absence.ABSENCE_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSpotAbsencesService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const absences = await Absence.find({ designatedSpot: _id });
        if (absences) res.status(HttpStatus.OK).json({ absences });
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.absence.ABSENCE_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getUserAbsenceStatusForDateService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { checkDate, checkTime }: { checkDate?: Date, checkTime?: string } = req.query
        const date = checkDate ? new Date(checkDate) : new Date()
        if (!_id || _id == undefined || _id == 'undefined') res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.parkingSpot.SPOT_ID_NOT_ENTERED })
        else {
            const spot: any = await ParkingSpot.findById(_id).populate({ path: 'parkingLot', select: 'messageSendTime' })
            if (spot) {
                const startOfDay = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0);
                const endOfDay = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 23, 59, 59);
                const userPresenceDate = spot.userPresenceDate;
                if (userPresenceDate >= startOfDay && userPresenceDate <= endOfDay) res.status(HttpStatus.OK).json({ status: true, message: 'Already submitted' })
                else {
                    const absence = await Absence.findOne({ designatedSpot: _id, startDate: { $lte: date }, endDate: { $gte: date }, status: 'pending' })
                    if (absence) res.status(HttpStatus.OK).json({ status: true, message: 'Already have an absence' })
                    else {
                        const timeDiff = getTimeRelation(checkTime ? checkTime : '05:00', spot?.parkingLot.messageSendTime)
                        if (timeDiff == 'within-2-hours') res.status(HttpStatus.OK).json({ status: false, message: 'Time is in between two hours' })
                        else if (timeDiff == 'before') res.status(HttpStatus.OK).json({ status: true, message: 'It is too soon' })
                        else if (timeDiff == 'after' || timeDiff == 'within-4-hours') res.status(HttpStatus.OK).json({ status: true, message: 'Time has passed' })
                    }
                }
            } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const updateAbsenceService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { status }: { status: string } = req.body;
        if (absenceEnums.includes(status)) {
            if (status == 'canceled') {
                const absence = await Absence.findById(_id)
                if (absence) {
                    const query = {
                        $and: [
                            {
                                parkingSpot: absence.designatedSpot
                            },
                            {
                                $or: [
                                    { parkingStatus: { $ne: reservationEnums[0] } },
                                    { parkingStatus: { $ne: reservationEnums[1] } }
                                ]
                            },
                            {
                                reservationDate: { $gte: absence.startDate, $lte: absence.endDate }
                            }
                        ]
                    }
                    const reservations = await Reservation.find(query)
                    if (reservations.length > 0) {
                        res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.absence.CANNOT_CANCEL_ABSENCE + ` Reservation Date(s): ${reservations.map(reservation => moment(reservation.reservationDate).format('DD-MM-YYYY')).join(', ')}.` })
                    } else {
                        const absence = await Absence.findByIdAndUpdate(_id, { status, }, { new: true, omitUndefined: true, });
                        res.status(HttpStatus.OK).json(absence)
                    }
                } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.absence.ABSENCE_NOT_FOUND })
            } else {
                const absence = await Absence.findByIdAndUpdate(_id, { status, }, { new: true, omitUndefined: true, });
                if (absence) res.status(HttpStatus.OK).json({ message: HttpMessages.absence.ABSENCE_UPDATE_SUCCESS })
                else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.absence.ABSENCE_NOT_FOUND })
            }

        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.absence.ABSENCE_INVALID_STATUS })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const deleteAbsenceService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const absence = await Absence.findById(_id);
        if (absence) {
            // const query = {
            //     $and: [
            //         { parkingSpot: absence.designatedSpot },
            //         { parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] } },
            //         { reservationDate: { $gte: absence.startDate, $lte: absence.endDate } }
            //     ]
            // }
            const reservations = await Reservation.find({ parkingSpot: absence.designatedSpot, parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] }, reservationDate: { $gte: absence.startDate, $lte: absence.endDate } })
            if (reservations?.length > 0) {
                res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.absence.CANNOT_DELETE_ABSENCE })
            } else {
                await Absence.findByIdAndDelete(_id)
                res.status(HttpStatus.OK).json({ message: HttpMessages.absence.ABSENCE_DELETED_SUCCESS })
            }
        }
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.absence.ABSENCE_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}