import { Request, Response, } from 'express';
import { generateHashPassword, generateTemporaryPassword } from "../../helpers/generators"
import { isDate90DaysOld, isPasswordValid, validateEmail, validatePassword } from "../../helpers/validators"
import { sendEmail } from "../../middlewares/email-messaging"
import Admin from "../../models/admin"
import HttpMessages from "../../services/constants/http-messages"
import HttpStatus from "../../services/constants/http-statuses"
import Messages from "../../services/constants/messages"
import ParkingLot from '../../models/parking-lot';
import { signJwt } from '../../middlewares/authentication';

interface IAddAdminInterface {
    email: string,
    password: string,
    confirmPassword: string,
    firstName?: string,
    middleName?: string,
    lastName?: string,
    avatar?: string,
    phoneNumber: {
        countryCode: string,
        number: number
    },
    generateRandom: boolean,
    role?: string,
    currentAdmin?: string,
}


export const createAdminService = async (req: Request, res: Response) => {
    try {
        const { email, password, confirmPassword, firstName, middleName, lastName, avatar, role, phoneNumber, generateRandom, currentAdmin }: IAddAdminInterface = req.body
        if (!validateEmail(email)) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL })
        } else if (!validatePassword(password) && !generateRandom) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_PASSWORD })
        } else if (!firstName) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.admin.ADMIN_DETAILS_MISSING })
        } else if (await Admin.findOne({ email: String(email).toLowerCase() })) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.admin.ADMIN_ALREADY_EXISTS })
        } else if (password != confirmPassword && !generateRandom) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.PASSWORDS_NOT_MATCH })
        } else {
            const currAdmin = await Admin.findById(currentAdmin)
            const generatedPassword = generateTemporaryPassword()
            const hashedPassword = await generateHashPassword(generateRandom ? generatedPassword.password : password)
            const admin = new Admin({
                avatar,
                email: String(email).toLowerCase(),
                password: hashedPassword,
                firstName,
                middleName,
                lastName,
                role: currAdmin?.role == 'sub-admin' ? 'sub-admin' : role,
                phoneNumber,
            })
            const savedAdmin = await admin.save()
            const emailSent = await sendEmail(email, Messages.emailSubjects.createAdmin, Messages.emailDetails.createAdmin + `Link: ${process.env.LIVE_BASE_URL}\nEmail: ${email}\nPassword: ${generateRandom ? generatedPassword.password : password}`)
            if (generateRandom) res.status(HttpStatus.CREATED).json({ ...savedAdmin.toObject(), password: generatedPassword, emailSent })
            else res.status(HttpStatus.CREATED).json(savedAdmin)
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSingleAdminService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const admin = await Admin.findById(_id).select(['-password', '-temporaryPassword'])
        if (admin)
            res.status(HttpStatus.OK).json(admin)
        else
            res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getParkingLotAdminsService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const parkingLot = await ParkingLot.findById(_id).populate({ path: 'admins', select: ['firstName', 'lastName', 'phoneNumber'] }).select(['admins'])
        if (parkingLot) {
            const { admins } = parkingLot
            res.status(HttpStatus.OK).json(admins)
        }
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingLot.PARKING_LOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAllAdminsService = async (req: Request, res: Response) => {
    try {
        const { currentAdmin, page = 1, limit = 10, filter }: { page?: number, limit?: number, currentAdmin?: string, filter?: string } = req.query
        const query: any = {
            _id: { $ne: currentAdmin },
        }
        if (filter) query.role = filter
        if (currentAdmin) {
            const admin = await Admin.findById(currentAdmin)
            if (admin?.role == 'sub-admin') query.role = 'sub-admin'
        }
        const totalCount = await Admin.countDocuments(query);
        const totalPages = Math.ceil(totalCount / limit);
        const admins = await Admin.find(query).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1)).select(['-password', '-temporaryPassword'])
        res.status(HttpStatus.OK).json({ totalPages, totalCount, admins })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const deleteAdminService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const admin = await Admin.findById(_id)
        if (admin) {
            if (admin.parkingLots.length > 0) {
                res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.admin.ADMIN_DELETE_DEPENDENCIES, })
            } else {
                await Admin.findByIdAndDelete(_id)
                res.status(HttpStatus.OK).json({ message: HttpMessages.admin.ADMIN_DELETED_SUCCESS })
            }
        }
        else
            res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const updateAdminService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const { firstName, middleName, lastName, avatar, role }: { firstName?: string, middleName?: string, lastName?: string, avatar?: string, role?: string } = req.body
        const admin = await Admin.findByIdAndUpdate(_id, { firstName, middleName, lastName, avatar, role }, { omitUndefined: true, new: true, })
        if (admin)
            res.status(HttpStatus.OK).json(admin)
        else
            res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const loginAdminService = async (req: Request, res: Response) => {
    try {
        const { email, password }: { email: string, password: string } = req.body
        if (!validateEmail(email)) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL })
        } else {
            const admin = await Admin.findOne({ email: String(email).toLowerCase() })
            if (admin) {
                if (admin.failedLoginAttempts > 2) {
                    res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.admin.ACCOUNT_LOCKED })
                } else {
                    const isValid = isPasswordValid(password, admin.password)
                    if (isValid) {
                        const isPasswordExpired = isDate90DaysOld(new Date(admin.passwordUpdatedAt))
                        if (isPasswordExpired) {
                            res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.admin.PASSWORD_EXPIRED })
                        } else {
                            admin.temporaryPassword = null
                            await admin.save()
                            const newAdmin = await Admin.findById(admin._id).select(['-password', '-temporaryPassword'])
                            const token = signJwt(admin?._id)
                            res.status(HttpStatus.OK).json({ admin: newAdmin, token })
                        }
                    } else {
                        admin.failedLoginAttempts++
                        admin.failedAttemptsDates.push(new Date())
                        await admin.save()
                        if (admin.failedLoginAttempts > 2) admin.accountLockInstances.push(new Date())
                        await admin.save()
                        res.status(HttpStatus.UNAUTHORIZED).json({ message: HttpMessages.admin.FAILED_LOGIN_ATTEMPT + `You have ${3 - admin.failedLoginAttempts} left.` })
                    }
                }
            } else
                res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const forgotPasswordService = async (req: Request, res: Response) => {
    try {
        const { email, isDev }: { email: string, isDev: boolean } = req.body
        if (!validateEmail(email)) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL })
        } else {
            const admin = await Admin.findOne({ email })
            if (admin) {
                if (admin.failedLoginAttempts > 2) {
                    res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.admin.ACCOUNT_LOCKED })
                } else {
                    const { expiry, password } = generateTemporaryPassword()
                    admin.temporaryPassword = await generateHashPassword(isDev ? 'Dev@12345' : password)
                    admin.temporaryPasswordExpiry = expiry
                    await admin.save()
                    await sendEmail(admin.email, Messages.emailSubjects.temporaryPasswordCreated, Messages.emailDetails.temporaryPasswordCreated + `Email: ${admin.email}\nPassword: ${isDev ? 'Dev@12345' : password}`)
                    res.status(HttpStatus.CREATED).json({ email, message: HttpMessages.admin.TEMPORARY_PASSWORD_CREATED })
                }
            } else
                res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const loginWithTemporaryPasswordService = async (req: Request, res: Response) => {
    try {
        const { email, temporaryPassword, newPassword, confirmPassword }: { email: string, temporaryPassword: string, newPassword: string, confirmPassword: string } = req.body
        if (!validateEmail(email)) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL })
        } else {
            const admin = await Admin.findOne({ email: String(email).toLowerCase() })
            if (admin) {
                if (admin.failedLoginAttempts > 2) {
                    res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.admin.ACCOUNT_LOCKED })
                } else {
                    const isValid = isPasswordValid(temporaryPassword, admin.temporaryPassword)
                    if (isValid) {
                        if (new Date(admin.temporaryPasswordExpiry) > new Date()) {
                            if (newPassword === confirmPassword) {
                                const hashedPassword = await generateHashPassword(newPassword)
                                admin.password = hashedPassword
                                admin.temporaryPassword = null
                                admin.passwordUpdatedAt = Date.now()
                                await admin.save()
                                res.status(HttpStatus.OK).json({ message: HttpMessages.admin.PASSWORD_UPDATE_SUCCESS })
                            } else
                                res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.validations.PASSWORDS_NOT_MATCH })
                        } else
                            res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.validations.TEMP_PASSWORD_EXPIRED })
                    } else
                        res.status(HttpStatus.UNAUTHORIZED).json({ message: HttpMessages.admin.FAILED_TEMPORARY_LOGIN })
                }
            } else
                res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const resetPasswordService = async (req: Request, res: Response) => {
    try {
        const { email, oldPassword, newPassword, confirmPassword }: { email: string, oldPassword: string, newPassword: string, confirmPassword: string } = req.body
        if (!validateEmail(email)) {
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL })
        } else {
            const admin = await Admin.findOne({ email })
            if (admin) {
                if (admin.failedLoginAttempts > 2) {
                    res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.admin.ACCOUNT_LOCKED })
                } else {
                    const isValid = isPasswordValid(oldPassword, admin.password)
                    if (isValid) {
                        if (newPassword === confirmPassword) {
                            const hashedPassword = await generateHashPassword(newPassword)
                            admin.password = hashedPassword
                            admin.temporaryPassword = null
                            admin.passwordUpdatedAt = Date.now()
                            await admin.save()
                            res.status(HttpStatus.OK).json({ message: HttpMessages.admin.PASSWORD_UPDATE_SUCCESS })
                        } else
                            res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.validations.PASSWORDS_NOT_MATCH })
                    } else
                        res.status(HttpStatus.UNAUTHORIZED).json({ message: HttpMessages.admin.WRONG_OLD_PASSWORD })
                }
            } else
                res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const unlockAdminAccountService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params
        const admin = await Admin.findByIdAndUpdate(_id, { failedLoginAttempts: 0, failedAttemptsDates: [], })
        if (admin) res.status(HttpStatus.OK).json({ message: HttpMessages.admin.ACCOUNT_UNLOCKED })
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}