import { Request, Response, } from 'express';
import HttpMessages from '../../services/constants/http-messages';
import HttpStatus from '../../services/constants/http-statuses';
import Complaint from '../../models/complaint';
import ParkingSpot from '../../models/parking-spot';
import User from '../../models/user';
import Admin from '../../models/admin';
import ParkingLot from '../../models/parking-lot';
import { sendMessage } from '../../middlewares/email-messaging';
import Vehicle from '../../models/vehicle';
import { generateComplaintsReportCsv } from '../../helpers/generators';
const complaintEnums = ['pending', 'opened', 'in-progress', 'closed', 'completed']

export const createComplaintService = async (req: Request, res: Response) => {
    try {
        const { parkingLot, complaintBy, details, }: { parkingLot: string, complaintBy: string, details: string, } = req.body;
        if (complaintBy && details) {
            const complaint = new Complaint({
                parkingLot,
                complaintBy,
                details,
            })
            if (parkingLot) {
                const found = await ParkingLot.findById(parkingLot).populate({ path: 'admins', select: ['phoneNumber'] }).select(['name'])
                const user = await User.findById(complaintBy).select(['firstName', 'phoneNumber'])
                if (found && user) {
                    const { admins } = found
                    if (admins.length > 0) {
                        const message = `דיווח משתמש בחניון ${found.name}, ממשתמש בשם: ${user.firstName} טלפון: ${user.phoneNumber.number}. ההודעה: ${details}`
                        await Promise.all(admins.map(async (admin: any) => {
                            return await sendMessage(admin.phoneNumber, message)
                        }))
                    }
                }
            }
            await complaint.save();
            res.status(HttpStatus.CREATED).json(complaint);
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.complaint.COMPLAINT_DETAILS_MISSING })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getAllComplaintsService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10, status, complaintBy, currentAdmin }: { page?: number, limit?: number, status?: string, complaintBy?: string, currentAdmin?: string, } = req.query
        let filter: any = {}
        if (currentAdmin) {
            const admin = await Admin.findById(currentAdmin).select(['parkingLots', 'role'])
            if (admin && admin.role == 'sub-admin') {
                const { parkingLots } = admin
                filter.parkingLot = { $in: parkingLots }
            }
        }
        if (status) filter.status = status
        if (complaintBy) filter.complaintBy = complaintBy
        const totalCount = await Complaint.countDocuments(filter);
        const totalPages = Math.ceil(totalCount / limit);
        const complaints = await Complaint.find(filter).sort({ updatedAt: -1, }).limit(limit).skip(limit * (page - 1))
            .populate('parkingLot', ['name',])
            .populate('complaintBy', ['firstName', 'middleName', 'lastName', 'phoneNumber', 'email',])
        res.status(HttpStatus.OK).json({ totalPages, totalCount, complaints });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getComplaintsReportService = async (req: Request, res: Response) => {
    try {
        const { currentAdmin }: { currentAdmin?: string, } = req.query
        let filter: any = {}
        if (currentAdmin) {
            const admin = await Admin.findById(currentAdmin).select(['parkingLots', 'role'])
            if (admin && admin.role == 'sub-admin') {
                const { parkingLots } = admin
                filter.parkingLot = { $in: parkingLots }
            }
        }
        const complaints = await Complaint.find(filter).sort({ createdAt: -1, })
            .populate('parkingLot', ['name',])
            .populate('complaintBy', ['firstName', 'middleName', 'lastName', 'phoneNumber', 'email',])
        const finalComplaints = await Promise.all(complaints.map((complaint) => {
            return {
                ...complaint.toObject(),
                status: complaint?.status == 'pending' ? 'פתוח' : complaint?.status == 'completed' ? 'סגור' : complaint?.status
            }
        }))
        const { filePath } = await generateComplaintsReportCsv(finalComplaints)
        res.status(HttpStatus.OK).json({ path: process.env.LIVE_BASE_URL + filePath });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const getSingleComplaintService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const complaint = await Complaint.findById(_id);
        if (complaint) res.status(HttpStatus.OK).json(complaint);
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.complaint.COMPLAINT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const updateComplaintService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { status }: { status: string } = req.body;
        if (complaintEnums.includes(status)) {
            const complaint = await Complaint.findByIdAndUpdate(_id, { status, }, { new: true, omitUndefined: true, });
            if (complaint) res.status(HttpStatus.OK).json(complaint)
            else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.complaint.COMPLAINT_NOT_FOUND })
        } else res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.complaint.COMPLAINT_INVALID_STATUS })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const spotTakenComplaintService = async (req: Request, res: Response) => {
    try {
        const { complaintBy, registration, parkingSpot, lang = "heb", }: { complaintBy: string, registration: string, parkingSpot: string, lang?: string, } = req.body
        const complaintByUser: any = await User.findById(complaintBy).populate('vehicles')
        if (complaintByUser) {
            const complaintAgainstVehicle = await Vehicle.findOne({ registration: { $regex: new RegExp(registration.toUpperCase(), 'i') } })
            if (complaintAgainstVehicle) {
                const user = await User.findById(complaintAgainstVehicle.user)
                if (user) {
                    const spot: any = await ParkingSpot.findById(parkingSpot).populate({ path: 'parkingLot', select: ['admins'], populate: { path: 'admins', select: 'phoneNumber' } })
                    if (spot) {
                        const complaintAgainstUser: any = await User.findById(user._id).populate('vehicles')
                        if (complaintAgainstUser) {
                            const sms = `שם המתלונן: ${complaintByUser.firstName} ${complaintByUser.lastName}
מספר טלפון המתלונן:  ${complaintByUser.phoneNumber.number}
מספר רכב חוסם: ${complaintAgainstVehicle.registration}
שם רכב החוסם: ${complaintAgainstUser.firstName} ${complaintAgainstUser.lastName}
מספר טלפון רכב חוסם:  ${complaintAgainstUser.phoneNumber.number}
מספר חניה שדווחה כחסומה: ${spot?.parkingCode}`
                            //                             const sms = `שם המתלונן: ${complaintByUser.firstName} ${complaintByUser.lastName}
                            // מספר טלפון המתלונן:  ${complaintByUser.phoneNumber.number}
                            // מספר רכב חוסם: ${complaintByUser.vehicles?.[0]?.registration}
                            // שם רכב החוסם: ${complaintAgainstUser.firstName} ${complaintAgainstUser.lastName}
                            // מספר טלפון רכב חוסם:  ${complaintAgainstUser.phoneNumber.number}
                            // מספר חניה שדווחה כחסומה: ${spot?.parkingCode}`
                            //                             const sms = `שם המתלונן: ${complaintByUser.firstName} ${complaintByUser.lastName}\n
                            // מספר טלפון המתלונן: ${complaintByUser.phoneNumber.countryCode + complaintByUser.phoneNumber.number}
                            // מספר רכב חוסם: ${registration}
                            // מספר טלפון רכב חוסם: ${complaintAgainstUser.phoneNumber.countryCode + complaintAgainstUser.phoneNumber.number}
                            // מספר חניה שדווחה כחסומה: ${spot?.parkingCode}`
                            const complaint = new Complaint({
                                parkingLot: spot.parkingLot?._id,
                                complaintBy,
                                // details: `Car with the plate number ${registration} was reported to unauthorizedly park in parking spot ${spot?.parkingCode}.\n\n${complaintAgainstUser?.phoneNumber?.countryCode + complaintAgainstUser?.phoneNumber?.number}`,
                                details: sms
                            })
                            const messagesSent: any[] = []
                            if (spot?.parkingLot?.admins) {
                                await Promise.all(spot?.parkingLot?.admins?.map(async (admin: any) => {
                                    const messageSent = await sendMessage(admin?.phoneNumber, sms)
                                    messagesSent.push(messageSent)
                                }))
                            }
                            const saved = await complaint.save()
                            res.status(HttpStatus.CREATED).json({ complaint: saved, messagesSent })
                        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
                    } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
                } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
            } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.vehicle.VEHICLE_NOT_FOUND })
        } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })

        // if (spot) {

        //     if (complaintByUser) {
        //         const complaintAgainstUser = await User.findById(vehicle?.user)
        //         if (complaintAgainstUser) {
        //             const complaint = new Complaint({
        //                 parkingLot: spot.parkingLot,
        //                 complaintBy,
        //                 details: `Car with the plate number ${registration} was reported to unauthorizedly park in parking spot ${spot?.parkingCode}.\n\n${complaintAgainstUser?.phoneNumber?.countryCode + complaintAgainstUser?.phoneNumber?.number}`,
        //             })
        //             const saved = await complaint.save()
        //             res.status(HttpStatus.CREATED).json({ complaint: saved })
        //         } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
        //     } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND })
        // } else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.parkingSpot.PARKING_SPOT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}

export const deleteComplaintService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const complaint = await Complaint.findByIdAndDelete(_id);
        if (complaint) res.status(HttpStatus.OK).json({ message: HttpMessages.complaint.COMPLAINT_DELETED_SUCCESS });
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.complaint.COMPLAINT_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err })
    }
}