import express, { Request, Response, Application } from "express";
import HttpMessages from "../../services/constants/http-messages";
import HttpStatus from "../../services/constants/http-statuses";
import {
  isDate90DaysOld,
  isPasswordValid,
  validateEmail,
  validatePassword,
} from "../../helpers/validators";
import {
  generateGuestUserReportCsv,
  generateHashPassword,
  generateTemporaryPinCode,
  generateUserReportCsv,
  registrationSmsGenerator,
} from "../../helpers/generators";
import { signJwt, verifyToken } from "../../middlewares/authentication";
const app: Application = express();
import User, { IUserInterface } from "../../models/user";
import ParkingSpot from "../../models/parking-spot";
import Vehicle, { IVehicle } from "../../models/vehicle";
import { sendMessage } from "../../middlewares/email-messaging";
import { Schema } from "mongoose";
import Reservation from "../../models/reservation";
import { getCorrectDateTime, getStartAndEndOfDay } from "../../helpers/dates";
import { deleteFile, fileReaderUsers } from "../../helpers/files";
import ParkingLot from "../../models/parking-lot";
import Admin from "../../models/admin";
import { reservationEnums } from "../reservations/controllers";
import moment from "moment";

interface IAddUserInterface {
  email: string;
  password: string;
  confirmPassword: string;
  firstName?: string;
  middleName?: string;
  lastName?: string;
  avatar?: string;
  phoneNumber: {
    countryCode: string;
    number: number;
  };
  electricCharging: boolean;
  designatedParkingSpot: string[];
  generateRandom: boolean;
  make?: string;
  model?: string;
  registration?: string;
  userSpot?: string;
  disabled?: boolean;
  parkingLots: Schema.Types.ObjectId[];
  mainParkingLot: Schema.Types.ObjectId | null;
  permanentDesignated?: Schema.Types.ObjectId[];
  vehicles: { make?: string; model?: string; registration: string }[];
}
interface IUpdateUserInterface {
  email?: string;
  firstName?: string;
  middleName?: string;
  lastName?: string;
  avatar?: string;
  phoneNumber?: {
    countryCode: string;
    number: string;
  };
  electricCharging?: boolean;
  designatedParkingSpot?: string[];
  parkingLots?: string[];
  mainParkingLot?: string;
  disabled?: boolean;
  vehicles?: IVehicle[];
  reservationDate?:Date
}

interface IGetUsersInterface {
  page?: number;
  limit?: number;
  type?: string;
  parkingSpotType?: string;
  isAvailable?: boolean | string;
  search?: string;
  sort?: "asc" | "dsc";
  isGuest?:boolean
  sortBy?:
  | "updated"
  | "created"
  | "firstName"
  | "middleName"
  | "lastName"
  | "registration"
  | "email"
  | "phoneNumber"
  | "designated";
}
interface IGetGuestUsersInterface {
  page?: number;
  limit?: number;
  type?: string;
  parkingSpotType?: string;
  isAvailable?: boolean | string;
  sort?: "asc" | "dsc";
  sortBy?:
  | "created"
  | "firstName"
  | "registration"
  | "phoneNumber"
  | "disabled";
  parkingLot?: string;
}

interface IGuestUser {
  firstName: string;
  phoneNumber: {
    countryCode: string;
    number: number;
  };
  email: string;
  electricCharging: boolean;
  disabled: boolean;
  reservationDate: Date;
  vehicles: { make?: string; model?: string; registration: string }[];
  parkingLots: Schema.Types.ObjectId[];
  designatedParkingSpot: Schema.Types.ObjectId[];
  mainParkingLot: Schema.Types.ObjectId;
}

interface IUpdateGuestInterface {
  firstName?: string;
  phoneNumber?: {
    countryCode: string;
    number: number;
  };
  email?: string;
  electricCharging?: boolean;
  disabled?: boolean;
  reservationDate?: Date;
  vehicles?: { make?: string; model?: string; registration: string }[];
  parkingLots?: Schema.Types.ObjectId[];
  designatedParkingSpot?: Schema.Types.ObjectId[];
  mainParkingLot?: Schema.Types.ObjectId;
}

const updateUserDetailsHelper = async (req: Request) => {
  const { _id }: { _id?: string } = req.params;
  const {
    email,
    firstName,
    middleName,
    lastName,
    avatar,
    phoneNumber,
    electricCharging,
    parkingLots,
    mainParkingLot,
    disabled,
  }: IUpdateUserInterface = req.body;
  return await User.findByIdAndUpdate(
    _id,
    {
      email: email ? email.toLowerCase() : undefined,
      firstName,
      middleName,
      lastName,
      avatar,
      phoneNumber,
      electricCharging,
      parkingLots,
      mainParkingLot,
      disabled,
    },
    { omitUndefined: true, new: true }
  ).select(["-password", "-temporaryPassword"]);
};

const updateGuestDetailsHelper = async (req: Request) => {
  const { _id }: { _id?: string } = req.params;
  const {
    email,
    firstName,
    phoneNumber,
    electricCharging,
    parkingLots,
    mainParkingLot,
    disabled,
  }: IUpdateGuestInterface = req.body;
  return await User.findByIdAndUpdate(
    _id,
    {
      email: email ? email.toLowerCase() : undefined,
      firstName,
      phoneNumber,
      electricCharging,
      parkingLots,
      mainParkingLot,
      disabled,
    },
    { omitUndefined: true, new: true }
  ).select(["-password", "-temporaryPassword"]).populate({ path: "parkingLots", select: ["name"] });
};

const updateDesignatedSpotsHelper = async (req: Request) => {
  const { _id }: { _id?: string } = req.params;
  const {
    designatedParkingSpot,
    reservationDate
  }: IUpdateUserInterface | IUpdateGuestInterface = req.body;

  if (designatedParkingSpot) {
    const foundUser = await User.findById(_id).select("designatedParkingSpot");

    //find reservations
    let reservation = null
    if(reservationDate){
      const { startOfDay, endOfDay } = getStartAndEndOfDay(new Date());
       reservation = await Reservation.find({
        parkingSpot: { $in: designatedParkingSpot },
        reservationDate: { $gte: startOfDay, $lte: endOfDay },
        $or: [
          { parkingStatus: reservationEnums[0] },
          { parkingStatus: reservationEnums[1] },
        ],
      });
    }


    if (
      designatedParkingSpot.length == 0 &&
      foundUser?.designatedParkingSpot &&
      foundUser?.designatedParkingSpot.length > 0
    ) {
      await Promise.all(
        foundUser.designatedParkingSpot.map(async spot => {
          await Reservation.updateMany(
            {
              parkingSpot: spot,
              user: foundUser._id,
              parkingStatus: {
                $in: [reservationEnums[0], reservationEnums[1]],
              },
            },
            { parkingStatus: reservationEnums[4] }
          );
          return await ParkingSpot.findByIdAndUpdate(spot, {
            designatedTo: null,
            userPresenceDate: null,
          });
        })
      );
      await User.findByIdAndUpdate(
        _id,
        { designatedParkingSpot: [] },
        { new: true }
      );
    } else if (designatedParkingSpot.length > 0) {
      if (foundUser) {
        await Reservation.updateMany(
          {
            parkingSpot: { $nin: designatedParkingSpot },
            user: foundUser._id,
            parkingStatus: { $in: [reservationEnums[0], reservationEnums[1]] },
          },
          { parkingStatus: reservationEnums[4] }
        );
        await Promise.all(
          foundUser.designatedParkingSpot.map(async spot => {
            const updated = await ParkingSpot.findByIdAndUpdate(spot, {
              designatedTo: null,
              userPresenceDate: null,
            });
            return updated;
          })
        );
      }
      const date = getCorrectDateTime(new Date())
      await Promise.all(
        designatedParkingSpot.map(async spot => {
           const updated =  await ParkingSpot.findByIdAndUpdate(spot, {
            designatedTo: foundUser?._id,
            userPresenceDate:reservationDate ? date : null,
          });

          if(updated) {
            const reservation = new Reservation({
              user: foundUser?._id,
              parkingLot: updated?.parkingLot,
              level: updated?.level,
              parkingSpot: updated._id,
              reservationDate: date,
              isElectricChargingRequired: false,
              parkingStatus: "reserved",
            });
            await reservation.save();
          }

          return updated
        })
      );
      await User.findByIdAndUpdate(
        _id,
        { designatedParkingSpot: designatedParkingSpot },
        { new: true }
      );
    }
  }
};


const  findReservations = async (designatedParkingSpot:Schema.Types.ObjectId[], reservationDate:Date)=> {
  const { startOfDay, endOfDay } = getStartAndEndOfDay(reservationDate);
  
  const reservation = await Reservation.find({
    parkingSpot: { $in: designatedParkingSpot },
    reservationDate: { $gte: startOfDay, $lte: endOfDay },
    $or: [
      { parkingStatus: reservationEnums[0] },
      { parkingStatus: reservationEnums[1] },
    ],
  });

  return reservation;
}

async function assignParkingSpot(user:IUserInterface, designatedParkingSpot:Schema.Types.ObjectId[], reservationDate:Date) {
  const spots = await Promise.all(
    designatedParkingSpot.map(async spot => {
      const date = reservationDate;
      const updated = await ParkingSpot.findByIdAndUpdate(
        spot,
        { designatedTo: user._id, userPresenceDate: date },
        { new: true }
      );

      if (updated) {
        const reservation = new Reservation({
          user: user._id,
          parkingLot: updated?.parkingLot,
          level: updated?.level,
          parkingSpot: updated._id,
          reservationDate: date,
          isElectricChargingRequired: false,
          parkingStatus: "reserved",
        });

        await reservation.save();
        return updated._id;
      } else return null
    })
  );

  return spots;
}


const updateVehiclesHelper = async (
  req: Request,
  _id: string,
  vehicles: IVehicle[]
) => {
  const errors: string[] = [];
  const updatedVehicles = await Promise.all(
    vehicles.map(async vehicle => {
      try {
        let updatedVehicle;
        const existingVehicle = await Vehicle.findOne({
          registration: vehicle.registration,
          _id: { $ne: vehicle?._id },
        });
        if (existingVehicle) {
          errors.push(
            `Vehicle with registration ${vehicle.registration} already exists.`
          );
          return null;
        } else {
          if (vehicle._id)
            updatedVehicle = await Vehicle.findByIdAndUpdate(
              vehicle._id,
              vehicle,
              { new: true }
            );
          else updatedVehicle = await Vehicle.create(vehicle);
        }
        return updatedVehicle;
      } catch (error) {
        console.error("Error updating/creating vehicle:", error);
        return null;
      }
    })
  );
  if (errors.length > 0) {
    return { vehicles: [], errors };
  } else {
    const updatedVehiclesIds = updatedVehicles
      .filter(vehicle => vehicle !== null)
      .map((vehicle: any) => vehicle._id);
    await Vehicle.deleteMany({ _id: { $nin: updatedVehiclesIds }, user: _id });
    const updated = await Vehicle.find({ user: _id });
    await User.findByIdAndUpdate(_id, {
      vehicles: updated.map(vehicle => vehicle._id),
    });
    return { vehicles: updated, errors };
  }
};

export const createUserService = async (req: Request, res: Response) => {
  try {
    const {
      email,
      password,
      confirmPassword,
      firstName,
      middleName,
      lastName,
      avatar,
      phoneNumber,
      electricCharging,
      designatedParkingSpot,
      generateRandom,
      userSpot,
      parkingLots,
      mainParkingLot,
      disabled,
      vehicles,
    }: IAddUserInterface = req.body;
    if (!email && email != "" && !validateEmail(email)) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL });
    } else if (!validatePassword(password) && !generateRandom) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_PASSWORD });
    } else if (!firstName) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.user.USER_DETAILS_MISSING });
    } else if (
      (!email &&
        email != "" &&
        (await User.findOne({ email: email.toLowerCase() }))) ||
      (await User.findOne({ "phoneNumber.number": phoneNumber.number }))
    ) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.user.USER_ALREADY_EXISTS });
    } else if (password != confirmPassword && !generateRandom) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.PASSWORDS_NOT_MATCH });
    } else if (!phoneNumber?.countryCode && !phoneNumber?.number) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_PHONE_NUMBER });
    } else if (!vehicles || vehicles.length == 0) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_DETAILS_MISSING });
    } else if (
      vehicles &&
      vehicles?.length > 0 &&
      (await Vehicle.findOne({
        registration: { $in: vehicles?.map(vehicle => vehicle?.registration) },
      }))
    ) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_REGISTRATION_DUPLICATE });
    } else {
      const generatedPassword = generateTemporaryPinCode();
      const hashedPassword = await generateHashPassword(
        generateRandom ? generatedPassword.password : password
      );
      const user = new User({
        avatar,
        email: email ? email.toLowerCase() : "",
        temporaryPassword: hashedPassword,
        firstName,
        middleName,
        lastName,
        phoneNumber,
        electricCharging,
        designatedParkingSpot: designatedParkingSpot
          ? designatedParkingSpot
          : [],
        userSpot,
        parkingLots,
        mainParkingLot,
        disabled,
      });
      const userWithDesignatedSpot = await User.findOne({
        designatedParkingSpot: { $in: designatedParkingSpot },
      });
      const { startOfDay, endOfDay } = getStartAndEndOfDay(new Date());
      const reservation = await Reservation.find({
        parkingSpot: { $in: designatedParkingSpot },
        reservationDate: { $gte: startOfDay, $lte: endOfDay },
        $or: [
          { parkingStatus: reservationEnums[0] },
          { parkingStatus: reservationEnums[1] },
        ],
      });
      if (userWithDesignatedSpot && designatedParkingSpot) {
        res.status(HttpStatus.FORBIDDEN).json({
          message:
            HttpMessages.parkingSpot.DESIGNATED_SPOT_ALREADY_ASSIGNED_TO_USER,
        });
      } else if (reservation.length > 0 && designatedParkingSpot) {
        res.status(HttpStatus.FORBIDDEN).json({
          message: HttpMessages.parkingSpot.DESIGNATED_SPOT_HAS_RESERVATIONS,
        });
      } else {
        const messageToSend = registrationSmsGenerator(
          firstName,
          generateRandom,
          generatedPassword,
          password
        );
        if (designatedParkingSpot) {
          const spots = await Promise.all(
            designatedParkingSpot.map(async spot => {
              const date = getCorrectDateTime(new Date());
              const updated = await ParkingSpot.findByIdAndUpdate(
                spot,
                { designatedTo: user._id, userPresenceDate: date },
                { new: true }
              );
              if (updated) {
                const reservation = new Reservation({
                  user: user._id,
                  parkingLot: updated?.parkingLot,
                  level: updated?.level,
                  parkingSpot: updated._id,
                  reservationDate: date,
                  isElectricChargingRequired: false,
                  parkingStatus: "reserved",
                });
                await reservation.save();
                return updated._id;
              } else return null;
            })
          );
          if (spots.length > 0) {
            const designated = await Promise.all(
              spots?.filter(item => {
                if (item != null && item != undefined) return item;
              })
            );
            user.designatedParkingSpot =
              designated?.length > 0 ? designated : [];
          }
        }
        if (vehicles && vehicles.length > 0) {
          const createdVehicles = await Promise.all(
            vehicles.map(async vehicle => {
              const created = new Vehicle({
                user: user._id,
                make: vehicle?.make,
                model: vehicle?.model,
                registration: vehicle.registration.toUpperCase(),
              });
              return await created.save();
            })
          );
          user.vehicles = createdVehicles.map(vehicle => vehicle._id);
        }
        const savedUser = await user.save();
        const messageSent = await sendMessage(user.phoneNumber, messageToSend);
        res.status(HttpStatus.CREATED).json({ user: savedUser, messageSent });
      }
    }
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const createGuestUserService = async (req: Request, res: Response) => {
  try {
    const {
      firstName,
      phoneNumber,
      email,
      electricCharging,
      disabled,
      reservationDate,
      designatedParkingSpot,
      parkingLots,
      mainParkingLot,
      vehicles,
    }: IGuestUser = req.body;
    if (!email && email != "" && !validateEmail(email)) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_EMAIL });
      return
    } else if (!firstName) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.user.USER_DETAILS_MISSING });
      return
    } else if (
      (!email &&
        email != "" &&
        (await User.findOne({ email: email.toLowerCase() }))) ||
      (await User.findOne({ "phoneNumber.number": phoneNumber.number }))
    ) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.user.USER_ALREADY_EXISTS });
      return
    } else if (!phoneNumber?.countryCode && !phoneNumber?.number) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_PHONE_NUMBER });
      return
    } else if (!vehicles || vehicles.length == 0) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_DETAILS_MISSING });
      return
    } else if (
      vehicles &&
      vehicles?.length > 0 &&
      (await Vehicle.findOne({
        registration: { $in: vehicles?.map(vehicle => vehicle?.registration) },
      }))
    ) {
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_REGISTRATION_DUPLICATE });
      return
    } else {
      const user = new User({
        email: email ? email.toLowerCase() : "",
        firstName,
        phoneNumber,
        electricCharging,
        designatedParkingSpot: designatedParkingSpot
          ? designatedParkingSpot
          : [],
        parkingLots,
        mainParkingLot,
        disabled,
        isGuest: true,
      });
      const userWithDesignatedSpot = await User.findOne({
        designatedParkingSpot: { $in: designatedParkingSpot },
      });

      const reservation = await findReservations(designatedParkingSpot,reservationDate)

      if (userWithDesignatedSpot && designatedParkingSpot) {
        return res.status(HttpStatus.FORBIDDEN).json({
          message:
            HttpMessages.parkingSpot.DESIGNATED_SPOT_ALREADY_ASSIGNED_TO_USER,
        });
      } else if (reservation.length > 0 && designatedParkingSpot) {
        return res.status(HttpStatus.FORBIDDEN).json({
          message: HttpMessages.parkingSpot.DESIGNATED_SPOT_HAS_RESERVATIONS,
        });
      } else {
        const parkingLotsDetails = await ParkingLot.find({
          _id: { $in: parkingLots },
        });
        const designatedSpotsDetails: any[] = await ParkingSpot.find({
          _id: { $in: designatedParkingSpot },
        }).populate({ path: "level", select: ["name"] });
        const messageToSend = `${firstName} שלום,
הוזמנה לך חניה במתקן חברת חשמל.
תאריך ההזמנה: מ-${moment(reservationDate).format("DD-MMM-YYYY")}
שם המתקן: ${parkingLotsDetails?.map(parkingLot => parkingLot?.name)?.join(", ")}
כתובת: ${parkingLotsDetails
            ?.map(parkingLot => parkingLot?.location?.name)
            ?.join(", ")}, קישור לwaze: ${parkingLotsDetails
              ?.map(
                parkingLot =>
                  `https://waze.com/ul?q=${parkingLot?.location?.latitude},${parkingLot?.location?.longitude}`
              )
              ?.join(", ")}
מספר חניה: ${designatedSpotsDetails
            ?.map(designatedSpot => designatedSpot?.parkingCode)
            ?.join(", ")}
אזור/קומה: ${designatedSpotsDetails
            ?.map(designatedSpot => designatedSpot?.level?.name)
            ?.join(", ")}`;
        if (designatedParkingSpot) {
          const spots = await assignParkingSpot(user,designatedParkingSpot,reservationDate)
          if (spots.length > 0) {
            const designated = await Promise.all(
              spots?.filter(item => {
                if (item != null && item != undefined) return item;
              })
            );
            user.designatedParkingSpot =
              designated?.length > 0 ? designated : [];
          }
        }
        if (vehicles && vehicles.length > 0) {
          const createdVehicles = await Promise.all(
            vehicles.map(async vehicle => {
              const created = new Vehicle({
                user: user._id,
                make: vehicle?.make,
                model: vehicle?.model,
                registration: vehicle.registration.toUpperCase(),
              });
              return await created.save();
            })
          );
          user.vehicles = createdVehicles.map(vehicle => vehicle._id);
        }
        const savedUser = await user.save();
        const messageSent = await sendMessage(user.phoneNumber, messageToSend)
        res.status(HttpStatus.CREATED).json({ user: savedUser, messageSent, });
      }
    }
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const createMultipleUsersService = async (req: Request, res: Response) => {
  try {
    const {
      fileUrl,
      lang = "heb",
    }: { fileUrl: string; generateRandom: boolean; lang: string } = req.body;
    const usersFromFile = await fileReaderUsers(fileUrl);
    const users: { user: any; vehicles: any[] }[] = await Promise.all(
      usersFromFile.map(async (user, index) => {
        return await new Promise(async (resolve, reject) => {
          try {
            if (
              !user.email &&
              user.email != "" &&
              (await User.findOne({ email: user.email?.trim().toLowerCase() }))
            ) {
              reject({ type: "email-already-exist", index });
            } else if (
              await User.findOne({
                "phoneNumber.number": user.phoneNumber?.trim(),
              })
            ) {
              reject({ type: "phone-number-already-exist", index });
            } else if (
              await Vehicle.findOne({ registration: user.registration?.trim() })
            ) {
              reject({ type: "vehicle-registration-already-exist", index });
            } else {
              const mainParkingLot = await ParkingLot.findOne({
                name: user.mainParkingLot,
              });
              if (!mainParkingLot)
                reject({ type: "main-parking-lot-not-found", index });
              else {
                const parkingLots = await Promise.all(
                  user.allowedParkingLots.map(async (parkingLot: string) => {
                    try {
                      const lot = await ParkingLot.findOne({
                        name: parkingLot,
                      });
                      if (!lot)
                        reject({
                          type: "allowed-parking-lots-not-found",
                          index,
                        });
                      else return lot;
                    } catch (error) {
                      reject(error);
                    }
                  })
                );
                const designatedSpots = await Promise.all(
                  user.designatedSpots.map(async (spot: string) => {
                    try {
                      const parkingSpot = await ParkingSpot.findOne({
                        parkingLot: { $in: parkingLots.map(item => item?._id) },
                        parkingCode: spot,
                        designatedTo: null,
                        spotType: "designated",
                      });
                      if (!parkingSpot)
                        reject({ type: "designated-spot-not-found", index });
                      else return parkingSpot;
                    } catch (error) {
                      reject(error);
                    }
                  })
                );
                const generatedPassword = generateTemporaryPinCode();
                const hashedPassword = await generateHashPassword(
                  generatedPassword.password
                );
                const newUser = new User({
                  avatar: "",
                  email:
                    user.email != "" && user.email != undefined
                      ? user.email.trim().toLowerCase()
                      : "",
                  temporaryPassword: hashedPassword,
                  firstName: user.firstName,
                  middleName: "",
                  lastName: user.lastName,
                  phoneNumber: {
                    countryCode: "+972",
                    number: user.phoneNumber,
                  },
                  electricCharging: user.electricCharging,
                  designatedParkingSpot: designatedSpots.map(item => item?._id),
                  parkingLots: parkingLots.map(item => item?._id),
                  mainParkingLot: mainParkingLot?._id,
                  disabled: user.accessible,
                  dateOfBirth: user.dateOfBirth,
                });
                const vehicle = new Vehicle({
                  user: newUser._id,
                  make: user.make,
                  model: user.model,
                  registration: user.registration.toUpperCase(),
                });
                newUser.vehicles = [vehicle._id];
                resolve({ user: newUser, vehicles: [vehicle] });
              }
            }
          } catch (error) {
            reject(error);
          }
        });
      })
    );
    const savedUsers = await Promise.all(
      users.map(async (item: any) => {
        await Promise.all(
          item.user.designatedParkingSpot.map(async (spot: any) => {
            const parkingSpot = await ParkingSpot.findByIdAndUpdate(spot?._id, {
              designatedTo: item?.user?._id,
            });
            const date = getCorrectDateTime(new Date());
            const reservation = new Reservation({
              user: item?.user?._id,
              parkingLot: parkingSpot?.parkingLot,
              level: parkingSpot?.level,
              parkingSpot: parkingSpot?._id,
              reservationDate: date,
              isElectricChargingRequired: false,
              parkingStatus: "reserved",
            });
            await reservation.save();
          })
        );
        await Promise.all(
          item.vehicles.map(async (vehicle: any) => {
            await vehicle.save();
          })
        );
        return await item?.user?.save();
      })
    );
    // res.status(HttpStatus.CREATED).json(users)
    res.status(HttpStatus.CREATED).json(savedUsers);
  } catch (err: any) {
    const { fileUrl, lang = "heb", }: { fileUrl: string; generateRandom: boolean; lang: string } = req.body;
    await deleteFile(fileUrl);
    if (err?.type == "main-parking-lot-not-found") res.status(HttpStatus.NOT_FOUND).json({ message: "Main parking lot not found at row " + (Number(err?.index) + 1), });
    else if (err?.type == "allowed-parking-lots-not-found") res.status(HttpStatus.NOT_FOUND).json({ message: "Allowed parking lot not found at row " + (Number(err?.index) + 1), });
    else if (err?.type == "designated-spot-not-found") res.status(HttpStatus.NOT_FOUND).json({ message: "Designated parking spot not found at row " + (Number(err?.index) + 1), });
    else if (err?.type == "email-already-exist") res.status(HttpStatus.FORBIDDEN).json({ message: `${lang == "heb" ? HttpMessages.user.EMAIL_ALREADY_EXISTS_HEB : HttpMessages.user.EMAIL_ALREADY_EXISTS} ${Number(err?.index) + 1}`, });
    else if (err?.type == "phone-number-already-exist") res.status(HttpStatus.FORBIDDEN).json({ message: `${lang == "heb" ? HttpMessages.user.PHONE_ALREADY_EXISTS_HEB : HttpMessages.user.PHONE_ALREADY_EXISTS} ${Number(err?.index) + 1}`, });
    else if (err?.type == "vehicle-registration-already-exist") res.status(HttpStatus.FORBIDDEN).json({ message: `${lang == "heb" ? HttpMessages.user.REG_ALREADY_EXISTS_HEB : HttpMessages.user.REG_ALREADY_EXISTS} ${Number(err?.index) + 1}`, });
    else res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const getSingleUserService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const user = await User.findById(_id).select(["-password", "-temporaryPassword"]).populate({ path: "designatedParkingSpot", populate: "parkingLot" });
    if (user) res.status(HttpStatus.OK).json(user);
    else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const getAdminLotsUsersService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const {
      page = 1,
      limit = 10,
      type,
      parkingSpotType,
      isAvailable,
      search,
      sortBy = "updated",
      sort = "dsc",
    }: IGetUsersInterface = req.query;
    const sorting = sort == "asc" ? 1 : -1;
    const admin = await Admin.findById(_id).populate("parkingLots").select("parkingLots");
    if (admin) {
      const { parkingLots } = admin;
      let filter: any = { parkingLots: { $in: parkingLots }, $or: [{ isGuest: false }, { isGuest: undefined }] };
      if (search) {
        const searchRegex = new RegExp(search, "i"); // Case-insensitive regex
        filter.$or = [
          { email: searchRegex },
          { "phoneNumber.number": searchRegex },
          { firstName: searchRegex },
          { middleName: searchRegex },
          { lastName: searchRegex },
          { "vehicle.registration": searchRegex },
        ];
      }
      if (type == "locked") filter.failedLoginAttempts = { $gte: 2 };
      else if (type == "unlocked") filter.failedLoginAttempts = { $lte: 3 };
      if (parkingSpotType == "designated")
        filter.designatedParkingSpot = { $ne: null };
      else if (parkingSpotType == "regular")
        filter.designatedParkingSpot = null;
      const sortingFilter: any = {};
      if (sortBy == "updated") sortingFilter.updatedAt = sorting;
      else if (sortBy == "created") sortingFilter.createdAt = sorting;
      else if (sortBy == "firstName") sortingFilter.firstName = sorting;
      else if (sortBy == "middleName") sortingFilter.middleName = sorting;
      else if (sortBy == "lastName") sortingFilter.lastName = sorting;
      else if (sortBy == "registration") sortingFilter.vehicles = sorting;
      else if (sortBy == "email") sortingFilter.email = sorting;
      else if (sortBy == "phoneNumber") sortingFilter.phoneNumber = sorting;
      else if (sortBy == "designated")
        sortingFilter.designatedParkingSpot = sorting;
      const totalCount = await User.countDocuments(filter);
      const totalPages = Math.ceil(totalCount / limit);
      const users = await User.find(filter)
        .sort(sortingFilter)
        .limit(limit)
        .skip(limit * (page - 1))
        .select([
          "-password",
          "-temporaryPassword",
          "-passwordUpdatedAt",
          "-temporaryPasswordExpiry",
        ])
        .populate({
          path: "designatedParkingSpot",
          populate: { path: "parkingLot", select: ["name"] },
        })
        .populate({ path: "parkingLots", select: ["name"] })
        .populate("vehicles");
      const usersWithVehicles = await Promise.all(
        users.map(async (user: any) => {
          const userWithDesignatedSpots = await Promise.all(
            user.parkingLots.map(async (parkingLot: any) => {
              const designatedSpots = await ParkingSpot.find({
                parkingLot: parkingLot._id,
                spotType: "designated",
                designatedTo: isAvailable ? null : { $ne: null },
              });
              return {
                ...parkingLot.toObject(),
                designatedSpots,
              };
            })
          );
          return {
            ...user.toObject(),
            parkingLots: userWithDesignatedSpots,
          };
        })
      );
      res.status(HttpStatus.OK).json({ totalPages, totalCount, users: usersWithVehicles });
    } else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.admin.ADMIN_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const getAllUsersService = async (req: Request, res: Response) => {
  try {
    const {
      page = 1,
      limit = 10,
      type,
      parkingSpotType,
      isAvailable,
      search,
      sortBy = "updated",
      sort = "dsc",
    }: IGetUsersInterface = req.query;

    const isGuest =req.query?.isGuest == "false" ? false : undefined

    const sorting = sort == "asc" ? 1 : -1;
    let filter: any = {};
    if (search) {
      const searchRegex = new RegExp(search, "i"); // Case-insensitive regex
      if (searchRegex) {
        const sortingFilter: any = {};
        if (sortBy == "registration") sortingFilter.registration = sorting;

        const vehiclesPromise =  Vehicle.find({
          $or: [
            { registration: searchRegex },
            { make: searchRegex },
            { model: searchRegex },
          ],
        }).sort(sortingFilter);

        const parkingSpotsPromise = ParkingSpot.find({
          "parkingCode":searchRegex
        })
        const parkingLotsPromise = ParkingLot.find({
          "name":searchRegex
        })
        filter.$or = await Promise.all([vehiclesPromise,parkingSpotsPromise,parkingLotsPromise]).then(([vehicles,parkingSpots,parkingLots])=>{
          return  [
            { vehicles: { $in: vehicles.map(vehicle => vehicle.id) } },
            { designatedParkingSpot: { $in: parkingSpots.map(spot => spot.id) } },
            { parkingLots: { $in: parkingLots.map(spot => spot.id) } },
          ];
        })

      }
      filter.$or = [
        ...filter.$or,
        { email: searchRegex },
        { "phoneNumber.number": searchRegex },
        { firstName: searchRegex },
        { middleName: searchRegex },
        { lastName: searchRegex },
        { "vehicle.registration": searchRegex },
      ];
    }
    if (type == "locked") filter.failedLoginAttempts = { $gt: 2 };
    else if (type == "unlocked") filter.failedLoginAttempts = { $lt: 3 };
    if (parkingSpotType == "designated")
      filter.designatedParkingSpot = { $ne: null };
    else if (parkingSpotType == "regular") filter.designatedParkingSpot = null;
    const sortingFilter: any = {};
    if (sortBy == "updated") sortingFilter.updatedAt = sorting;
    else if (sortBy == "created") sortingFilter.createdAt = sorting;
    else if (sortBy == "firstName") sortingFilter.firstName = sorting;
    else if (sortBy == "middleName") sortingFilter.middleName = sorting;
    else if (sortBy == "lastName") sortingFilter.lastName = sorting;
    else if (sortBy == "registration") sortingFilter.vehicles = sorting;
    else if (sortBy == "email") sortingFilter.email = sorting;
    else if (sortBy == "phoneNumber") sortingFilter.phoneNumber = sorting;
    else if (sortBy == "designated") sortingFilter.designatedParkingSpot = sorting;

    filter = {
      ...filter,
      ...(typeof isGuest === "boolean" ? {isGuest} : {})
    }
    const totalCount = await User.countDocuments(filter);
    const totalPages = Math.ceil(totalCount / limit);
    const users = await User.find(filter)
      .sort(sortingFilter)
      .limit(limit)
      .skip(limit * (page - 1))
      .select([
        "-password",
        "-temporaryPassword",
        "-passwordUpdatedAt",
        "-temporaryPasswordExpiry",
      ])
      .populate({
        path: "designatedParkingSpot",
        populate: { path: "parkingLot", select: ["name"] },
      })
      .populate({ path: "parkingLots", select: ["name"] })
      .populate({ path: "vehicles" });
    const usersWithVehicles = await Promise.all(
      users.map(async (user: any) => {
        const userWithDesignatedSpots = await Promise.all(
          user.parkingLots.map(async (parkingLot: any) => {
            const designatedSpots = await ParkingSpot.find({
              parkingLot: parkingLot._id,
              spotType: "designated",
              designatedTo: isAvailable ? null : { $ne: null },
            });
            return {
              ...parkingLot.toObject(),
              designatedSpots,
            };
          })
        );
        return {
          ...user.toObject(),
          parkingLots: userWithDesignatedSpots,
        };
      })
    );
    res.status(HttpStatus.OK).json({ totalPages, totalCount, users: usersWithVehicles });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const getAllGuestUsersService = async (req: Request, res: Response) => {
  try {
    const {
      page = 1,
      limit = 10,
      type,
      parkingSpotType,
      isAvailable,
      sortBy = "created",
      sort = "dsc",
      parkingLot,
    }: IGetGuestUsersInterface = req.query;
    const sorting = sort == "asc" ? 1 : -1;
    let filter: any = { isGuest: true, parkingLots: { $in: [parkingLot] } };
    if (type == "locked") filter.failedLoginAttempts = { $gt: 2 };
    else if (type == "unlocked") filter.failedLoginAttempts = { $lt: 3 };
    if (parkingSpotType == "designated")
      filter.designatedParkingSpot = { $ne: null };
    else if (parkingSpotType == "regular") filter.designatedParkingSpot = null;
    const sortingFilter: any = {};
    if (sortBy == "created") sortingFilter.createdAt = sorting;
    else if (sortBy == "firstName") sortingFilter.firstName = sorting;
    else if (sortBy == "phoneNumber") sortingFilter.phoneNumber = sorting;
    else if (sortBy == "disabled") sortingFilter.disabled = sorting;
    else if (sortBy == "registration") sortingFilter.vehicles = sorting;
    else if (sortBy == "designated")
      sortingFilter.designatedParkingSpot = sorting;
    const totalCount = await User.countDocuments(filter);
    const totalPages = Math.ceil(totalCount / limit);
    const users = await User.find(filter)
      .sort(sortingFilter)
      .limit(limit)
      .skip(limit * (page - 1))
      .select([
        "-password",
        "-temporaryPassword",
        "-passwordUpdatedAt",
        "-temporaryPasswordExpiry",
      ])
      .populate({
        path: "designatedParkingSpot",
        populate: { path: "parkingLot", select: ["name"] },
      })
      .populate({ path: "parkingLots", select: ["name"] })
      .populate({ path: "vehicles" });
    const usersWithVehicles = await Promise.all(
      users.map(async (user: any) => {
        const userWithDesignatedSpots = await Promise.all(
          user.parkingLots.map(async (parkingLot: any) => {
            const designatedSpots = await ParkingSpot.find({
              parkingLot: parkingLot._id,
              spotType: "designated",
              designatedTo: isAvailable ? null : { $ne: null },
            });
            return {
              ...parkingLot.toObject(),
              designatedSpots,
            };
          })
        );
        return {
          ...user.toObject(),
          parkingLots: userWithDesignatedSpots,
        };
      })
    );
    res.status(HttpStatus.OK).json({ totalPages, totalCount, users: usersWithVehicles });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const searchGuestService = async (req: Request, res: Response) => {
  try {
    const { search }: { search?: string } = req.params;

    if (!search) res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.user.USER_DETAILS_MISSING });
    const regex = new RegExp(search, "igm");
    let guests = await User.find({
      $and: [
        { isGuest: true },
        {
          $or: [
            { firstName: { $regex: regex } },
            { "phoneNumber.countryCode": { $regex: regex } },
            { "phoneNumber.number": { $regex: regex } },
          ],
        },
      ],
    }).populate({ path: 'vehicles' }).populate({ path: 'designatedParkingSpot', populate: 'parkingLot' }).populate({ path: 'parkingLots' });


    console.log("guests",guests,regex)
    if (!guests.length) {
      const users = await User.find({}).populate({ path: 'vehicles' }).populate({ path: 'designatedParkingSpot', populate: 'parkingLot' }).populate({ path: 'parkingLots' })
      guests = users.filter(user => user.vehicles.some((vehicle: any) => regex.test(vehicle?.registration) ||  (user.phoneNumber && regex.test(user.phoneNumber.number)))) 
      if (!guests.length) {
        guests = users.filter(user => user.designatedParkingSpot.some((parkingSpot: any) => regex.test(parkingSpot?.parkingCode)));
      }
      
    }
    const usersWithVehicles = await Promise.all(
      guests.map(async (user: any) => {
        const userWithDesignatedSpots = await Promise.all(
          user.parkingLots.map(async (parkingLot: any) => {
            const designatedSpots = await ParkingSpot.find({
              parkingLot: parkingLot._id,
              spotType: "designated",
              designatedTo: { $ne: null },
            });
            return {
              ...parkingLot.toObject(),
              designatedSpots,
            };
          })
        );
        return {
          ...user.toObject(),
          parkingLots: userWithDesignatedSpots,
        };
      })
    )
    res.status(HttpStatus.OK).json({ users: usersWithVehicles });
  } catch (error) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: error });
  }
};

export const getUsersReportService = async (req: Request, res: Response) => {
  try {
    const { currentAdmin }: { currentAdmin?: string } = req.query;
    let filter: {} = {};
    if (currentAdmin) {
      const admin = await Admin.findById(currentAdmin)
        .populate("parkingLots")
        .select("parkingLots");
      if (admin) {
        const { parkingLots } = admin;
        filter = { parkingLots: { $in: parkingLots } };
      }
    }
    const users = await User.find(filter)
      .select([
        "firstName",
        "phoneNumber",
        "disabled",
        "electricCharging",
        "firstLoginDate",
        "designatedParkingSpot",
        "parkingLots",
        "createdAt",
        "updatedAt",
        "deviceToken",
      ])
      .populate({
        path: "designatedParkingSpot",
        select: ["parkingCode", "level"],
        populate: {
          path: "level",
          select: ["name"],
        },
      })
      .populate({ path: "parkingLots", select: ["name"] });
    const { filePath } = await generateUserReportCsv(users);
    res.status(HttpStatus.OK).json({ path: process.env.LIVE_BASE_URL + filePath });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const getGuestUsersReportService = async (req: Request, res: Response) => {
  try {
    const filter: any = { isGuest: true, };
    const users = await User.find(filter)
      .select([
        "-password",
        "-temporaryPassword",
        "-passwordUpdatedAt",
        "-temporaryPasswordExpiry",
        "-deviceToken",
        "-avatar",
        "-userSpot",
        "-failedLoginAttempts",
        "-failedAttemptsDates",
        "-accountLockInstances",
        "-mainParkingLot",
        "-dateOfBirth",
        "-firstLoginDate",
        "-lastLogin",
        "-email",
      ])
      .populate({
        path: "designatedParkingSpot",
        select: ['parkingCode']
      })
      .populate({ path: "parkingLots", select: ["name"] })
      .populate({ path: "vehicles", select: ['registration'] })
    const usersWithReservation = await Promise.all(users.map(async (user: any) => {
      const dbReservations = await Reservation.find({ user: user._id, parkingSpot: { $in: user.designatedParkingSpot.map((spot: any) => spot._id) } }).select(['reservationDate'])
      const reservations = dbReservations?.map((reservation: any) => reservation?.reservationDate?.toISOString())?.join(', ')
      const vehicles = user.vehicles.map((vehicle: any) => vehicle.registration).join(', ')
      const parkingLots = user.parkingLots.map((parkingLot: any) => parkingLot.name).join(', ')
      const designatedParkingSpot = user.designatedParkingSpot.map((spot: any) => spot.parkingCode).join(', ')
      const disabled = user.disabled ? 'כן' : 'לא'
      return {
        ...user?.toObject(),
        reservations,
        vehicles,
        parkingLots,
        designatedParkingSpot,
        disabled,
      }
    }))
    const { filePath } = await generateGuestUserReportCsv(usersWithReservation)
    res.status(HttpStatus.OK).json({ path: process.env.LIVE_BASE_URL + filePath });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const updateUserService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const { vehicles }: IUpdateUserInterface = req.body;
    const found = await User.findById(_id);
    if (!found)
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
    else {
      let updatedVehicles: any[] = [];
      if (vehicles) {
        const data = await updateVehiclesHelper(req, _id, vehicles);
        if (data.errors.length > 0)
          res.status(HttpStatus.BAD_REQUEST).json({
            message: HttpMessages.vehicle.VEHICLE_REGISTRATION_DUPLICATE,
          });
        else {
          const user = await updateUserDetailsHelper(req);
          await updateDesignatedSpotsHelper(req);
          updatedVehicles = data.vehicles;
          res.status(HttpStatus.OK).json({ user, vehicles: updatedVehicles });
        }
      } else {
        const user = await updateUserDetailsHelper(req);
        await updateDesignatedSpotsHelper(req);
        res.status(HttpStatus.OK).json({ user, vehicles: [] });
      }
    }
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const updateGuestService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const { vehicles }: IUpdateUserInterface = req.body;
    let updatedVehicles: any[] = [];
    const found = await User.findById(_id)
    if (!found)
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
    else {
      if (vehicles) {
        await Vehicle.findByIdAndDelete(_id);
        updatedVehicles = await Promise.all(
          vehicles.map(async ({_id,...vehicle}) => {
            return await Vehicle.create({ ...vehicle, user: _id });
          })
        );
        await User.findByIdAndUpdate(_id, {
          vehicles: updatedVehicles.map(vehicle => vehicle._id),
        });
      }
      const guest = await updateGuestDetailsHelper(req);
      await updateDesignatedSpotsHelper(req);
      res.status(HttpStatus.OK).json({ user: guest, vehicles: updatedVehicles || [] });
    }
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const loginUserService = async (req: Request, res: Response) => {
  try {
    const {
      phoneNumber,
      password,
      deviceToken,
    }: { phoneNumber: string; password: string; deviceToken: string } =
      req.body;
    if (phoneNumber && password && deviceToken) {
      const user = await User.findOne({ "phoneNumber.number": phoneNumber, $or: [{ isGuest: false }, { isGuest: undefined }] });
      if (user) {
        if (user.failedLoginAttempts > 2) {
          res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.user.ACCOUNT_LOCKED });
        } else {
          const isValid = isPasswordValid(password, user.password);
          if (isValid) {
            const isPasswordExpired = isDate90DaysOld(
              new Date(user.passwordUpdatedAt)
            );
            if (isPasswordExpired)
              res
                .status(HttpStatus.FORBIDDEN)
                .json({ message: HttpMessages.user.PASSWORD_EXPIRED });
            else {
              // user.temporaryPassword = null
              // user.deviceToken = deviceToken
              const savedUser = await User.findByIdAndUpdate(
                user._id,
                {
                  temporaryPassword: null,
                  deviceToken,
                  lastLogin: new Date(),
                  ...(!user?.firstLoginDate && { firstLoginDate: new Date() }),
                },
                { new: true, omitUndefined: true }
              )
                .select(["-password", "-temporaryPassword"])
                .populate({
                  path: "designatedParkingSpot",
                  populate: "parkingLot",
                });
              // if (!user?.firstLoginDate) await User.findByIdAndUpdate(user._id, { firstLoginDate: new Date() })
              // await user.save()
              // const finalUser = await User.findById(user._id).select(['-password', '-temporaryPassword']).populate({ path: 'designatedParkingSpot', populate: 'parkingLot' })
              const token = signJwt(user._id);
              res.status(HttpStatus.OK).json({ user: savedUser, token });
            }
          } else {
            const isTemporaryPassword = user.temporaryPassword
              ? isPasswordValid(password, user.temporaryPassword)
              : false;
            if (isTemporaryPassword) {
              res.status(HttpStatus.ACCEPTED).json({
                user: { phoneNumber: user.phoneNumber },
                message: HttpMessages.user.IS_TEMPORARY_PASSWORD,
              });
            } else {
              user.failedLoginAttempts++;
              user.failedAttemptsDates.push(new Date());
              await user.save();
              if (user.failedLoginAttempts > 2)
                user.accountLockInstances.push(new Date());
              await user.save();
              res.status(HttpStatus.UNAUTHORIZED).json({
                message:
                  HttpMessages.user.FAILED_LOGIN_ATTEMPT +
                  `You have ${3 - user.failedLoginAttempts} left.`,
              });
            }
          }
        }
      } else
        res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
    } else
      res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.validations.INVALID_LOGIN_CREDENTIALS });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const logoutUserService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const user = await User.findByIdAndUpdate(_id, { deviceToken: null });
    if (user)
      res.status(HttpStatus.OK).json({ message: HttpMessages.user.ACCOUNT_LOGGED_OUT });
    else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const forgotPasswordService = async (req: Request, res: Response) => {
  try {
    const {
      phoneNumber,
      isDev,
    }: {
      phoneNumber: { countryCode: string; number: string };
      isDev: boolean;
    } = req.body;
    const user = await User.findOne({
      "phoneNumber.number": phoneNumber.number,
    });
    if (user) {
      // if (user.failedLoginAttempts > 2) {
      //     res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.user.ACCOUNT_LOCKED })
      // } else {
      const { expiry, password } = generateTemporaryPinCode();
      // console.log(password, 'password')
      await User.findByIdAndUpdate(user._id, {
        temporaryPassword: await generateHashPassword(
          isDev ? "Dev@12345" : password
        ),
        temporaryPasswordExpiry: expiry,
        failedLoginAttempts: 0,
        failedAttemptsDates: [],
      });
      if (isDev) {
        res.status(HttpStatus.CREATED).json({
          phoneNumber,
          message: HttpMessages.user.TEMPORARY_PASSWORD_CREATED_PHONE,
        });
      } else {
        const sms = `${user.firstName
          }, ברוך הבא לספוטפארק, אפליקציית ניהול החניונים החדשה של חברת החשמל.\n
לאחר הורדת האפליקציה החדשה מהחנות יש למלא את  שם המשתמש שלך (מספר הטלפון הנייד שלך) והסיסמה הזמנית שלך היא: ${isDev ? "Dev@12345" : password
          }.
לאחר מכן, יש לשנות סיסמא כפי שמוסבר במצגת שקיבלת.\n
בכל בעיה שאתה נתקל אתה מוזמן לפנות לאדמיניסטרטור שלך כפי שמופיע במצגת שקיבלת.\n
לאחר ההתחברות לאפליקציה החדשה יש להסיר את האפליקציה הישנה שיצאה משימוש.\n
שים לב כי מספר החניה שלך נשאר ללא שינוי.\n
בהצלחה לכולנו!`;
        const messageSent = await sendMessage(user.phoneNumber, sms);
        res.status(HttpStatus.CREATED).json({
          phoneNumber,
          message: HttpMessages.user.TEMPORARY_PASSWORD_CREATED_PHONE,
          messageSent,
        });
      }
    } else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const loginWithTemporaryPasswordService = async (req: Request, res: Response) => {
  try {
    const {
      phoneNumber,
      temporaryPassword,
      newPassword,
      confirmPassword,
    }: {
      phoneNumber: { countryCode: string; number: string };
      temporaryPassword: string;
      newPassword: string;
      confirmPassword: string;
    } = req.body;
    const user = await User.findOne({
      "phoneNumber.number": phoneNumber.number,
    });
    if (user) {
      if (user.failedLoginAttempts > 2) {
        res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.user.ACCOUNT_LOCKED });
      } else {
        const isValid = isPasswordValid(
          temporaryPassword,
          user.temporaryPassword ? user.temporaryPassword : ""
        );
        if (isValid) {
          // if (new Date(user.temporaryPasswordExpiry) > new Date()) {
          if (newPassword === confirmPassword) {
            const hashedPassword = await generateHashPassword(newPassword);
            user.password = hashedPassword;
            user.temporaryPassword = null;
            user.passwordUpdatedAt = Date.now();
            await user.save();
            res
              .status(HttpStatus.OK)
              .json({ message: HttpMessages.user.PASSWORD_UPDATE_SUCCESS });
          } else
            res
              .status(HttpStatus.FORBIDDEN)
              .json({ message: HttpMessages.validations.PASSWORDS_NOT_MATCH });
          // } else res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.validations.TEMP_PASSWORD_EXPIRED })
        } else
          res.status(HttpStatus.UNAUTHORIZED).json({ message: HttpMessages.user.FAILED_TEMPORARY_LOGIN });
      }
    } else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const resetPasswordService = async (req: Request, res: Response) => {
  try {
    const {
      phoneNumber,
      oldPassword,
      newPassword,
      confirmPassword,
    }: {
      phoneNumber: { countryCode: string; number: string };
      oldPassword: string;
      newPassword: string;
      confirmPassword: string;
    } = req.body;
    const user = await User.findOne({
      "phoneNumber.number": phoneNumber.number,
    });
    if (user) {
      if (user.failedLoginAttempts > 2) {
        res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.user.ACCOUNT_LOCKED });
      } else {
        const isValid = isPasswordValid(oldPassword, user.password);
        if (isValid) {
          if (newPassword === confirmPassword) {
            const hashedPassword = await generateHashPassword(newPassword);
            user.password = hashedPassword;
            user.temporaryPassword = null;
            user.passwordUpdatedAt = Date.now();
            await user.save();
            res
              .status(HttpStatus.OK)
              .json({ message: HttpMessages.user.PASSWORD_UPDATE_SUCCESS });
          } else
            res
              .status(HttpStatus.FORBIDDEN)
              .json({ message: HttpMessages.validations.PASSWORDS_NOT_MATCH });
        } else
          res.status(HttpStatus.UNAUTHORIZED).json({ message: HttpMessages.user.WRONG_OLD_PASSWORD });
      }
    } else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const unlockUserAccountService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const user = await User.findByIdAndUpdate(_id, {
      failedLoginAttempts: 0,
      failedAttemptsDates: [],
    });
    if (user)
      res.status(HttpStatus.OK).json({ message: HttpMessages.user.ACCOUNT_UNLOCKED });
    else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const deleteAllUsersService = async (req: Request, res: Response) => {
  try {
    const users = await User.find();
    const deleted = await Promise.all(
      users.map(async user => {
        return new Promise(async resolve => {
          const { designatedParkingSpot } = user;
          let designatedFreed = null;
          if (designatedParkingSpot && designatedParkingSpot.length > 0) {
            designatedFreed = await ParkingSpot.updateMany(
              { _id: { $in: designatedParkingSpot } },
              { designatedTo: null, userPresenceDate: null }
            );
          }
          const reservationFreed = await Reservation.updateMany(
            {
              user: user._id,
              parkingStatus: {
                $in: [reservationEnums[0], reservationEnums[1]],
              },
            },
            { parkingStatus: reservationEnums[4] }
          );
          const vehiclesDeleted = await Vehicle.deleteMany({ user: user._id });
          const deletedUser = await User.findByIdAndDelete(user._id);
          resolve({
            deletedUser,
            reservationFreed,
            vehiclesDeleted,
            designatedFreed,
          });
        });
      })
    );
    res.status(HttpStatus.OK).json({ message: HttpMessages.user.ALL_USERS_DELETED, deleted });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};

export const deleteUserService = async (req: Request, res: Response) => {
  try {
    const { _id }: { _id?: string } = req.params;
    const user = await User.findByIdAndDelete(_id);
    if (user) {
      const { designatedParkingSpot, vehicles } = user;
      if (designatedParkingSpot && designatedParkingSpot.length > 0) {
        const designatedFreed = await ParkingSpot.updateMany(
          { _id: { $in: designatedParkingSpot } },
          { designatedTo: null, userPresenceDate: null }
        );
      }
      const reservationFreed = await Reservation.updateMany(
        {
          user: _id,
          parkingStatus: {
            $in: [
              reservationEnums[0],
              reservationEnums[1],
              reservationEnums[2],
            ],
          },
        },
        { parkingStatus: reservationEnums[4] }
      );
      const deletedVehicles = await Vehicle.deleteMany({
        $or: [{ user: _id }, { _id: { $in: vehicles } }],
      });
      res.status(HttpStatus.OK).json({ message: HttpMessages.user.USER_DELETED_SUCCESS });
    } else
      res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
  } catch (err) {
    res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
  }
};
