
import { Request, Response, } from 'express';
import HttpMessages from '../../services/constants/http-messages';
import HttpStatus from '../../services/constants/http-statuses';
import Vehicle from '../../models/vehicle';

export const createVehicleService = async (req: Request, res: Response) => {
    try {
        const { user, make, model, registration }: { user?: string, make?: string, model?: string, registration?: string } = req.body;
        if (user && make && model && registration) {
            const vehicle = new Vehicle({
                user,
                make,
                model,
                registration: registration.toUpperCase(),
            });
            await vehicle.save();
            res.status(HttpStatus.CREATED).json(vehicle);
        } else
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_DETAILS_MISSING });
    } catch (err: any) {
        if (err?.code == 11000)
            res.status(HttpStatus.FORBIDDEN).json({ message: HttpMessages.vehicle.VEHICLE_REGISTRATION_DUPLICATE })
        else
            res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const getAllVehiclesService = async (req: Request, res: Response) => {
    try {
        const { page = 1, limit = 10 }: { page?: number, limit?: number } = req.query
        const totalCount = await Vehicle.countDocuments();
        const totalPages = Math.ceil(totalCount / limit);
        const vehicles = await Vehicle.find().sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
        res.status(HttpStatus.OK).json({ totalPages, totalCount, vehicles });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const getSingleVehicleService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const vehicle = await Vehicle.findById(_id);
        if (vehicle)
            res.status(HttpStatus.OK).json(vehicle);
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.vehicle.VEHICLE_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const getUserVehiclesService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { page = 1, limit = 10 }: { page?: number, limit?: number } = req.query
        const totalCount = await Vehicle.countDocuments({ user: _id });
        const totalPages = Math.ceil(totalCount / limit);
        const vehicles = await Vehicle.find({ user: _id }).sort({ updatedAt: -1 }).limit(limit).skip(limit * (page - 1))
        res.status(HttpStatus.OK).json({ totalPages, totalCount, vehicles });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const getUserFromRegistrationService = async (req: Request, res: Response) => {
    try {
        const { registration }: { registration?: string, } = req.params
        const query = { registration: { $regex: new RegExp(registration.toUpperCase(), 'i') } }
        const vehicle = await Vehicle.findOne(query).sort({ updatedAt: -1 }).populate({ path: 'user', select: 'phoneNumber', })
        if (vehicle) res.status(HttpStatus.OK).json(vehicle);
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.user.USER_NOT_FOUND });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const searchFromRegistrationService = async (req: Request, res: Response) => {
    try {
        const { registration }: { registration?: string } = req.query;
        if (!registration) res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_DETAILS_MISSING })
        else {
            const regex = new RegExp(registration, 'i')
            const vehicles = await Vehicle.find({ registration: { $regex: regex } })
            // const users = await User.find({ 'vehicle.registration': { $regex: regex } });
            // const vehicles = await Promise.all(users.map(user => user.vehicle))
            res.status(HttpStatus.OK).json(vehicles);
        }
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const updateVehicleService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const { make, model, registration }: { make?: string, model?: string, registration?: string } = req.body;
        if (make && model && registration) {
            const vehicle = await Vehicle.findByIdAndUpdate(_id, { make, model, registration, }, { new: true });
            res.status(HttpStatus.OK).json(vehicle);
        } else
            res.status(HttpStatus.BAD_REQUEST).json({ message: HttpMessages.vehicle.VEHICLE_DETAILS_MISSING });
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}

export const deleteVehicleService = async (req: Request, res: Response) => {
    try {
        const { _id }: { _id?: string } = req.params;
        const vehicle = await Vehicle.findByIdAndDelete(_id);
        if (vehicle) res.status(HttpStatus.OK).json({ message: HttpMessages.vehicle.VEHICLE_DELETED_SUCCESS });
        else res.status(HttpStatus.NOT_FOUND).json({ message: HttpMessages.vehicle.VEHICLE_NOT_FOUND })
    } catch (err) {
        res.status(HttpStatus.INTERNAL_SERVER_ERROR).json({ message: err });
    }
}